
(doc=>{
	Array.prototype.unique = function() { return this.filter((v, i, a) => a.indexOf(v) === i); };
	Array.prototype.validUnique = function(fn) {
		return this.filter((v, i, a) => (fn ? fn(v) : v) && a.indexOf(v) === i);
	};

	// full = Monday, December 12, 2022 at 12:16:21 PM Central European Standard Time
	// long = December 12, 2022 at 12:16:21 PM GMT+1
	// medium = Dec 12, 2022, 12:16:21 PM
	// short = 12/12/22, 12:16 PM
	let formats = {
//		LT   : {timeStyle: 'short'}, // Issue in Safari
		LT   : {hour: 'numeric', minute: 'numeric'},
		LLL  : {dateStyle: 'long', timeStyle: 'short'}
	};

	// Format momentjs/PHP date formats to Intl.DateTimeFormat
	// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl/DateTimeFormat
	Date.prototype.format = function (options, UTC, hourCycle) {
		if (typeof options == 'string') {
			if (formats[options]) {
				options = formats[options];
			} else {
				console.log('Date.format('+options+')');
				options = {};
			}
		}
		if (hourCycle) {
			options.hourCycle = hourCycle;
		}
		let el = doc.documentElement;
		return this.toLocaleString(el.dataset.dateLang || el.lang, options);
	};

	Element.prototype.closestWithin = function(selector, parent) {
		const el = this.closest(selector);
		return (el && el !== parent && parent.contains(el)) ? el : null;
	};

	Element.fromHTML = string => {
		const template = doc.createElement('template');
		template.innerHTML = string.trim();
		return template.content.firstChild;
	};

	/**
	 * Every time the function is executed,
	 * it will delay the execution with the given amount of milliseconds.
	 */
	if (!Function.prototype.debounce) {
		Function.prototype.debounce = function(ms) {
			let func = this, timer;
			return function(...args) {
				timer && clearTimeout(timer);
				timer = setTimeout(()=>{
					func.apply(this, args);
					timer = 0;
				}, ms);
			};
		};
	}

	/**
	 * No matter how many times the event is executed,
	 * the function will be executed only once, after the given amount of milliseconds.
	 */
	if (!Function.prototype.throttle) {
		Function.prototype.throttle = function(ms) {
			let func = this, timer;
			return function(...args) {
				timer = timer || setTimeout(()=>{
						func.apply(this, args);
						timer = 0;
					}, ms);
			};
		};
	}

})(document);

/**
 * Modified version of https://github.com/Bernardo-Castilho/dragdroptouch
 * This is to only support Firefox Mobile.
 * Because touchstart must call preventDefault() to prevent scrolling
 * but then it doesn't work native in Chrome on Android
 */

(doc => {
	let ua = navigator.userAgent.toLowerCase();
	// Chrome on mobile supports drag & drop
	if (ua.includes('mobile') && ua.includes('gecko/')) {

		let opt = { passive: false, capture: false },

			dropEffect = 'move',
			effectAllowed = 'all',
			data = {},

			dataTransfer,
			dragSource,
			isDragging,
			allowDrop,
			lastTarget,
			lastTouch,
			holdInterval,

			img;

/*
		class DataTransferItem
		{
			get kind() { return 'string'; }
		}
*/
		/** https://developer.mozilla.org/en-US/docs/Web/API/DataTransfer */
		class DataTransfer
		{
			get dropEffect() { return dropEffect; }
			set dropEffect(value) { dropEffect = value; }

			get effectAllowed() { return effectAllowed; }
			set effectAllowed(value) { effectAllowed = value; }

			get files() { return []; }
			get items() { return []; } // DataTransferItemList
			get types() { return Object.keys(data); }

			clearData(type) {
				if (type != null) {
					delete data[type];
				} else {
					data = {};
				}
			}

			getData(type) {
				return data[type] || '';
			}

			setData(type, value) {
				data[type] = value;
			}

			constructor() {
				this.setDragImage = setDragImage;
			}
		}

		const
		htmlDrag = b => doc.documentElement.classList.toggle('firefox-drag', b),

		setDragImage = (src, xOffset, yOffset) => {
			img?.remove();
			if (src) {
				// create drag image from custom element or drag source
				img = src.cloneNode(true);
				copyStyle(src, img);
				img._x = xOffset == null ? src.clientWidth / 2 : xOffset;
				img._y = yOffset == null ? src.clientHeight / 2 : yOffset;
			}
		},

		// clear all members
		reset = () => {
			if (dragSource) {
				clearInterval(holdInterval);
				// dispose of drag image element
				img?.remove();
				isDragging && dispatchEvent(lastTouch, 'dragend', dragSource);
				img = dragSource = lastTouch = lastTarget = dataTransfer = holdInterval = null;
				isDragging = allowDrop = false;
				htmlDrag(false);
			}
		},

		// get point for a touch event
		getPoint = e => {
			e = e.touches ? e.touches[0] : e;
			return { x: e.clientX, y: e.clientY };
		},

		touchend = e => {
			if (dragSource) {
				// finish dragging
				allowDrop && 'touchcancel' !== e.type && dispatchEvent(lastTouch, 'drop', lastTarget);
				reset();
			}
		},

		// get the element at a given touch event
		getTarget = pt => {
			let el = doc.elementFromPoint(pt.x, pt.y);
			while (el && getComputedStyle(el).pointerEvents == 'none') {
				el = el.parentElement;
			}
			return el;
		},

		// move the drag image element
		moveImage = pt => {
			requestAnimationFrame(() => {
				if (img) {
					img.style.left = Math.round(pt.x - img._x) + 'px';
					img.style.top = Math.round(pt.y - img._y) + 'px';
				}
			});
		},

		copyStyle = (src, dst) => {
			// remove potentially troublesome attributes
			['id','class','style','draggable'].forEach(att => dst.removeAttribute(att));
			// copy canvas content
			if (src instanceof HTMLCanvasElement) {
				let cSrc = src, cDst = dst;
				cDst.width = cSrc.width;
				cDst.height = cSrc.height;
				cDst.getContext('2d').drawImage(cSrc, 0, 0);
			}
			// copy style (without transitions)
			let cs = getComputedStyle(src);
			Object.entries(cs).forEach(([key, value]) => key.includes('transition') || (dst.style[key] = value));
			dst.style.pointerEvents = 'none';
			// and repeat for all children
			let i = src.children.length;
			while (i--) copyStyle(src.children[i], dst.children[i]);
		},

		// return false when cancelled
		dispatchEvent = (e, type, target) => {
			if (e && target) {
				let evt = new Event(type, {bubbles:true,cancelable:true});
				evt.button = 0;
				evt.buttons = 1;
				// copy event properties into new event
				['altKey','ctrlKey','metaKey','shiftKey'].forEach(k => evt[k] = e[k]);
				let src = e.touches ? e.touches[0] : e;
				['pageX','pageY','clientX','clientY','screenX','screenY','offsetX','offsetY'].forEach(k => evt[k] = src[k]);
				if (dragSource) {
					evt.dataTransfer = dataTransfer;
				}
				return target.dispatchEvent(evt);
			}
			return false;
		};

/*
		doc.addEventListener('pointerdown', e => {
			doc.addEventListener('pointermove', e => {
				e.clientX
			});
			doc.setPointerCapture(e.pointerId);
		});
		doc.addEventListener('pointerup', e => {
			doc.releasePointerCapture(e.pointerId);
		});
*/
		doc.addEventListener('touchstart', e => {
			// clear all variables
			reset();
			// ignore events that have been handled or that involve more than one touch
			if (e && !e.defaultPrevented && e.touches && e.touches.length < 2) {
				// get nearest draggable element
				dragSource = e.target.closest('[draggable]');
				if (dragSource) {
					// get ready to start dragging
					lastTouch = e;
//					dragSource.style.userSelect = 'none';

					// 1000 ms to wait, chrome on android triggers dragstart in 600
					holdInterval = setTimeout(() => {
						// start dragging
						dataTransfer = new DataTransfer();
						if ((isDragging = dispatchEvent(e, 'dragstart', dragSource))) {
							htmlDrag(true);

							let pt = getPoint(e);

							// create drag image from custom element or drag source
							img || setDragImage(dragSource);
							let style = img.style;
							style.top = style.left = '-9999px';
							style.position = 'fixed';
							style.pointerEvents = 'none';
							style.zIndex = '999999999';
							// add image to document
							moveImage(pt);
							doc.body.append(img);

							dispatchEvent(e, 'dragenter', getTarget(pt));
						} else {
							reset();
						}
					}, 1000);
				}
			}
		}, opt);

		doc.addEventListener('touchmove', e => {
			if (isDragging) {
				// continue dragging
				let pt = getPoint(e),
					target = getTarget(pt);
				lastTouch = e;
				if (target != lastTarget) {
					dispatchEvent(e, 'dragleave', lastTarget);
					dispatchEvent(e, 'dragenter', target);
					lastTarget = target;
				}
				moveImage(pt);
				allowDrop = !dispatchEvent(e, 'dragover', target);
			} else {
				reset();
			}
		}, opt);

		doc.addEventListener('touchend', touchend);
		doc.addEventListener('touchcancel', touchend);
	}

})(document);


(win => {

let
	scope = {},
	_scope = 'all';

const
	doc = document,
	// On Mac we use ⌘ else the Ctrl key
	meta = /Mac OS X/.test(navigator.userAgent) ? 'meta' : 'ctrl',
	_scopes = {
		all: {}
	},
	toArray = v => Array.isArray(v) ? v : v.split(/\s*,\s*/),

	exec = (event, cmd) => {
		try {
			// call the handler and stop the event if neccessary
			if (!event.defaultPrevented && cmd(event) === false) {
				event.preventDefault();
				event.stopPropagation();
			}
		} catch (e) {
			console.error(e);
		}
	},

	shortcuts = {
		on: () => doc.addEventListener('keydown', keydown),
		off: () => doc.removeEventListener('keydown', keydown),
		add: (keys, modifiers, scopes, method) => {
			if (null == method) {
				method = scopes;
				scopes = 'all';
			}
			toArray(scopes).forEach(scope => {
				if (!_scopes[scope]) {
					_scopes[scope] = {};
				}
				toArray(keys).forEach(key => {
					key = key.toLowerCase();
					if (!_scopes[scope][key]) {
						_scopes[scope][key] = {};
					}
					modifiers = toArray(modifiers)
						.map(key => 'meta' == key ? meta : key)
						.unique().sort().join('+');
					if (!_scopes[scope][key][modifiers]) {
						_scopes[scope][key][modifiers] = [];
					}
					_scopes[scope][key][modifiers].push(method);
				});
			});
		},
		setScope: value => {
			_scope = value || 'all';
			scope = _scopes[_scope] || {};
			console.log('Shortcuts scope set to: ' + _scope);
		},
		getScope: () => _scope,
		getMetaKey: () => 'meta' === meta ? '⌘' : 'Ctrl'
	},

	keydown = event => {
		let key = (event.key || '').toLowerCase().replace(' ','space'),
			modifiers = ['alt','ctrl','meta','shift'].filter(v => event[v+'Key']).join('+');
		scope[key]?.[modifiers]?.forEach(cmd => exec(event, cmd));
		!event.defaultPrevented && _scope !== 'all' && _scopes.all[key]?.[modifiers]?.forEach(cmd => exec(event, cmd));
	};

win.shortcuts = shortcuts;

shortcuts.on();

})(this);

/*!!
 * Hasher <http://github.com/millermedeiros/hasher>
 * @author Miller Medeiros
 * @version 1.1.2 (2012/10/31 03:19 PM)
 * Released under the MIT License
 */

(global => {

    //--------------------------------------------------------------------------------------
    // Private
    //--------------------------------------------------------------------------------------

    const
    _hashValRegexp = /#(.*)$/,
    _hashRegexp = /^[#/]+/,
    _hashTrim = /^\/+/g,
    _trimHash = hash => hash?.replace(_hashTrim, '') || '',
    _getWindowHash = () => {
        //parsed full URL instead of getting window.location.hash because Firefox decode hash value (and all the other browsers don't)
        var result = _hashValRegexp.exec( location.href );
        return result?.[1] ? decodeURIComponent(result[1]) : '';
    },
    _registerChange = newHash => {
        if (_hash !== newHash) {
            var oldHash = _hash;
            _hash = newHash; //should come before event dispatch to make sure user can get proper value inside event handler
            _dispatch(_trimHash(newHash), _trimHash(oldHash));
        }
    },
    _setHash = (path, replace) => {
        path = path ? '/' + path.replace(_hashRegexp, '') : path;
        if (path !== _hash){
            // we should store raw value
            _registerChange(path);
            if (path === _hash) {
                path = '#' + encodeURI(path)
                // we check if path is still === _hash to avoid error in
                // case of multiple consecutive redirects [issue #39]
                replace
                    ? location.replace(path)
                    : (location.hash = path);
            }
        }
    },
    _dispatch = (...args) => hasher.active && _bindings.forEach(callback => callback(...args)),

    //--------------------------------------------------------------------------------------
    // Public (API)
    //--------------------------------------------------------------------------------------

    hasher = /** @lends hasher */ {
        clear : () => {
            _bindings = [];
            hasher.active = true;
        },

        /**
         * Signal dispatched when hash value changes.
         * - pass current hash as 1st parameter to listeners and previous hash value as 2nd parameter.
         * @type signals.Signal
         */
        active : true,
        add : callback => _bindings.push(callback),

        /**
         * Start listening/dispatching changes in the hash/history.
         * <ul>
         *   <li>hasher won't dispatch CHANGE events by manually typing a new value or pressing the back/forward buttons before calling this method.</li>
         * </ul>
         */
        init : () => _dispatch(_trimHash(_hash)),

        /**
         * Set Hash value, generating a new history record.
         * @param {...string} path    Hash value without '#'.
         * @example hasher.setHash('lorem/ipsum/dolor') -> '#/lorem/ipsum/dolor'
         */
        setHash : path => _setHash(path),

        /**
         * Set Hash value without keeping previous hash on the history record.
         * @param {...string} path    Hash value without '#'.
         * @example hasher.replaceHash('lorem/ipsum/dolor') -> '#/lorem/ipsum/dolor'
         */
        replaceHash : path => _setHash(path, true)
    };

    var _hash = _getWindowHash(),
        _bindings = [];

    addEventListener('hashchange', () => _registerChange(_getWindowHash()));

    global.hasher = hasher;
})(this);

/** @license
 * Crossroads.js <http://millermedeiros.github.com/crossroads.js>
 * Released under the MIT license
 * Author: Miller Medeiros
 * Version: 0.7.1 - Build: 93 (2012/02/02 09:29 AM)
 */

(global => {

    const isFunction = obj => typeof obj === 'function';

    // Crossroads --------
    //====================

    global.Crossroads = class Crossroads {

        constructor() {
            this._routes = [];
        }

        addRoute(pattern, callback) {
            var route = new Route(pattern, callback, this);
            this._routes.push(route);
            return route;
        }

        parse(request) {
            request = request || '';
            var i = 0,
                routes = this._routes,
                n = routes.length,
                route;
            //should be decrement loop since higher priorities are added at the end of array
            while (n--) {
                route = routes[n];
                if ((!i || route.greedy) && route.match(request)) {
                    route.callback?.(...route._getParamsArray(request));
                    ++i;
                }
            }
        }
    }

    // Route --------------
    //=====================

    class Route {

        constructor(pattern, callback, router) {
            var isRegexPattern = pattern instanceof RegExp;
            Object.assign(this, {
                greedy: false,
                rules: {},
                _router: router,
                _pattern: pattern,
                _paramsIds: isRegexPattern ? null : captureVals(PARAMS_REGEXP, pattern),
                _optionalParamsIds: isRegexPattern ? null : captureVals(OPTIONAL_PARAMS_REGEXP, pattern),
                _matchRegexp: isRegexPattern ? pattern : compilePattern(pattern),
                callback: isFunction(callback) ? callback : null
            });
        }

        match(request) {
            // validate params even if regexp.
            var values = this._getParamsObject(request);
            return this._matchRegexp.test(request)
             && 0 == Object.entries(this.rules).filter(([key, validationRule]) => {
                var val = values[key],
                    isValid = false;
                if (key === 'normalize_'
                 || (val == null && this._optionalParamsIds?.includes(key))) {
                    isValid = true;
                }
                else if (validationRule instanceof RegExp) {
                    isValid = validationRule.test(val);
                }
                else if (Array.isArray(validationRule)) {
                    isValid = validationRule.includes(val);
                }
                else if (isFunction(validationRule)) {
                    isValid = validationRule(val, request, values);
                }
                // fail silently if validationRule is from an unsupported type
                return !isValid;
            }).length;
        }

        _getParamsObject(request) {
            var values = getParamValues(request, this._matchRegexp) || [],
                n = values.length;
            if (this._paramsIds) {
                while (n--) {
                    values[this._paramsIds[n]] = values[n];
                }
            }
            return values;
        }

        _getParamsArray(request) {
            var norm = this.rules.normalize_;
            return isFunction(norm)
                ? norm(request, this._getParamsObject(request))
                : getParamValues(request, this._matchRegexp);
        }

    }



    // Pattern Lexer ------
    //=====================

    const
        ESCAPE_CHARS_REGEXP = /[\\.+*?^$[\](){}/'#]/g, //match chars that should be escaped on string regexp
        UNNECESSARY_SLASHES_REGEXP = /\/$/g, //trailing slash
        OPTIONAL_SLASHES_REGEXP = /([:}]|\w(?=\/))\/?(:)/g, //slash between `::` or `}:` or `\w:`. $1 = before, $2 = after
        REQUIRED_SLASHES_REGEXP = /([:}])\/?(\{)/g, //used to insert slash between `:{` and `}{`

        REQUIRED_PARAMS_REGEXP = /\{([^}]+)\}/g, //match everything between `{ }`
        OPTIONAL_PARAMS_REGEXP = /:([^:]+):/g, //match everything between `: :`
        PARAMS_REGEXP = /(?:\{|:)([^}:]+)(?:\}|:)/g, //capture everything between `{ }` or `: :`

        //used to save params during compile (avoid escaping things that
        //shouldn't be escaped).
        SAVE_REQUIRED_PARAMS = '__CR_RP__',
        SAVE_OPTIONAL_PARAMS = '__CR_OP__',
        SAVE_REQUIRED_SLASHES = '__CR_RS__',
        SAVE_OPTIONAL_SLASHES = '__CR_OS__',
        SAVED_REQUIRED_REGEXP = new RegExp(SAVE_REQUIRED_PARAMS, 'g'),
        SAVED_OPTIONAL_REGEXP = new RegExp(SAVE_OPTIONAL_PARAMS, 'g'),
        SAVED_OPTIONAL_SLASHES_REGEXP = new RegExp(SAVE_OPTIONAL_SLASHES, 'g'),
        SAVED_REQUIRED_SLASHES_REGEXP = new RegExp(SAVE_REQUIRED_SLASHES, 'g'),

        captureVals = (regex, pattern) => {
            var vals = [], match;
            while ((match = regex.exec(pattern))) {
                vals.push(match[1]);
            }
            return vals;
        },

        getParamValues = (request, regexp) => {
            var vals = regexp.exec(request);
            vals?.shift();
            return vals;
        },
        compilePattern = pattern => {
            return new RegExp('^' + (pattern
                ? pattern
                    // tokenize, save chars that shouldn't be escaped
                    .replace(UNNECESSARY_SLASHES_REGEXP, '')
                    .replace(OPTIONAL_SLASHES_REGEXP, '$1'+ SAVE_OPTIONAL_SLASHES +'$2')
                    .replace(REQUIRED_SLASHES_REGEXP, '$1'+ SAVE_REQUIRED_SLASHES +'$2')
                    .replace(OPTIONAL_PARAMS_REGEXP, SAVE_OPTIONAL_PARAMS)
                    .replace(REQUIRED_PARAMS_REGEXP, SAVE_REQUIRED_PARAMS)
                    .replace(ESCAPE_CHARS_REGEXP, '\\$&')
                    // untokenize
                    .replace(SAVED_OPTIONAL_SLASHES_REGEXP, '\\/?')
                    .replace(SAVED_REQUIRED_SLASHES_REGEXP, '\\/')
                    .replace(SAVED_OPTIONAL_REGEXP, '([^\\/]+)?/?')
                    .replace(SAVED_REQUIRED_REGEXP, '([^\\/]+)')
                : ''
            ) + '/?$'); //trailing slash is optional
        };

})(this);

/* RainLoop Webmail (c) RainLoop Team | MIT */
(doc => {
	const
		defined = v => undefined !== v,
		/**
		 * @param {*} aItems
		 * @param {Function} fFileCallback
		 * @param {number=} iLimit = 20
		 */
		getDataFromFiles = (aItems, fFileCallback, iLimit) =>
		{
			if (aItems?.length)
			{
				let
					oFile,
					iCount = 0,
					bCallLimit = false
				;

				[...aItems].forEach(oItem => {
					if (oItem) {
						if (iLimit && iLimit < ++iCount) {
							if (!bCallLimit) {
								bCallLimit = true;
//								fLimitCallback(iLimit);
							}
						} else {
							oFile = getDataFromFile(oItem);
							oFile && fFileCallback(oFile);
						}
					}
				});
			}
		},

		addEventListeners = (element, obj) =>
			Object.entries(obj).forEach(([key, value]) => element.addEventListener(key, value)),

		/**
		 * @param {*} oFile
		 * @return {Object}
		 */
		getDataFromFile = oFile =>
		{
			return oFile.size
				? {
					fileName: (oFile.name || '').replace(/^.*\/([^/]*)$/, '$1'),
					size: oFile.size,
					file: oFile
				}
				: null; // Folder
		},

		eventContainsFiles = oEvent => oEvent.dataTransfer.types.includes('Files');

	class Queue extends Array
	{
		push(fn, ...args) {
			super.push([fn, args]);
			this.call();
		}
		call() {
			if (!this.running) {
				this.running = true;
				let f;
				while ((f = this.shift())) f[0](...f[1]);
				this.running = false;
			}
		}
	}

	/**
	 * @constructor
	 * @param {Object=} options
	 */
	class Jua
	{
		constructor(options)
		{
			let timer,
				el = options.clickElement;

			const self = this,
				timerStart = fn => {
					timerStop();
					timer = setTimeout(fn, 200);
				},
				timerStop = () => {
					timer && clearTimeout(timer);
					timer = 0;
				};

			self.oEvents = {
				onSelect: null,
				onStart: null,
				onComplete: null,
				onProgress: null,
				onDragEnter: null,
				onDragLeave: null,
				onBodyDragEnter: null,
				onBodyDragLeave: null
			};

			self.oXhrs = {};
			self.oUids = {};
			self.options = Object.assign({
					action: '',
					name: 'uploader',
					limit: 0,
//					clickElement:
//					dragAndDropElement:
				}, options || {});
			self.oQueue = new Queue();

			// clickElement
			if (el) {
				el.style.position = 'relative';
				el.style.overflow = 'hidden';
				if ('inline' === el.style.display) {
					el.style.display = 'inline-block';
				}

				self.generateNewInput(el);
			}

			el = options.dragAndDropElement;
			if (el) {
				addEventListeners(doc, {
					dragover: oEvent => {
						if (eventContainsFiles(oEvent)) {
							timerStop();
							if (el.contains(oEvent.target)) {
								oEvent.dataTransfer.dropEffect = 'copy';
								oEvent.stopPropagation();
							} else {
								oEvent.dataTransfer.dropEffect = 'none';
							}
							oEvent.preventDefault();
						}
					},
					dragenter: oEvent => {
						if (eventContainsFiles(oEvent)) {
							timerStop();
							oEvent.preventDefault();
							self.runEvent('onBodyDragEnter', oEvent);
							if (el.contains(oEvent.target)) {
								timerStop();
								self.runEvent('onDragEnter', el, oEvent);
							}
						}
					},
					dragleave: oEvent => {
						if (eventContainsFiles(oEvent)) {
							let oRelatedTarget = doc.elementFromPoint(oEvent.clientX, oEvent.clientY);
							if (!oRelatedTarget || !el.contains(oRelatedTarget)) {
								self.runEvent('onDragLeave', el, oEvent);
							}
							timerStart(() => self.runEvent('onBodyDragLeave', oEvent))
						}
					},
					drop: oEvent => {
						if (eventContainsFiles(oEvent)) {
							timerStop();
							oEvent.preventDefault();
							if (el.contains(oEvent.target)) {
								getDataFromFiles(
									oEvent.files || oEvent.dataTransfer.files,
									oFile => {
										if (oFile) {
											self.addFile(oFile);
										}
									},
									self.options.limit
								);
							}
						}
						self.runEvent('onDragLeave', oEvent);
						self.runEvent('onBodyDragLeave', oEvent);
					}
				});
			}
		}

		/**
		 * @param {string} sName
		 * @param {Function} fFunc
		 */
		on(sName, fFunc)
		{
			this.oEvents[sName] = fFunc;
			return this;
		}

		/**
		 * @param {string} sName
		 */
		runEvent(sName, ...aArgs)
		{
			this.oEvents[sName]?.apply(null, aArgs);
		}

		/**
		 * @param {string} sName
		 */
		getEvent(sName)
		{
			return this.oEvents[sName] || null;
		}

		/**
		 * @param {Object} oFileInfo
		 */
		addFile(oFileInfo)
		{
			const sUid = 'jua-uid-' + Jua.randomId(16) + '-' + (Date.now().toString()),
				fOnSelect = this.getEvent('onSelect');
			if (oFileInfo && (!fOnSelect || (false !== fOnSelect(sUid, oFileInfo))))
			{
				this.oUids[sUid] = true;
				this.oQueue.push((...args) => this.uploadTask(...args), sUid, oFileInfo);
			}
			else
			{
				this.cancel(sUid);
			}
		}

		/**
		 * @param {string} sUid
		 * @param {?} oFileInfo
		 */
		uploadTask(sUid, oFileInfo)
		{
			if (false === this.oUids[sUid] || !oFileInfo || !oFileInfo.file)
			{
				return false;
			}

			try
			{
				const
					self = this,
					oXhr = new XMLHttpRequest(),
					oFormData = new FormData(),
					sAction = this.options.action,
					fStartFunction = this.getEvent('onStart'),
					fProgressFunction = this.getEvent('onProgress')
				;

				oXhr.open('POST', sAction, true);

				if (fProgressFunction && oXhr.upload)
				{
					oXhr.upload.onprogress = oEvent => {
						if (oEvent && oEvent.lengthComputable && defined(oEvent.loaded) && defined(oEvent.total))
						{
							fProgressFunction(sUid, oEvent.loaded, oEvent.total);
						}
					};
				}

				oXhr.onreadystatechange = () => {
					if (4 === oXhr.readyState)
					{
						delete self.oXhrs[sUid];
						let bResult = false,
							oResult = null;
						if (200 === oXhr.status)
						{
							try
							{
								oResult = JSON.parse(oXhr.responseText);
								bResult = true;
							}
							catch (e)
							{
								console.error(e);
							}
						}
						this.getEvent('onComplete')(sUid, bResult, oResult);
					}
				};

				fStartFunction && fStartFunction(sUid);

				oFormData.append(this.options.name, oFileInfo.file);

				oXhr.send(oFormData);

				this.oXhrs[sUid] = oXhr;
				return true;
			}
			catch (oError)
			{
				console.error(oError)
			}

			return false;
		}

		generateNewInput(oClickElement)
		{
			if (oClickElement)
			{
				const self = this,
					limit = self.options.limit,
					oInput = doc.createElement('input'),
					onClick = ()=>oInput.click();

				oInput.type = 'file';
				oInput.tabIndex = -1;
				oInput.style.display = 'none';
				oInput.multiple = 1 != limit;

				oClickElement.addEventListener('click', onClick);

				oInput.addEventListener('input', () => {
					const fFileCallback = oFile => {
						self.addFile(oFile);
						setTimeout(() => {
							oInput.remove();
							oClickElement.removeEventListener('click', onClick);
							self.generateNewInput(oClickElement);
						}, 10);
					};
					if (oInput.files?.length) {
						getDataFromFiles(oInput.files, fFileCallback, limit);
					} else {
						fFileCallback({
							fileName: oInput.value.split(/\\\//).pop(),
							size: null,
							file : null
						});
					}
				});
			}
		}

		/**
		 * @param {string} sUid
		 */
		cancel(sUid)
		{
			this.oUids[sUid] = false;
			if (this.oXhrs[sUid])
			{
				try
				{
					this.oXhrs[sUid].abort && this.oXhrs[sUid].abort();
				}
				catch (oError)
				{
					console.error(oError);
				}

				delete this.oXhrs[sUid];
			}
		}
	}

	Jua.randomId = len => {
		let arr = new Uint8Array((len || 32) / 2);
		crypto.getRandomValues(arr);
		return arr.map(dec => dec.toString(16).padStart(2,'0')).join('');
	}

	this.Jua = Jua;

})(document);

/*!
	* Native JavaScript for Bootstrap v3.0.10 (https://thednp.github.io/bootstrap.native/)
	* Copyright 2015-2020 © dnp_theme
	* Licensed under MIT (https://github.com/thednp/bootstrap.native/blob/master/LICENSE)
	*/

(doc => {
	const
		setFocus = element => element.focus ? element.focus() : element.setActive(),
		isArrow = e => 'ArrowUp' === e.key || 'ArrowDown' === e.key;

	this.BSN = {
		Dropdown: function(toggleBtn) {
			let menu, menuItems = [];
			const self = this,
				parent = toggleBtn.parentNode,
				preventEmptyAnchor = e => {
					const t = e.target;
					('#' === (t.href || t.parentNode?.href)?.slice(-1)) && e.preventDefault();
				},
				open = bool => {
					menu?.classList.toggle('show', bool);
					parent.classList.toggle('show', bool);
					toggleBtn.setAttribute('aria-expanded', bool);
					toggleBtn.open = bool;
					if (bool) {
						toggleBtn.removeEventListener('click',clickHandler);
					} else {
						setTimeout(() => toggleBtn.addEventListener('click',clickHandler), 1);
					}
				},
				toggleEvents = () => {
					const action = (toggleBtn.open ? 'add' : 'remove') + 'EventListener';
					doc[action]('click',dismissHandler);
					doc[action]('keydown',preventScroll);
					doc[action]('keyup',keyHandler);
					doc[action]('focus',dismissHandler);
				},
				dismissHandler = e => {
					const eventTarget = e.target;
					if ((!menu.contains(eventTarget) && !toggleBtn.contains(eventTarget)) || e.type !== 'focus') {
						self.hide();
						preventEmptyAnchor(e);
					}
				},
				clickHandler = e => {
					self.show();
					preventEmptyAnchor(e);
				},
				preventScroll = e => isArrow(e) && e.preventDefault(),
				keyHandler = e => {
					if ('Escape' === e.key) {
						self.toggle();
					} else if (isArrow(e)) {
						let activeItem = doc.activeElement,
							isMenuButton = activeItem === toggleBtn,
							idx = isMenuButton ? 0 : menuItems.indexOf(activeItem);
						if (parent.contains(activeItem)) {
							if (!isMenuButton) {
								idx = 'ArrowUp' === e.key
									? (idx > 1 ? idx-1 : 0)
									: (idx < menuItems.length-1 ? idx+1 : idx);
							}
							menuItems[idx] && setFocus(menuItems[idx]);
						} else {
							console.log('activeElement not in menu');
						}
					}
				};
			self.show = () => {
				menu = parent.querySelector('.dropdown-menu');
				menuItems = [...menu.querySelectorAll('A')].filter(item => 'none' != item.parentNode.style.display);
				!('tabindex' in menu) && menu.setAttribute('tabindex', '0');
				open(true);
				setTimeout(() => {
					setFocus( menu.getElementsByTagName('INPUT')[0] || toggleBtn );
					toggleEvents();
				},1);
			};
			self.hide = () => {
				open(false);
				toggleEvents();
				setFocus(toggleBtn);
			};
			self.toggle = () => toggleBtn.open ? self.hide() : self.show();
			open(false);
			toggleBtn.Dropdown = self;
		}
	};

})(document);

/*!
 * Knockout JavaScript library v3.5.1-sm
 * (c) The Knockout.js team - http://knockoutjs.com/
 * License: MIT (http://www.opensource.org/licenses/mit-license.php)
 */

(V=>{function O(a,b){return null===a||ea[typeof a]?a===b:!1}function fa(a,b){var d;return()=>{d||(d=setTimeout(()=>{d=0;a()},b))}}function ha(a,b){var d;return()=>{clearTimeout(d);d=setTimeout(a,b)}}function ia(a,b){b?.dispose?.()}function ja(a,b){var d=this.Pb,f=d[x];f.Z||(this.Ra&&this.va[b]?(d.mb(b,a,this.va[b]),this.va[b]=null,--this.Ra):f.v[b]||d.mb(b,a,f.A?{W:a}:d.Hb(a)),a.ha&&a.Lb())}function W(a){a=c.g.j.get(a,H);var b=a?.F;b&&(a.F=null,b.Bb())}function X(a,b){for(var d,f=c.m.firstChild(b);d=
f;)f=c.m.nextSibling(d),Y(a,d);c.l.notify(b,c.l.D)}function Y(a,b){var d=a;if(1===b.nodeType||c.qb.ac(b))d=Z(b,null,a).bindingContextForDescendants;d&&!b.matches?.("SCRIPT,TEXTAREA,TEMPLATE")&&X(d,b)}function ka(a){var b=[],d={},f=[];c.g.M(a,function e(l){if(!d[l]){var g=c.h[l];g&&(g.after&&(f.push(l),g.after.forEach(k=>{if(a[k]){if(f.includes(k))throw Error("Cannot combine the following bindings, because they have a cyclic dependency: "+f.join(", "));e(k)}}),f.length--),b.push({key:l,xb:g}));d[l]=
!0}});return b}function Z(a,b,d){var f=c.g.j.Wa(a,H,{}),h=f.Mb;if(!b){if(h)throw Error("You cannot apply bindings multiple times to the same element.");f.Mb=!0}h||(f.context=d);f.Ya||(f.Ya={});if(b&&"function"!==typeof b)var l=b;else{var e=c.o(()=>{if(l=b?b(d,a):c.qb.Sb(a,d))d[I]?.(),d[aa]?.();return l},{s:a});l&&e.isActive()||(e=null)}var g=d,k;if(l){var n=e?m=>()=>e()[m]():m=>l[m],p={get:m=>l[m]&&n(m)(),has:m=>m in l};c.l.D in l&&c.l.subscribe(a,c.l.D,()=>{var m=l[c.l.D]();if(m){var q=c.m.childNodes(a);
q.length&&m(q,c.dataFor(q[0]))}});c.l.fa in l&&(g=c.l.cb(a,d),c.l.subscribe(a,c.l.fa,()=>{var m=l[c.l.fa]();m&&c.m.firstChild(a)&&m(a)}));ka(l).forEach(m=>{var q=m.xb.init,r=m.xb.update,t=m.key;if(8===a.nodeType&&!c.m.ba[t])throw Error("The binding '"+t+"' cannot be used with virtual elements");try{"function"==typeof q&&c.u.J(()=>{var u=q(a,n(t),p,g.$data,g);if(u&&u.controlsDescendantBindings){if(void 0!==k)throw Error("Multiple bindings ("+k+" and "+t+") are trying to control descendant bindings of the same element. You cannot use these bindings together on the same element.");
k=t}}),"function"==typeof r&&c.o(()=>r(a,n(t),p,g.$data,g),{s:a})}catch(u){throw u.message='Unable to process binding "'+t+": "+l[t]+'"\nMessage: '+u.message,u;}})}f=void 0===k;return{shouldBindDescendants:f,bindingContextForDescendants:f&&g}}function P(a,b){return a&&a instanceof c.da?a:new c.da(a,void 0,void 0,b)}var J=V.document,Q={},c="undefined"!==typeof Q?Q:{};c.S=(a,b)=>{a=a.split(".");for(var d=c,f=0,h=a.length-1;f<h;f++)d=d[a[f]];d[a[h]]=b};c.version="3.5.1-sm";c.g={extend:(a,b)=>b?Object.assign(a,
b):a,M:(a,b)=>a&&Object.entries(a).forEach(d=>b(d[0],d[1])),Va:a=>[...a.childNodes].forEach(b=>c.removeNode(b)),Ab:a=>{a=[...a];var b=(a[0]?.ownerDocument||J).createElement("div");a.forEach(d=>b.append(c.ea(d)));return b},ua:(a,b)=>Array.prototype.map.call(a,b?d=>c.ea(d.cloneNode(!0)):d=>d.cloneNode(!0)),qa:(a,b)=>{c.g.Va(a);b&&a.append(...b)},xa:(a,b)=>{if(a.length){for(b=8===b.nodeType&&b.parentNode||b;a.length&&a[0].parentNode!==b;)a.splice(0,1);for(;1<a.length&&a[a.length-1].parentNode!==b;)--a.length;
if(1<a.length){b=a[0];var d=a[a.length-1];for(a.length=0;b!==d;)a.push(b),b=b.nextSibling;a.push(d)}}return a},Gb:a=>null==a?"":a.trim?a.trim():a.toString().replace(/^[\s\xa0]+|[\s\xa0]+$/g,""),Ua:a=>a.ownerDocument.documentElement.contains(1!==a.nodeType?a.parentNode:a),Ib:(a,b)=>{if(!a?.nodeType)throw Error("element must be a DOM node when calling triggerEvent");a.dispatchEvent(new Event(b))},i:a=>c.T(a)?a():a,bb:(a,b)=>a.textContent=c.g.i(b)};c.S("utils",c.g);c.S("unwrap",c.g.i);(()=>{let a=0,
b="__ko__"+Date.now(),d=new WeakMap;c.g.j={get:(f,h)=>(d.get(f)||{})[h],set:(f,h,l)=>{d.has(f)?d.get(f)[h]=l:d.set(f,{[h]:l});return l},Wa(f,h,l){return this.get(f,h)||this.set(f,h,l)},clear:f=>d.delete(f),Y:()=>a++ +b}})();c.g.O=(()=>{var a=c.g.j.Y(),b={1:1,8:1,9:1},d={1:1,9:1};const f=(e,g)=>{var k=c.g.j.get(e,a);g&&!k&&(k=new Set,c.g.j.set(e,a,k));return k},h=e=>{var g=f(e);g&&(new Set(g)).forEach(k=>k(e));c.g.j.clear(e);d[e.nodeType]&&l(e.childNodes,!0)},l=(e,g)=>{for(var k=[],n,p=0;p<e.length;p++)if(!g||
8===e[p].nodeType)if(h(k[k.length]=n=e[p]),e[p]!==n)for(;p--&&!k.includes(e[p]););};return{addDisposeCallback:(e,g)=>{if("function"!=typeof g)throw Error("Callback must be a function");f(e,1).add(g)},ab:(e,g)=>{var k=f(e);k&&(k.delete(g),k.size||c.g.j.set(e,a,null))},ea:e=>{c.u.J(()=>{b[e.nodeType]&&(h(e),d[e.nodeType]&&l(e.getElementsByTagName("*")))});return e},removeNode:e=>{c.ea(e);e.parentNode&&e.parentNode.removeChild(e)}}})();c.ea=c.g.O.ea;c.removeNode=c.g.O.removeNode;c.S("utils.domNodeDisposal",
c.g.O);c.extenders={debounce:(a,b)=>a.Ea(d=>ha(d,b)),rateLimit:(a,b)=>a.Ea(d=>fa(d,b)),notify:(a,b)=>{a.ma="always"==b?null:O}};var ea={undefined:1,"boolean":1,number:1,string:1};class la{constructor(a,b,d){this.W=a;this.gb=b;this.za=d;this.Ka=!1;this.I=this.ga=null}dispose(){this.Ka||(this.I&&c.g.O.ab(this.ga,this.I),this.Ka=!0,this.za(),this.W=this.gb=this.za=this.ga=this.I=null)}s(a){this.ga=a;c.g.O.addDisposeCallback(a,this.I=this.dispose.bind(this))}}c.U=function(){Object.setPrototypeOf(this,
R);R.init(this)};var R={init:a=>{a.V=new Map;a.V.set("change",new Set);a.lb=1},subscribe:function(a,b,d){var f=this;d=d||"change";var h=new la(f,b?a.bind(b):a,()=>{f.V.get(d).delete(h);f.La?.(d)});f.Na?.(d);f.V.has(d)||f.V.set(d,new Set);f.V.get(d).add(h);return h},B(a,b){b=b||"change";"change"===b&&this.Ha();if(this.oa(b)){b="change"===b&&this.Jb||new Set(this.V.get(b));try{c.u.pb(),b.forEach(d=>{d.Ka||d.gb(a)})}finally{c.u.end()}}},ya(){return this.lb},Vb(a){return this.ya()!==a},Ha(){++this.lb},
Ea(a){var b=this,d=c.T(b),f,h,l,e,g;b.sa||(b.sa=b.B,b.B=(n,p)=>{p&&"change"!==p?"beforeChange"===p?b.ib(n):b.sa(n,p):b.jb(n)});var k=a(()=>{b.ha=!1;d&&e===b&&(e=b.hb?b.hb():b());var n=h||g&&b.Ca(l,e);g=h=f=!1;n&&b.sa(l=e)});b.jb=(n,p)=>{p&&b.ha||(g=!p);b.Jb=new Set(b.V.get("change"));b.ha=f=!0;e=n;k()};b.ib=n=>{f||(l=n,b.sa(n,"beforeChange"))};b.kb=()=>{g=!0};b.Lb=()=>{b.Ca(l,b.G(!0))&&(h=!0)}},oa(a){return(this.V.get(a)||[]).size},Ca(a,b){return!this.ma||!this.ma(a,b)},toString:()=>"[object Object]",
extend:function(a){var b=this;a&&c.g.M(a,(d,f)=>{d=c.extenders[d];"function"==typeof d&&(b=d(b,f)||b)});return b}};c.U.fn=Object.setPrototypeOf(R,Function.prototype);c.Zb=a=>"function"==typeof a?.subscribe&&"function"==typeof a.B;(()=>{let a=[],b,d=0;c.u={pb:f=>{a.push(b);b=f},end:()=>b=a.pop(),Eb:f=>{if(b){if(!c.Zb(f))throw Error("Only subscribable things can act as dependencies");b.Nb.call(b.Ob,f,f.Kb||(f.Kb=++d))}},J(f,h,l){try{return a.push(b),b=void 0,f.apply(h,l||[])}finally{b=a.pop()}},na:()=>
b?.o.na(),Da:()=>b?.Da,o:()=>b?.o}})();const B=Symbol("_latestValue");c.aa=a=>{function b(){if(0<arguments.length)return b.Ca(b[B],arguments[0])&&(b.fb(),b[B]=arguments[0],b.valueHasMutated()),this;c.u.Eb(b);return b[B]}b[B]=a;Object.defineProperty(b,"length",{get:()=>null==b[B]?void 0:b[B].length});c.U.fn.init(b);return Object.setPrototypeOf(b,K)};var K={toJSON:function(){let a=this[B];return a?.toJSON?.()||a},ma:O,G(){return this[B]},valueHasMutated:function(){this.B(this[B],"spectate");this.B(this[B])},
fb(){this.B(this[B],"beforeChange")}};Object.setPrototypeOf(K,c.U.fn);var E=c.aa.dc="__ko_proto__";K[E]=c.aa;c.T=a=>{if((a="function"==typeof a&&a[E])&&a!==K[E]&&a!==c.o.fn[E])throw Error("Invalid object that looks like an observable; possibly from another Knockout instance");return!!a};c.yb=a=>"function"==typeof a&&(a[E]===K[E]||a[E]===c.o.fn[E]&&a.Wb);c.S("observable",c.aa);c.S("isObservable",c.T);c.S("observable.fn",K);c.observableArray=a=>{a=a||[];if("object"!=typeof a||!("length"in a))throw Error("The argument passed when initializing an observable array must be an array, or null, or undefined.");
return Object.setPrototypeOf(c.aa(a),c.observableArray.fn).extend({trackArrayChanges:!0})};const ba=Symbol("IS_OBSERVABLE_ARRAY");c.observableArray.fn=Object.setPrototypeOf({[ba]:1,remove:function(a){for(var b=this.G(),d=!1,f="function"!=typeof a||c.T(a)?e=>e===a:a,h=b.length;h--;){var l=b[h];if(f(l)){if(b[h]!==l)throw Error("Array modified during remove; cannot remove item");d||this.fb();d=!0;b.splice(h,1)}}d&&this.valueHasMutated()}},c.aa.fn);Object.getOwnPropertyNames(Array.prototype).forEach(a=>
{"function"===typeof Array.prototype[a]&&"constructor"!=a&&("copyWithin fill pop push reverse shift sort splice unshift".split(" ").includes(a)?c.observableArray.fn[a]=function(...b){var d=this.G();this.fb();this.rb(d,a,b);b=d[a](...b);this.valueHasMutated();return b===d?this:b}:c.observableArray.fn[a]=function(...b){return this()[a](...b)})});c.isObservableArray=a=>!(!a||!a[ba]);c.extenders.trackArrayChanges=(a,b)=>{function d(){function m(){if(g){var q=[].concat(a.G()||[]);if(a.oa("arrayChange")){if(!h||
1<g)h=c.g.sb(k,q,a.Pa);var r=h}k=q;h=null;g=0;r?.length&&a.B(r,"arrayChange")}}f?m():(f=!0,e=a.subscribe(()=>++g,null,"spectate"),k=[].concat(a.G()||[]),h=null,l=a.subscribe(m))}a.Pa={};"object"==typeof b&&c.g.extend(a.Pa,b);a.Pa.sparse=!0;if(!a.rb){var f=!1,h=null,l,e,g=0,k,n=a.Na,p=a.La;a.Na=m=>{n?.call(a,m);"arrayChange"===m&&d()};a.La=m=>{p?.call(a,m);"arrayChange"!==m||a.oa("arrayChange")||(l?.dispose(),e?.dispose(),e=l=null,f=!1,k=void 0)};a.rb=(m,q,r)=>{function t(A,v,G){return u[u.length]=
{status:A,value:v,index:G}}if(f&&!g){var u=[],y=m.length,w=r.length,z=0;switch(q){case "push":z=y;case "unshift":for(m=0;m<w;m++)t("added",r[m],z+m);break;case "pop":z=y-1;case "shift":y&&t("deleted",m[z],z);break;case "splice":z=Math.min(Math.max(0,0>r[0]?y+r[0]:r[0]),y);y=1===w?y:Math.min(z+(r[1]||0),y);w=z+w-2;q=Math.max(y,w);for(var C=[],F=[],M=2;z<q;++z,++M)z<y&&F.push(t("deleted",m[z],z)),z<w&&C.push(t("added",r[M],z));c.g.wb(F,C);break;default:return}h=u}}}};var x=Symbol("_state");c.o=(a,b)=>
{function d(){if(0<arguments.length){if("function"!==typeof f)throw Error("Cannot write a value to a ko.computed unless you specify a 'write' option. If you wish to read the current value, don't pass any parameters.");f(...arguments);return this}h.Z||c.u.Eb(d);(h.X||h.A&&d.pa())&&d.R();return h.K}"object"===typeof a?b=a:(b=b||{},a&&(b.read=a));if("function"!=typeof b.read)throw Error("Pass a function that returns the value of the ko.computed");var f=b.write,h={K:void 0,$:!0,X:!0,Ba:!1,eb:!1,Z:!1,
$a:!1,A:!1,Db:b.read,s:b.disposeWhenNodeIsRemoved||b.s||null,ka:b.disposeWhen||b.ka,Ta:null,v:{},H:0,vb:null};d[x]=h;d.Wb="function"===typeof f;c.U.fn.init(d);Object.setPrototypeOf(d,N);b.pure&&(h.$a=!0,h.A=!0,c.g.extend(d,ma));h.s&&(h.eb=!0,h.s.nodeType||(h.s=null));h.A||d.R();h.s&&d.isActive()&&c.g.O.addDisposeCallback(h.s,h.Ta=()=>{d.dispose()});return d};var N={ma:O,na(){return this[x].H},Tb(){var a=[];c.g.M(this[x].v,(b,d)=>a[d.ia]=d.W);return a},Xa(a){if(!this[x].H)return!1;var b=this.Tb();
return b.includes(a)||!!b.find(d=>d.Xa&&d.Xa(a))},mb(a,b,d){if(this[x].$a&&b===this)throw Error("A 'pure' computed must not be called recursively");this[x].v[a]=d;d.ia=this[x].H++;d.ja=b.ya()},pa(){var a,b=this[x].v;for(a in b)if(Object.prototype.hasOwnProperty.call(b,a)){var d=b[a];if(this.ra&&d.W.ha||d.W.Vb(d.ja))return!0}},jc(){this[x].Ba||this.ra?.(!1)},isActive(){var a=this[x];return a.X||0<a.H},kc(){this.ha?this[x].X&&(this[x].$=!0):this.ub()},Hb(a){return a.subscribe(this.ub,this)},ub(){var a=
this,b=a.throttleEvaluation;0<=b?(clearTimeout(this[x].vb),this[x].vb=setTimeout(()=>a.R(!0),b)):a.ra?a.ra(!0):a.R(!0)},R(a){var b=this[x],d=b.ka,f=!1;if(!b.Ba&&!b.Z){if(b.s&&!c.g.Ua(b.s)||d?.()){if(!b.eb){this.dispose();return}}else b.eb=!1;try{b.Ba=!0,f=this.Rb(a)}finally{b.Ba=!1}return f}},Rb(a){var b=this[x],d=b.$a?void 0:!b.H;var f={Pb:this,va:b.v,Ra:b.H};c.u.pb({Ob:f,Nb:ja,o:this,Da:d});b.v={};b.H=0;a:{try{var h=b.Db();break a}finally{c.u.end(),f.Ra&&!b.A&&c.g.M(f.va,ia),b.$=b.X=!1}h=void 0}b.H?
f=this.Ca(b.K,h):(this.dispose(),f=!0);f&&(b.A?this.Ha():this.B(b.K,"beforeChange"),b.K=h,this.B(b.K,"spectate"),!b.A&&a&&this.B(b.K),this.kb&&this.kb());d&&this.B(b.K,"awake");return f},G(a){var b=this[x];(b.X&&(a||!b.H)||b.A&&this.pa())&&this.R();return b.K},Ea(a){var b=this;c.U.fn.Ea.call(b,a);b.hb=()=>{b[x].A||(b[x].$?b.R():b[x].X=!1);return b[x].K};b.ra=d=>{b.ib(b[x].K);b[x].X=!0;d&&(b[x].$=!0);b.jb(b,!d)}},dispose:function(){var a=this[x];!a.A&&a.v&&c.g.M(a.v,(b,d)=>d.dispose?.());a.s&&a.Ta&&
c.g.O.ab(a.s,a.Ta);a.v=void 0;a.H=0;a.Z=!0;a.$=!1;a.X=!1;a.A=!1;a.s=void 0;a.ka=void 0;a.Db=void 0}},ma={Na(a){var b=this,d=b[x];if(!d.Z&&d.A&&"change"==a){d.A=!1;if(d.$||b.pa())d.v=null,d.H=0,b.R()&&b.Ha();else{var f=[];c.g.M(d.v,(h,l)=>f[l.ia]=h);f.forEach((h,l)=>{var e=d.v[h],g=b.Hb(e.W);g.ia=l;g.ja=e.ja;d.v[h]=g});b.pa()&&b.R()&&b.Ha()}d.Z||b.B(d.K,"awake")}},La(a){var b=this[x];b.Z||"change"!=a||this.oa("change")||(c.g.M(b.v,(d,f)=>{f.dispose&&(b.v[d]={W:f.W,ia:f.ia,ja:f.ja},f.dispose())}),b.A=
!0,this.B(void 0,"asleep"))},ya(){var a=this[x];a.A&&(a.$||this.pa())&&this.R();return c.U.fn.ya.call(this)}};Object.setPrototypeOf(N,c.U.fn);var S=c.aa.dc;N[S]=c.o;c.o.fn=N;c.S("computed",c.o);c.isComputed=a=>"function"==typeof a&&a[S]===N[S];c.Cb=a=>{if("function"===typeof a)return c.o(a,{pure:!0});a={...a,pure:!0};return c.o(a)};c.C={N:a=>{switch(a.nodeName){case "OPTION":return!0===a.__ko__hasDomDataOptionValue__?c.g.j.get(a,c.h.options.Za):a.value;case "SELECT":return 0<=a.selectedIndex?c.C.N(a.options[a.selectedIndex]):
void 0;default:return a.value}},Ia:(a,b,d)=>{switch(a.nodeName){case "OPTION":"string"===typeof b?(c.g.j.set(a,c.h.options.Za,void 0),delete a.__ko__hasDomDataOptionValue__,a.value=b):(c.g.j.set(a,c.h.options.Za,b),a.__ko__hasDomDataOptionValue__=!0,a.value="number"===typeof b?b:"");break;case "SELECT":for(var f=-1,h=""===b||null==b,l=a.options.length,e;l--;)if(e=c.C.N(a.options[l]),e==b||""===e&&h){f=l;break}if(d||0<=f||h&&1<a.size)a.selectedIndex=f;break;default:a.value=null==b?"":b}}};c.P=(()=>
{var a=RegExp("\"(?:\\\\.|[^\"])*\"|'(?:\\\\.|[^'])*'|`(?:\\\\.|[^`])*`|/\\*(?:[^*]|\\*+[^*/])*\\*+/|//.*\n|/(?:\\\\.|[^/])+/w*|[^\\s:,/][^,\"'`{}()/:[\\]]*[^\\s,\"'`{}()/:[\\]]|[^\\s]","g"),b=/[\])"'A-Za-z0-9_$]+$/,d={"in":1,"return":1,"typeof":1},f=h=>{h=c.g.Gb(h);123===h.charCodeAt(0)&&(h=h.slice(1,-1));h+="\n,";var l=[],e=h.match(a),g=[],k=0;if(1<e.length){for(var n=0,p;p=e[n++];){var m=p.charCodeAt(0);if(44===m){if(0>=k){l.push(q&&g.length?{key:q,value:g.join("")}:{unknown:q||g.join("")});var q=
k=0;g=[];continue}}else if(58===m){if(!k&&!q&&1===g.length){q=g.pop();continue}}else if(47===m&&1<p.length&&(47===p.charCodeAt(1)||42===p.charCodeAt(1)))continue;else 47===m&&n&&1<p.length?(m=e[n-1].match(b))&&!d[m[0]]&&(h=h.slice(h.indexOf(p)+1),e=h.match(a),n=-1,p="/"):40===m||123===m||91===m?++k:41===m||125===m||93===m?--k:q||g.length||34!==m&&39!==m||(p=p.slice(1,-1));g.push(p)}if(0<k)throw Error("Unbalanced parentheses, braces, or brackets");}return l};return{Oa:[],bc:f,cc:h=>{var l=[],e=(g,
k)=>{var n=c.h[g];n?.preprocess&&!n.preprocess(k,g,e)||l.push("'"+g+"':function(){return "+(k+" }"))};f(h).forEach(g=>e(g.key||g.unknown,g.value));return l.join(",")},$b:(h,l)=>-1<h.findIndex(e=>e.key==l),Ja:(h,l,e,g,k)=>{if(h&&c.T(h))!c.yb(h)||k&&h.G()===g||h(g);else throw Error(`${e} , must be observable`);}}})();(()=>{function a(e){return 8==e.nodeType&&f.test(e.nodeValue)}function b(e){return 8==e.nodeType&&h.test(e.nodeValue)}function d(e,g){for(var k=e,n=1,p=[];k=k.nextSibling;){if(b(k)&&(c.g.j.set(k,
l,!0),!--n))return p;p.push(k);a(k)&&++n}if(!g)throw Error("Cannot find closing comment tag to match: "+e.nodeValue);return null}var f=/^\s*ko(?:\s+([\s\S]+))?\s*$/,h=/^\s*\/ko\s*$/,l="__ko_matchedEndComment__";c.m={ba:{},childNodes:e=>a(e)?d(e):e.childNodes,la:e=>{a(e)?(e=d(e))&&[...e].forEach(g=>c.removeNode(g)):c.g.Va(e)},qa:(e,g)=>{a(e)?(c.m.la(e),e.after(...g)):c.g.qa(e,g)},prepend:(e,g)=>{a(e)?e.nextSibling.before(g):e.prepend(g)},Yb:(e,g,k)=>{k?k.after(g):c.m.prepend(e,g)},firstChild:e=>{if(a(e))return e=
e.nextSibling,!e||b(e)?null:e;let g=e.firstChild;if(g&&b(g))throw Error("Found invalid end comment, as the first child of "+e);return g},nextSibling:e=>{if(a(e)){var g=d(e,void 0);e=g?(g.length?g[g.length-1]:e).nextSibling:null}if((g=e.nextSibling)&&b(g)){if(b(g)&&!c.g.j.get(g,l))throw Error("Found end comment without a matching opening comment, as child of "+e);return null}return g},Ub:a,ic:e=>(e=e.nodeValue.match(f))?e[1]:null}})();const ca=new Map;c.qb=new class{ac(a){switch(a.nodeType){case 1:return null!=
a.getAttribute("data-bind");case 8:return c.m.Ub(a)}return!1}Sb(a,b){a:{switch(a.nodeType){case 1:var d=a.getAttribute("data-bind");break a;case 8:d=c.m.ic(a);break a}d=null}if(d)try{let h=ca.get(d);if(!h){var f="with($context){with($data||{}){return{"+c.P.cc(d)+"}}}";h=new Function("$context","$element",f);ca.set(d,h)}return h(b,a)}catch(h){throw h.message="Unable to parse bindings.\nBindings value: "+d+"\nMessage: "+h.message,h;}return null}};const I=Symbol("_subscribable"),L=Symbol("_ancestorBindingInfo"),
aa=Symbol("_dataDependency"),da={},H=c.g.j.Y();c.h={};c.da=class{constructor(a,b,d,f,h){var l=this,e=a===da,g=e?void 0:a,k="function"==typeof g&&!c.T(g),n=h?.dataDependency;a=()=>{var m=k?g():g,q=c.g.i(m);b?(c.g.extend(l,b),L in b&&(l[L]=b[L])):(l.$parents=[],l.$root=q,l.ko=c);l[I]=p;e?q=l.$data:(l.$rawData=m,l.$data=q);d&&(l[d]=q);f?.(l,b,q);if(b?.[I]&&!c.u.o().Xa(b[I]))b[I]();n&&(l[aa]=n);return l.$data};if(h?.exportDependencies)a();else{var p=c.Cb(a);p.G();p.isActive()?p.ma=null:l[I]=void 0}}createChildContext(a,
b,d,f){!f&&b&&"object"==typeof b&&(f=b,b=f.as,d=f.extend);return new c.da(a,this,b,(h,l)=>{h.$parentContext=l;h.$parent=l.$data;h.$parents=(l.$parents||[]).slice(0);h.$parents.unshift(h.$parent);d&&d(h)},f)}extend(a,b){return new c.da(da,this,null,d=>c.g.extend(d,"function"==typeof a?a(d):a),b)}};class na{constructor(a,b,d){this.I=a;this.ga=b;this.ta=new Set;this.D=!1;b.F||c.g.O.addDisposeCallback(a,W);d?.F&&(d.F.ta.add(a),this.za=d)}Bb(){this.za?.F?.Qb(this.I)}Qb(a){this.ta.delete(a);this.ta.size||
this.tb?.()}tb(){this.D=!0;this.ga.F&&!this.ta.size&&(this.ga.F=null,c.g.O.ab(this.I,W),c.l.notify(this.I,c.l.fa),this.Bb())}}c.l={D:"childrenComplete",fa:"descendantsComplete",subscribe:(a,b,d,f,h)=>{var l=c.g.j.Wa(a,H,{});l.wa||(l.wa=new c.U);h?.notifyImmediately&&l.Ya[b]&&c.u.J(d,f,[a]);return l.wa.subscribe(d,f,b)},notify:(a,b)=>{var d=c.g.j.get(a,H);if(d&&(d.Ya[b]=!0,d.wa?.B(a,b),b==c.l.D))if(d.F)d.F.tb();else if(void 0===d.F&&d.wa?.oa(c.l.fa))throw Error("descendantsComplete event not supported for bindings on this node");
},cb:(a,b)=>{var d=c.g.j.Wa(a,H,{});d.F||(d.F=new na(a,d,b[L]));return b[L]==d?b:b.extend(f=>{f[L]=d})}};c.hc=a=>(a=c.g.j.get(a,H))&&a.context;c.applyBindingAccessorsToNode=(a,b,d)=>Z(a,b,P(d));c.ob=(a,b)=>{1!==b.nodeType&&8!==b.nodeType||X(P(a),b)};c.nb=function(a,b,d){if(2>arguments.length){if(b=J.body,!b)throw Error("ko.applyBindings: could not find document.body; has the document been loaded?");}else if(!b||1!==b.nodeType&&8!==b.nodeType)throw Error("ko.applyBindings: first parameter should be your view model; second parameter should be a DOM node");
Y(P(a,d),b)};c.dataFor=a=>(a=a&&[1,8].includes(a.nodeType)&&c.hc(a))?a.$data:void 0;c.S("bindingHandlers",c.h);c.S("applyBindings",c.nb);(()=>{var a=Object.create(null),b=new Map;c.components={get:(l,e)=>{if(b.has(l))e(b.get(l));else{var g=a[l];g?g.subscribe(e):(g=a[l]=new c.U,g.subscribe(e),h(l,k=>{b.set(l,k);delete a[l];g.B(k)}))}},register:(l,e)=>{if(!e)throw Error("Invalid configuration for "+l);if(d[l])throw Error("Component "+l+" is already registered");d[l]=e}};var d=Object.create(null),f=
(l,e)=>{throw Error(`Component '${l}': ${e}`);},h=(l,e)=>{var g={},k=d[l]||{},n=k.template;k=k.viewModel;if(n){n.element||f(l,"Unknown template value: "+n);n=n.element;var p=J.getElementById(n);p||f(l,"Cannot find element with ID "+n);p.matches("TEMPLATE")||f(l,"Template Source Element not a <template>");g.template=c.g.ua(p.content.childNodes)}k&&("function"!==typeof k.createViewModel&&f(l,"Unknown viewModel value: "+k),g.createViewModel=k.createViewModel);e(g.template&&g.createViewModel?g:null)}})();
(()=>{var a=0;c.h.component={init:(b,d,f,h,l)=>{var e,g,k,n=()=>{var m=e&&e.dispose;"function"===typeof m&&m.call(e);k&&k.dispose();g=e=k=null},p=[...c.m.childNodes(b)];c.m.la(b);c.g.O.addDisposeCallback(b,n);c.o(()=>{var m=c.g.i(d());if("string"!==typeof m){var q=c.g.i(m.params);m=c.g.i(m.name)}if(!m)throw Error("No component name specified");var r=c.l.cb(b,l),t=g=++a;c.components.get(m,u=>{if(g===t){n();if(!u)throw Error("Unknown component '"+m+"'");var y=u.template;if(!y)throw Error("Component '"+
m+"' has no template");c.m.qa(b,c.g.ua(y));e=u.createViewModel(q,{element:b,templateNodes:p});c.ob(r.createChildContext(e,{extend:w=>{w.$component=e;w.$componentTemplateNodes=p}}),b)}})},{s:b});return{controlsDescendantBindings:!0}}};c.m.ba.component=!0})();c.h.attr={update:(a,b)=>{b=c.g.i(b())||{};c.g.M(b,function(d,f){f=c.g.i(f);var h=d.indexOf(":");h="lookupNamespaceURI"in a&&0<h&&a.lookupNamespaceURI(d.slice(0,h));var l=!1===f||null==f;l?h?a.removeAttributeNS(h,d):a.removeAttribute(d):(f=f.toString(),
h?a.setAttributeNS(h,d,f):a.setAttribute(d,f));"name"===d&&(a.name=l?"":f)})}};(()=>{c.h.checked={after:["value","attr"],init:function(a,b,d){var f="checkbox"==a.type,h="radio"==a.type;if(f||h){const m=c.Cb(()=>{if(d.has("checkedValue"))return c.g.i(d.get("checkedValue"));if(n)return d.has("value")?c.g.i(d.get("value")):a.value});var l=()=>{if(!c.u.Da()){var q=a.checked,r=m();if(q||!h&&!c.u.na()){var t=c.u.J(b);if(g){var u=k?t.G():t,y=p;p=r;y!==r?q&&(u.push(r),u.remove(y)):q?u.push(r):u.remove(r);
k&&c.yb(t)&&t(u)}else f&&(void 0===r?r=q:q||(r=void 0)),c.P.Ja(t,d,"checked",r,!0)}}},e=b(),g=f&&c.g.i(e)instanceof Array,k=!(g&&e.push&&e.splice),n=h||g,p=g?m():void 0;h&&!a.name&&c.h.uniqueName.init(a,function(){return!0});c.o(l,null,{s:a});a.addEventListener("click",l);c.o(()=>{var q=c.g.i(b()),r=m();g?(a.checked=q.includes(r),p=r):a.checked=f&&void 0===r?!!q:m()===q},null,{s:a});e=void 0}}};c.h.checkedValue={update:function(a,b){a.value=c.g.i(b())}}})();var T=(a,b,d)=>b&&b.split(/\s+/).forEach(f=>
a.classList.toggle(f,d));c.h.css={update:(a,b)=>{b=c.g.i(b());"object"==typeof b?c.g.M(b,(d,f)=>{f=c.g.i(f);T(a,d,!!f)}):(b=c.g.Gb(b),T(a,a.__ko__cssValue,!1),a.__ko__cssValue=b,T(a,b,!0))}};c.h.enable={update:(a,b)=>{(b=c.g.i(b()))&&a.disabled?a.removeAttribute("disabled"):b||a.disabled||(a.disabled=!0)}};c.h.disable={update:(a,b)=>c.h.enable.update(a,()=>!c.g.i(b()))};c.h.event={init:function(a,b,d,f,h){c.g.M(b()||{},l=>{"string"==typeof l&&a.addEventListener(l,(...e)=>{var g=b()[l];if(g)try{f=
h.$data;var k=g.apply(f,[f,...e])}finally{!0!==k&&e[0].preventDefault()}})})}};c.h.foreach={zb:a=>()=>{var b=a(),d=c.T(b)?b.G():b;if(!d||"number"==typeof d.length)return{foreach:b};c.g.i(b);return{foreach:d.data,as:d.as,beforeRemove:d.beforeRemove}},init:(a,b)=>c.h.template.init(a,c.h.foreach.zb(b)),update:(a,b,d,f,h)=>c.h.template.update(a,c.h.foreach.zb(b),d,f,h)};c.P.Oa.foreach=!1;c.m.ba.foreach=!0;c.h.hasfocus={init:(a,b,d)=>{var f=l=>{a.__ko_hasfocusUpdating=!0;l=a.ownerDocument.activeElement===
a;c.P.Ja(b(),d,"hasfocus",l,!0);a.__ko_hasfocusLastValue=l;a.__ko_hasfocusUpdating=!1},h=f.bind(null,!0);f=f.bind(null,!1);a.addEventListener("focus",h);a.addEventListener("focusin",h);a.addEventListener("blur",f);a.addEventListener("focusout",f);a.__ko_hasfocusLastValue=!1},update:(a,b)=>{b=!!c.g.i(b());a.__ko_hasfocusUpdating||a.__ko_hasfocusLastValue===b||(b?a.focus():a.blur())}};c.h.html={init:()=>({controlsDescendantBindings:!0}),update:(a,b)=>{c.g.Va(a);b=c.g.i(b());if(null!=b){const d=J.createElement("template");
d.innerHTML="string"!=typeof b?b.toString():b;a.appendChild(d.content)}}};(()=>{function a(b,d,f){c.h[b]={init:(h,l,e,g,k)=>{var n,p={};d&&(p={as:e.get("as"),exportDependencies:!0});var m=e.has(c.l.fa);c.o(()=>{var q=c.g.i(l()),r=!f!==!q,t=!n;m&&(k=c.l.cb(h,k));if(r){p.dataDependency=c.u.o();var u=d?k.createChildContext("function"==typeof q?q:l,p):c.u.na()?k.extend(null,p):k}t&&c.u.na()&&(n=c.g.ua(c.m.childNodes(h),!0));r?(t||c.m.qa(h,c.g.ua(n)),c.ob(u,h)):(c.m.la(h),c.l.notify(h,c.l.D))},{s:h});
return{controlsDescendantBindings:!0}}};c.P.Oa[b]=!1;c.m.ba[b]=!0}a("if");a("ifnot",!1,!0);a("with",!0)})();var U={};c.h.options={init:a=>{if(!a.matches("SELECT"))throw Error("options binding applies only to SELECT elements");let b=a.length;for(;b--;)a.remove(b);return{controlsDescendantBindings:!0}},update:(a,b,d)=>{var f=a.multiple,h=0!=a.length&&f?a.scrollTop:null,l=c.g.i(b()),e=d.get("valueAllowUnset")&&d.has("value"),g={},k=[];b=()=>Array.from(a.options).filter(r=>r.selected);var n=(r,t,u)=>
{var y=typeof t;return"function"==y?t(r):"string"==y?r[t]:u},p=(r,t)=>{q&&e?c.l.notify(a,c.l.D):k.length&&(r=k.includes(c.C.N(t[0])),t[0].selected=r,q&&!r&&c.u.J(c.g.Ib,null,[a,"change"]))};e||(f?k=b().map(c.C.N):0<=a.selectedIndex&&k.push(c.C.N(a.options[a.selectedIndex])));if(l){"undefined"==typeof l.length&&(l=[l]);var m=l.filter(r=>r||null==r);d.has("optionsCaption")&&(l=c.g.i(d.get("optionsCaption")),null!=l&&m.unshift(U))}var q=!1;g.beforeRemove=r=>a.removeChild(r);l=p;d.has("optionsAfterRender")&&
"function"==typeof d.get("optionsAfterRender")&&(l=(r,t)=>{p(r,t);c.u.J(d.get("optionsAfterRender"),null,[t[0],r!==U?r:void 0])});c.g.Fb(a,m,(r,t,u)=>{u.length&&(k=!e&&u[0].selected?[c.C.N(u[0])]:[],q=!0);t=a.ownerDocument.createElement("option");r===U?(c.g.bb(t,d.get("optionsCaption")),c.C.Ia(t,void 0)):(u=n(r,d.get("optionsValue"),r),c.C.Ia(t,c.g.i(u)),r=n(r,d.get("optionsText"),u),c.g.bb(t,r));return[t]},g,l);e||(m=k.length,(f?m&&b().length<m:m&&0<=a.selectedIndex?c.C.N(a.options[a.selectedIndex])!==
k[0]:m||0<=a.selectedIndex)&&c.u.J(c.g.Ib,null,[a,"change"]));(e||c.u.Da())&&c.l.notify(a,c.l.D);h&&20<Math.abs(h-a.scrollTop)&&(a.scrollTop=h)}};c.h.options.Za=c.g.j.Y();c.h.style={update:(a,b)=>{c.g.M(c.g.i(b()||{}),(d,f)=>{f=c.g.i(f);if(null==f||!1===f)f="";if(/^--/.test(d))a.style.setProperty(d,f);else{d=d.replace(/-(\w)/g,(l,e)=>e.toUpperCase());var h=a.style[d];a.style[d]=f;f===h||a.style[d]!=h||isNaN(f)||(a.style[d]=f+"px")}})}};c.h.submit={init:(a,b,d,f,h)=>{if("function"!=typeof b())throw Error("The value for a submit binding must be a function");
a.addEventListener("submit",l=>{var e=b();try{var g=e.call(h.$data,a)}finally{!0!==g&&l.preventDefault()}})}};c.h.text={init:()=>({controlsDescendantBindings:!0}),update:(a,b)=>{8===a.nodeType&&(a.text||a.after(a.text=J.createTextNode("")),a=a.text);c.g.bb(a,b())}};c.m.ba.text=!0;c.h.textInput={init:(a,b,d)=>{var f=a.value,h,l,e=()=>{clearTimeout(h);l=h=void 0;var k=a.value;f!==k&&(f=k,c.P.Ja(b(),d,"textInput",k))},g=()=>{var k=c.g.i(b());null==k&&(k="");void 0!==l&&k===l?setTimeout(g,4):a.value!==
k&&(a.value=k,f=a.value)};a.addEventListener("input",e);a.addEventListener("change",e);a.addEventListener("blur",e);c.o(g,{s:a})}};c.h.textinput={preprocess:(a,b,d)=>d("textInput",a)};c.h.value={init:(a,b,d)=>{var f=a.matches("SELECT"),h=a.matches("INPUT");if(!h||"checkbox"!=a.type&&"radio"!=a.type){var l=new Set,e=d.get("valueUpdate"),g=null,k=()=>{g=null;var m=b(),q=c.C.N(a);c.P.Ja(m,d,"value",q)};e&&("string"==typeof e?l.add(e):e.forEach(m=>l.add(m)),l.delete("change"));l.forEach(m=>{var q=k;(m||
"").startsWith("after")&&(q=()=>{g=c.C.N(a);setTimeout(k,0)},m=m.slice(5));a.addEventListener(m,q)});var n=h&&"file"==a.type?()=>{var m=c.g.i(b());null==m||""===m?a.value="":c.u.J(k)}:()=>{var m=c.g.i(b()),q=c.C.N(a);if(null!==g&&m===g)setTimeout(n,0);else if(m!==q||void 0===q)f?(q=d.get("valueAllowUnset"),c.C.Ia(a,m,q),q||m===c.C.N(a)||c.u.J(k)):c.C.Ia(a,m)};if(f){var p;c.l.subscribe(a,c.l.D,()=>{p?d.get("valueAllowUnset")?n():k():(a.addEventListener("change",k),p=c.o(n,{s:a}))},null,{notifyImmediately:!0})}else a.addEventListener("change",
k),c.o(n,{s:a})}else c.applyBindingAccessorsToNode(a,{checkedValue:b})},update:()=>{}};c.h.visible={update:(a,b)=>{b=c.g.i(b());var d="none"!=a.style.display;b&&!d?a.style.display="":d&&!b&&(a.style.display="none")}};c.h.hidden={update:(a,b)=>a.hidden=!!c.g.i(b())};(function(a){c.h[a]={init:function(b,d,f,h,l){return c.h.event.init.call(this,b,()=>({[a]:d()}),f,h,l)}}})("click");(()=>{let a=c.g.j.Y();class b{constructor(f){this.Sa=f}Fa(...f){let h=this.Sa;if(!f.length)return c.g.j.get(h,a)||(11===
this.I?h.content:1===this.I?h:void 0);c.g.j.set(h,a,f[0])}}class d extends b{constructor(f){super(f);f&&(this.I=f.matches("TEMPLATE")&&f.content?f.content.nodeType:1)}}c.Ga={Sa:d,Ma:b}})();(()=>{var a=(e,g,k)=>{var n;for(g=c.m.nextSibling(g);e&&(n=e)!==g;)e=c.m.nextSibling(n),k(n,e)},b=(e,g)=>{if(e.length){var k=e[0],n=k.parentNode;a(k,e[e.length-1],p=>{1!==p.nodeType&&8!==p.nodeType||c.nb(g,p)});c.g.xa(e,n)}},d=(e,g,k,n)=>{var p=(e&&(e.nodeType?e:0<e.length?e[0]:null)||k||{}).ownerDocument;if("string"==
typeof k){p=p||J;p=p.getElementById(k);if(!p)throw Error("Cannot find template with ID "+k);k=new c.Ga.Sa(p)}else if([1,8].includes(k.nodeType))k=new c.Ga.Ma(k);else throw Error("Unknown template type: "+k);k=(k=k.Fa?k.Fa():null)?[...k.cloneNode(!0).childNodes]:null;if("number"!=typeof k.length||0<k.length&&"number"!=typeof k[0].nodeType)throw Error("Template engine must return an array of DOM nodes");p=!1;switch(g){case "replaceChildren":c.m.qa(e,k);p=!0;break;case "ignoreTargetNode":break;default:throw Error("Unknown renderMode: "+
g);}p&&(b(k,n),"replaceChildren"==g&&c.l.notify(e,c.l.D));return k},f=(e,g,k)=>c.T(e)?e():"function"===typeof e?e(g,k):e;c.ec=function(e,g,k,n){k=k||{};var p=p||"replaceChildren";if(n){var m=n.nodeType?n:0<n.length?n[0]:null;return c.o(()=>{var q=g instanceof c.da?g:new c.da(g,null,null,null,{exportDependencies:!0}),r=f(e,q.$data,q);d(n,p,r,q,k)},{ka:()=>!m||!c.g.Ua(m),s:m})}console.log("no targetNodeOrNodeArray")};c.fc=(e,g,k,n,p)=>{function m(w,z){c.u.J(c.g.Fb,null,[n,w,t,k,u,z]);c.l.notify(n,c.l.D)}
var q,r=k.as,t=(w,z)=>{q=p.createChildContext(w,{as:r,extend:C=>{C.$index=z;r&&(C[r+"Index"]=z)}});w=f(e,w,q);return d(n,"ignoreTargetNode",w,q,k)},u=(w,z)=>{b(z,q);q=null};if(!k.beforeRemove&&c.isObservableArray(g)){m(g.G());var y=g.subscribe(w=>{m(g(),w)},null,"arrayChange");y.s(n);return y}return c.o(()=>{var w=c.g.i(g)||[];"undefined"==typeof w.length&&(w=[w]);m(w)},{s:n})};var h=c.g.j.Y(),l=c.g.j.Y();c.h.template={init:(e,g)=>{g=c.g.i(g());if("string"==typeof g||"name"in g)c.m.la(e);else if("nodes"in
g){g=g.nodes||[];if(c.T(g))throw Error('The "nodes" option must be a plain, non-observable array.');let k=g[0]?.parentNode;k&&c.g.j.get(k,l)||(k=c.g.Ab(g),c.g.j.set(k,l,!0));(new c.Ga.Ma(e)).Fa(k)}else if(g=c.m.childNodes(e),g.length)g=c.g.Ab(g),(new c.Ga.Ma(e)).Fa(g);else throw Error("Anonymous template defined, but no template content was provided");return{controlsDescendantBindings:!0}},update:(e,g,k,n,p)=>{var m=g();g=c.g.i(m);k=!0;n=null;"string"==typeof g?g={}:(m="name"in g?g.name:e,"if"in g&&
(k=c.g.i(g["if"])),k&&"ifnot"in g&&(k=!c.g.i(g.ifnot)),k&&!m&&(k=!1));"foreach"in g?n=c.fc(m,k&&g.foreach||[],g,e,p):k?(k=p,"data"in g&&(k=p.createChildContext(g.data,{as:g.as,exportDependencies:!0})),n=c.ec(m,k,g,e)):c.m.la(e);p=n;c.g.j.get(e,h)?.dispose?.();c.g.j.set(e,h,!p||p.isActive&&!p.isActive()?void 0:p)}};c.P.Oa.template=e=>{e=c.P.bc(e);return 1==e.length&&e[0].unknown||c.P.$b(e,"name")?null:"This template engine does not support anonymous templates nested within its templates"};c.m.ba.template=
!0})();c.g.wb=(a,b,d)=>{var f=0,h,l=b.length;l&&a.every(e=>{h=b.findIndex(g=>e.value===g.value);0<=h&&(e.moved=b[h].index,b[h].moved=e.index,b.splice(h,1),f=h=0,--l);f+=l;return l&&(!d||f<d)})};c.g.sb=(()=>{var a=(b,d,f,h,l)=>{for(var e=Math.min,g=Math.max,k=[],n=-1,p=b.length,m,q=d.length,r=q-p||1,t=p+q+1,u,y,w;++n<=p;)for(y=u,k.push(u=[]),w=e(q,n+r),m=g(0,n-1);m<=w;m++)u[m]=m?n?b[n-1]===d[m-1]?y[m-1]:e(y[m]||t,u[m-1]||t)+1:m+1:n+1;e=[];g=[];r=[];n=p;for(m=q;n||m;)q=k[n][m]-1,m&&q===k[n][m-1]?g.push(e[e.length]=
{status:f,value:d[--m],index:m}):n&&q===k[n-1][m]?r.push(e[e.length]={status:h,value:b[--n],index:n}):(--m,--n,l.sparse||e.push({status:"retained",value:d[m]}));c.g.wb(r,g,!l.dontLimitMoves&&10*p);return e.reverse()};return(b,d,f)=>{f="boolean"===typeof f?{dontLimitMoves:f}:f||{};b=b||[];d=d||[];return b.length<d.length?a(b,d,"added","deleted",f):a(d,b,"deleted","added",f)}})();(()=>{function a(f,h,l,e,g){var k=[],n=c.o(()=>{var p=h(l,g,c.g.xa(k,f))||[];if(0<k.length){var m=k.nodeType?[k]:k;if(0<
m.length){var q=m[0],r=q.parentNode;p.forEach(t=>r.insertBefore(t,q));m.forEach(t=>c.removeNode(t))}e&&c.u.J(e,null,[l,p,g])}k.length=0;k.push(...p)},{s:f,ka:()=>!!k.find(c.g.Ua)});return{L:k,Qa:n.isActive()?n:void 0}}var b=c.g.j.Y(),d=c.g.j.Y();c.g.Fb=(f,h,l,e,g,k)=>{h=h||[];"undefined"==typeof h.length&&(h=[h]);e=e||{};var n=c.g.j.get(f,b),p=[],m=0,q=0,r=[],t=[],u=[],y=0,w=v=>{A={ca:v,Aa:c.aa(q++)};p.push(A)},z=v=>{A=n[v];A.Aa(q++);c.g.xa(A.L,f);p.push(A)};if(n){if(!k||n&&n._countWaitingForRemove)k=
c.g.sb(Array.prototype.map.call(n,D=>D.ca),h,{dontLimitMoves:e.dontLimitMoves,sparse:!0});let v,G;for(k.forEach(D=>{v=D.moved;G=D.index;switch(D.status){case "deleted":for(;m<G;)z(m++);void 0===v&&(A=n[m],A.Qa&&(A.Qa.dispose(),A.Qa=void 0),c.g.xa(A.L,f).length&&(e.beforeRemove&&(p.push(A),y++,A.ca===d?A=null:u[A.Aa.G()]=A),A&&r.push.apply(r,A.L)));m++;break;case "added":for(;q<G;)z(m++);void 0!==v?(t.push(p.length),z(v)):w(D.value)}});q<h.length;)z(m++);p._countWaitingForRemove=y}else h.forEach(w);
c.g.j.set(f,b,p);r.forEach(e.beforeRemove?c.ea:c.removeNode);var C,F,M=v=>{c.m.Yb(f,v,F);F=v};k=f.ownerDocument.activeElement;if(t.length)for(;void 0!=(C=t.shift());){var A=p[C];for(F=void 0;C;)if(h=p[--C].L,h?.length){F=h[h.length-1];break}A.L.forEach(M)}p.forEach(v=>{v.L||c.g.extend(v,a(f,l,v.ca,g,v.Aa));v.L.forEach(M);!v.Xb&&g&&(g(v.ca,v.L,v.Aa),v.Xb=!0,F=v.L[v.L.length-1])});f.ownerDocument.activeElement!=k&&k?.focus();((v,G)=>{v&&G.forEach(D=>D?.L.forEach(oa=>v(oa,C,D.ca)))})(e.beforeRemove,
u);u.forEach(v=>v&&(v.ca=d))}})();V.ko=Q})(this);

/* Copyright © 2011-2015 by Neil Jenkins. MIT Licensed. */
/* eslint max-len: 0 */

/**
	TODO: modifyBlocks function doesn't work very good.
	For example you have: UL > LI > [cursor here in text]
	Then create blockquote at cursor, the result is: BLOCKQUOTE > UL > LI
	not UL > LI > BLOCKQUOTE
*/

(doc => {

const
	SHOW_ELEMENT = 1,                // NodeFilter.SHOW_ELEMENT,
	SHOW_TEXT = 4,                   // NodeFilter.SHOW_TEXT,
	SHOW_ELEMENT_OR_TEXT = 5,

	// source/Constants.ts
	ELEMENT_NODE = 1,                // Node.ELEMENT_NODE,
	TEXT_NODE = 3,                   // Node.TEXT_NODE,
	DOCUMENT_FRAGMENT_NODE = 11,     // Node.DOCUMENT_FRAGMENT_NODE,
	ZWS = '\u200B',
	ua = navigator.userAgent,
	isMac = /Mac OS X/.test(ua),
	isIOS = /iP(?:ad|hone)/.test(ua) || (isMac && !!navigator.maxTouchPoints),
	isAndroid = /Android/.test(ua),
	isWebKit = /WebKit\//.test(ua),
	ctrlKey = isMac || isIOS ? 'meta-' : 'ctrl-',
	cantFocusEmptyTextNodes = isWebKit,
	// Use [^ \t\r\n] instead of \S so that nbsp does not count as white-space
	notWS = /[^ \t\r\n]/,

	// source/node/Category.ts
//	phrasingElements = 'ABBR,AUDIO,B,BDO,BR,BUTTON,CANVAS,CITE,CODE,COMMAND,DATA,DATALIST,DFN,EM,EMBED,I,IFRAME,IMG,INPUT,KBD,KEYGEN,LABEL,MARK,MATH,METER,NOSCRIPT,OBJECT,OUTPUT,PROGRESS,Q,RUBY,SAMP,SCRIPT,SELECT,SMALL,SPAN,STRONG,SUB,SUP,SVG,TEXTAREA,TIME,VAR,VIDEO,WBR',
	inlineNodeNames = /^(?:#text|A|ABBR|ACRONYM|B|BR|BD[IO]|CITE|CODE|DATA|DEL|DFN|EM|FONT|HR|I|IMG|INPUT|INS|KBD|Q|RP|RT|RUBY|S|SAMP|SMALL|SPAN|STR(IKE|ONG)|SU[BP]|TIME|U|VAR|WBR)$/,
	leafNodeNames = new Set(["BR", "HR", "IMG"]),
	listNodeNames = new Set(["OL", "UL"]),
	UNKNOWN = 0,
	INLINE = 1,
	BLOCK = 2,
	CONTAINER = 3,
	isLeaf = node => isElement(node) && leafNodeNames.has(node.nodeName),
	getNodeCategory = node => {
		switch (node.nodeType) {
		case TEXT_NODE:
			return INLINE;
		case ELEMENT_NODE:
		case DOCUMENT_FRAGMENT_NODE:
			if (nodeCategoryCache.has(node)) {
				return nodeCategoryCache.get(node);
			}
			break;
		default:
			return UNKNOWN;
		}

		let nodeCategory =
			Array.prototype.every.call(node.childNodes, isInline)
			? (inlineNodeNames.test(node.nodeName) ? INLINE : BLOCK)
			// Malformed HTML can have block tags inside inline tags. Need to treat
			// these as containers rather than inline. See #239.
			: CONTAINER;
		nodeCategoryCache.set(node, nodeCategory);
		return nodeCategory;
	},
	isInline = node => getNodeCategory(node) === INLINE,
	isBlock = node => getNodeCategory(node) === BLOCK,
	isContainer = node => getNodeCategory(node) === CONTAINER,

	// source/node/Node.ts
	createElement = (tag, props, children) => {
		const el = doc.createElement(tag);
		if (props instanceof Array) {
			children = props;
			props = null;
		}
		setAttributes(el, props);
		children && el.append(...children);
		return el;
	},
	areAlike = (node, node2) => {
		return !isLeaf(node) && (
			node.nodeType === node2.nodeType &&
			node.nodeName === node2.nodeName &&
			node.nodeName !== "A" &&
			node.className === node2.className &&
			node.style?.cssText === node2.style?.cssText
		);
	},
	hasTagAttributes = (node, tag, attributes) => {
		return node.nodeName === tag && Object.entries(attributes || {}).every(([k,v]) => node.getAttribute(k) === v);
	},
	getNearest = (node, root, tag, attributes) => {
		while (node && node !== root) {
			if (hasTagAttributes(node, tag, attributes)) {
				return node;
			}
			node = node.parentNode;
		}
		return null;
	},
	getNodeBefore = (node, offset) => {
		let children = node.childNodes;
		while (offset && isElement(node)) {
			node = children[offset - 1];
			children = node.childNodes;
			offset = children.length;
		}
		return node;
	},
	getNodeAfterOffset = (node, offset) => {
		if (isElement(node)) {
			const children = node.childNodes;
			if (offset < children.length) {
				node = children[offset];
			} else {
				while (node && !node.nextSibling) {
					node = node.parentNode;
				}
				node && (node = node.nextSibling);
			}
		}
		return node;
	},
	getLength = node =>
		isElement(node) || node instanceof DocumentFragment ? node.childNodes.length : node.length || 0,
	empty = node => {
		const frag = doc.createDocumentFragment(),
			childNodes = node.childNodes;
		childNodes && frag.append(...childNodes);
		return frag;
	},
	detach = node => {
//		node.remove();
		node.parentNode?.removeChild(node);
	},
	getClosest = (node, root, selector) => {
		node = (node && !node.closest) ? node.parentElement : node;
		node = node?.closest(selector);
		return (node && root.contains(node)) ? node : null;
	},

	// source/node/Whitespace.ts
	notWSTextNode = node => isElement(node) ? isBrElement(node) : notWS.test(node.data),
	isLineBreak = (br, isLBIfEmptyBlock) => {
		let walker, block = br.parentNode;
		while (isInline(block)) {
			block = block.parentNode;
		}
		walker = createTreeWalker(block, SHOW_ELEMENT_OR_TEXT, notWSTextNode);
		walker.currentNode = br;
		return !!walker.nextNode() || (isLBIfEmptyBlock && !walker.previousNode());
	},
	// --- Workaround for browsers that can't focus empty text nodes ---
	// WebKit bug: https://bugs.webkit.org/show_bug.cgi?id=15256
	// Walk down the tree starting at the root and remove any ZWS. If the node only
	// contained ZWS space then remove it too. We may want to keep one ZWS node at
	// the bottom of the tree so the block can be selected. Define that node as the
	// keepNode.
	removeZWS = (root, keepNode) => {
		const walker = createTreeWalker(root, SHOW_TEXT);
		let parent, node, index;
		while (node = walker.nextNode()) {
			while ((index = node.data.indexOf(ZWS)) > -1  && (!keepNode || node.parentNode !== keepNode)) {
				if (node.length === 1) {
					do {
						parent = node.parentNode;
						node.remove();
						node = parent;
						walker.currentNode = parent;
					} while (isInline(node) && !getLength(node));
					break;
				} else {
					node.deleteData(index, 1);
				}
			}
		}
	},

	// source/range/Boundaries.ts
	START_TO_START = 0, // Range.START_TO_START
	START_TO_END = 1,   // Range.START_TO_END
	END_TO_END = 2,     // Range.END_TO_END
	END_TO_START = 3,   // Range.END_TO_START
	isNodeContainedInRange = (range, node, partial = true) => {
		const nodeRange = doc.createRange();

		nodeRange.selectNode(node);

		return partial
			// Node must not finish before range starts or start after range finishes.
			? range.compareBoundaryPoints(END_TO_START, nodeRange) < 0
				&& range.compareBoundaryPoints(START_TO_END, nodeRange) > 0
			// Node must start after range starts and finish before range finishes
			: range.compareBoundaryPoints(START_TO_START, nodeRange) < 1
				&& range.compareBoundaryPoints(END_TO_END, nodeRange) > -1;
	},
	moveRangeBoundariesDownTree = range => {
		let { startContainer, startOffset, endContainer, endOffset } = range;
		let child;

		while (!isTextNode(startContainer)) {
			child = startContainer.childNodes[startOffset];
			if (!child || isLeaf(child)) {
				if (startOffset) {
					child = startContainer.childNodes[startOffset - 1];
					if (isTextNode(child)) {
						let textChild = child;
						let prev;
						while (!textChild.length && (prev = textChild.previousSibling) && isTextNode(prev)) {
							textChild.remove();
							textChild = prev;
						}
						startContainer = textChild;
						startOffset = textChild.data.length;
					}
				}
				break;
			}
			startContainer = child;
			startOffset = 0;
		}
		if (endOffset) {
			while (!isTextNode(endContainer)) {
				child = endContainer.childNodes[endOffset - 1];
				if (!child || isLeaf(child)) {
					if (isBrElement(child) && !isLineBreak(child)) {
						--endOffset;
						continue;
					}
					break;
				}
				endContainer = child;
				endOffset = getLength(endContainer);
			}
		} else {
			while (!isTextNode(endContainer)) {
				child = endContainer.firstChild;
				if (!child || isLeaf(child)) {
					break;
				}
				endContainer = child;
			}
		}

		range.setStart(startContainer, startOffset);
		range.setEnd(endContainer, endOffset);
	},
	moveRangeBoundariesUpTree = (range, startMax, endMax, root) => {
		let startContainer = range.startContainer;
		let startOffset = range.startOffset;
		let endContainer = range.endContainer;
		let endOffset = range.endOffset;
		let parent;

		if (!startMax) {
			startMax = range.commonAncestorContainer;
		}
		if (!endMax) {
			endMax = startMax;
		}

		while (!startOffset && startContainer !== startMax && startContainer !== root) {
			parent = startContainer.parentNode;
			startOffset = indexOf(parent.childNodes, startContainer);
			startContainer = parent;
		}

		while (endContainer !== endMax && endContainer !== root) {
			if (!isTextNode(endContainer) && isBrElement(endContainer.childNodes[endOffset]) && !isLineBreak(endContainer.childNodes[endOffset])) {
				++endOffset;
			}
			if (endOffset !== getLength(endContainer)) {
				break;
			}
			parent = endContainer.parentNode;
			endOffset = indexOf(parent.childNodes, endContainer) + 1;
			endContainer = parent;
		}

		range.setStart(startContainer, startOffset);
		range.setEnd(endContainer, endOffset);
	},
	moveRangeBoundaryOutOf = (range, tag, root) => {
		let parent = getClosest(range.endContainer, root, tag);
		if (parent && (parent = parent.parentNode)) {
			const clone = range.cloneRange();
			moveRangeBoundariesUpTree(clone, parent, parent, root);
			if (clone.endContainer === parent) {
				range.setStart(clone.endContainer, clone.endOffset);
				range.setEnd(clone.endContainer, clone.endOffset);
			}
		}
		return range;
	},

	// source/node/MergeSplit.ts
	fixCursor = (node) => {
		let fixer = null;
		if (!isTextNode(node)) {
			if (isInline(node)) {
				let child = node.firstChild;
				if (cantFocusEmptyTextNodes) {
					while (child && isTextNode(child) && !child.data) {
						node.removeChild(child);
						child = node.firstChild;
					}
				}
				if (!child) {
					fixer = doc.createTextNode(cantFocusEmptyTextNodes ? ZWS : "");
				}
			} else if (isElement(node) && !node.querySelector("BR")) {
				fixer = createElement("BR");
			}
			if (fixer) {
				try {
					node.appendChild(fixer);
				} catch (error) {
					didError({
						name: 'Squire: fixCursor – ' + error,
						message: 'Parent: ' + node.nodeName + '/' + node.innerHTML +
							' appendChild: ' + fixer.nodeName
					});
				}
			}
		}
		return node;
	},
	// Recursively examine container nodes and wrap any inline children.
	fixContainer = (container, root) => {
		let wrapper, isBR;
		// Not live, and fast
		[...container.childNodes].forEach(child => {
			isBR = isBrElement(child);
			if (!isBR && child.parentNode == root && isInline(child)
//			 && (blockTag !== "DIV" || (child.matches && !child.matches(phrasingElements)))
			) {
				wrapper = wrapper || createElement("DIV");
				wrapper.append(child);
			} else if (isBR || wrapper) {
				wrapper = wrapper || createElement("DIV");
				fixCursor(wrapper);
				child[isBR ? "replaceWith" : "before"](wrapper);
				wrapper = null;
			}
			isContainer(child) && fixContainer(child, root);
		});
		wrapper && container.append(fixCursor(wrapper));
		return container;
	},
	split = (node, offset, stopNode, root) => {
		if (isTextNode(node) && node !== stopNode) {
			if (typeof offset !== "number") {
				throw new Error("Offset must be a number to split text node!");
			}
			if (!node.parentNode) {
				throw new Error("Cannot split text node with no parent!");
			}
			return split(node.parentNode, node.splitText(offset), stopNode, root);
		}
		let nodeAfterSplit = typeof offset === "number" ? offset < node.childNodes.length ? node.childNodes[offset] : null : offset;
		const parent = node.parentNode;
		if (!parent || node === stopNode || !isElement(node)) {
			return nodeAfterSplit;
		}
		const clone = node.cloneNode(false);
		while (nodeAfterSplit) {
			const next = nodeAfterSplit.nextSibling;
			clone.append(nodeAfterSplit);
			nodeAfterSplit = next;
		}
		if (node instanceof HTMLOListElement && getClosest(node, root, "BLOCKQUOTE")) {
			clone.start = (+node.start || 1) + node.childNodes.length - 1;
		}
		fixCursor(node);
		fixCursor(clone);
		node.after(clone);
		return split(parent, clone, stopNode, root);
	},
	_mergeInlines = (node, fakeRange) => {
		const children = node.childNodes;
		let l = children.length;
		let frags = [];
		let child, prev;
		while (l--) {
			child = children[l];
			prev = l && children[l - 1];
			if (prev && isInline(child) && areAlike(child, prev)/* && !leafNodeNames.has(child.nodeName)*/) {
				if (fakeRange.startContainer === child) {
					fakeRange.startContainer = prev;
					fakeRange.startOffset += getLength(prev);
				}
				if (fakeRange.endContainer === child) {
					fakeRange.endContainer = prev;
					fakeRange.endOffset += getLength(prev);
				}
				if (fakeRange.startContainer === node) {
					if (fakeRange.startOffset > l) {
						--fakeRange.startOffset;
					} else if (fakeRange.startOffset === l) {
						fakeRange.startContainer = prev;
						fakeRange.startOffset = getLength(prev);
					}
				}
				if (fakeRange.endContainer === node) {
					if (fakeRange.endOffset > l) {
						--fakeRange.endOffset;
					} else if (fakeRange.endOffset === l) {
						fakeRange.endContainer = prev;
						fakeRange.endOffset = getLength(prev);
					}
				}
				detach(child);
				if (isTextNode(child)) {
					prev.appendData(child.data);
				} else {
					frags.unshift(empty(child));
				}
			} else if (isElement(child)) {
				child.append(...frags);
				frags = [];
				_mergeInlines(child, fakeRange);
			}
		}
	},
	mergeInlines = (node, range) => {
		node = isTextNode(node) ? node.parentNode : node;
		if (isElement(node)) {
			const fakeRange = {
				startContainer: range.startContainer,
				startOffset: range.startOffset,
				endContainer: range.endContainer,
				endOffset: range.endOffset
			};
			_mergeInlines(node, fakeRange);
			range.setStart(fakeRange.startContainer, fakeRange.startOffset);
			range.setEnd(fakeRange.endContainer, fakeRange.endOffset);
		}
	},
	mergeWithBlock = (block, next, range, root) => {
		let container = next;
		let parent;
		let offset;
		while ((parent = container.parentNode) && parent !== root && isElement(parent) && parent.childNodes.length === 1) {
			container = parent;
		}
		detach(container);
		offset = block.childNodes.length;
		// Remove extra <BR> fixer if present.
		const last = block.lastChild;
		if (isBrElement(last)) {
			last.remove();
			--offset;
		}
		block.append(empty(next));
		range.setStart(block, offset);
		range.collapse(true);
		mergeInlines(block, range);
	},
	mergeContainers = (node, root) => {
		const prev = node.previousSibling;
		const first = node.firstChild;
		const isListItem = node.nodeName === "LI";
		// Do not merge LIs, unless it only contains a UL
		if (isListItem && (!first || !listNodeNames.has(first.nodeName))) {
			return;
		}
		let needsFix, block;
		if (prev && areAlike(prev, node)) {
			if (!isContainer(prev)) {
				if (!isListItem) {
					return;
				}
				block = createElement("DIV");
				block.append(empty(prev));
				prev.append(block);
			}
			detach(node);
			needsFix = !isContainer(node);
			prev.append(empty(node));
			if (needsFix) {
				fixContainer(prev, root);
			}
			if (first) {
				mergeContainers(first, root);
			}
		} else if (isListItem) {
			block = createElement("DIV");
			node.insertBefore(block, first);
			fixCursor(block);
		}
	},

	// source/Clean.ts
	styleToSemantic = {
		fontWeight: {
			regexp: /^bold|^700/i,
			replace: () => createElement("B")
		},
		fontStyle: {
			regexp: /^italic/i,
			replace: () => createElement("I")
		},
		fontFamily: {
			regexp: notWS,
			replace: (doc, family) => createElement("SPAN", {
				style: 'font-family:' + family
			})
		},
		fontSize: {
			regexp: notWS,
			replace: (doc, size) => createElement("SPAN", {
				style: 'font-size:' + size
			})
		},
		textDecoration: {
			regexp: /^underline/i,
			replace: () => createElement("U")
		}
	/*
		textDecoration: {
			regexp: /^line-through/i,
			replace: doc => createElement("S")
		}
	*/
	},
	replaceStyles = node => {
		const style = node.style;
		let newTreeBottom, newTreeTop, css, el;
		Object.entries(styleToSemantic).forEach(([attr,converter])=>{
			css = style[attr];
			if (css && converter.regexp.test(css)) {
				el = converter.replace(doc, css);
				if (el.nodeName === node.nodeName && el.className === node.className) {
					return;
				}
				if (!newTreeTop) {
					newTreeTop = el;
				}
				if (newTreeBottom) {
					newTreeBottom.append(el);
				}
				newTreeBottom = el;
				node.style[attr] = '';
			}
		});
		if (newTreeTop && newTreeBottom) {
			newTreeBottom.append(empty(node));
			if (node.style.cssText) {
				node.append(newTreeTop);
			} else {
				node.replaceWith(newTreeTop);
			}
		}
		return newTreeBottom || node;
	},
	replaceWithTag = tag => node => {
		let el = createElement(tag);
		Array.prototype.forEach.call(node.attributes, attr => el.setAttribute(attr.name, attr.value));
		node.replaceWith(el);
		el.append(empty(node));
		return el;
	},
	fontSizes = {
		1: 'x-small',
		2: "small",
		3: "medium",
		4: "large",
		5: 'x-large',
		6: 'xx-large',
		7: 'xxx-large',
		'-1': "smaller",
		'+1': "larger"
	},
	stylesRewriters = {
		STRONG: replaceWithTag("B"),
		EM: replaceWithTag("I"),
		INS: replaceWithTag("U"),
		STRIKE: replaceWithTag("S"),
		SPAN: replaceStyles,
		FONT: node => {
			const face = node.face;
			const size = node.size;
			let color = node.color;
			let newTag = createElement("SPAN");
			let css = newTag.style;
			newTag.style.cssText = node.style.cssText;
			if (face) {
				css.fontFamily = face;
			}
			if (size) {
				css.fontSize = fontSizes[size];
			}
			if (color && /^#?([\dA-F]{3}){1,2}$/i.test(color)) {
				if (color.charAt(0) !== "#") {
					color = "#" + color;
				}
				css.color = color;
			}
			node.replaceWith(newTag);
			newTag.append(empty(node));
			return newTag;
		},
	//	KBD:
	//	VAR:
	//	CODE:
	//	SAMP:
		TT: node => {
			let el = createElement("SPAN", {
				style: 'font-family:menlo,consolas,"courier new",monospace'
			});
			node.replaceWith(el);
			el.append(empty(node));
			return el;
		}
	},
	allowedBlock = /^(?:A(?:DDRESS|RTICLE|SIDE|UDIO)|BLOCKQUOTE|CAPTION|D(?:[DLT]|IV)|F(?:IGURE|IGCAPTION|OOTER)|H[1-6]|HEADER|L(?:ABEL|EGEND|I)|O(?:L|UTPUT)|P(?:RE)?|SECTION|T(?:ABLE|BODY|D|FOOT|H|HEAD|R)|COL(?:GROUP)?|UL)$/,
	blacklist = new Set(["HEAD", "META", "STYLE"]),
	/*
		Two purposes:

		1. Remove nodes we don't want, such as weird <o:p> tags, comment nodes
		   and whitespace nodes.
		2. Convert inline tags into our preferred format.
	*/
	cleanTree = (node, preserveWS) => {
		const children = node.childNodes;
		let nonInlineParent, i = children.length, child, nodeName, childLength;
//			startsWithWS, endsWithWS, data, sibling;

		nonInlineParent = node;
		while (isInline(nonInlineParent)) {
			nonInlineParent = nonInlineParent.parentNode;
		}
//		const walker = createTreeWalker(nonInlineParent, SHOW_ELEMENT_OR_TEXT);

		while (i--) {
			child = children[i];
			nodeName = child.nodeName;
			if (isElement(child)) {
				childLength = child.childNodes.length;
				if (stylesRewriters[nodeName]) {
					child = stylesRewriters[nodeName](child);
				} else if (blacklist.has(nodeName)) {
					child.remove();
					continue;
				} else if (!allowedBlock.test(nodeName) && !isInline(child)) {
					i += childLength;
					child.replaceWith(empty(child));
					continue;
				}
				if (childLength) {
					cleanTree(child, preserveWS || (nodeName === "PRE"));
				}
/*
			} else {
				if (isTextNode(child)) {
					data = child.data;
					startsWithWS = !notWS.test(data.charAt(0));
					endsWithWS = !notWS.test(data.charAt(data.length - 1));
					if (preserveWS || (!startsWithWS && !endsWithWS)) {
						continue;
					}
					// Iterate through the nodes; if we hit some other content
					// before the start of a new block we don't trim
					if (startsWithWS) {
						walker.currentNode = child;
						while (sibling = walker.previousPONode()) {
							nodeName = sibling.nodeName;
							if (nodeName === "IMG" || (nodeName === '#text' && notWS.test(sibling.data))) {
								break;
							}
							if (!isInline(sibling)) {
								sibling = null;
								break;
							}
						}
						data = data.replace(/^[ \r\n]+/g, sibling ? ' ' : '');
					}
					if (endsWithWS) {
						walker.currentNode = child;
						while (sibling = walker.nextNode()) {
							if (nodeName === "IMG" || (nodeName === '#text' && notWS.test(sibling.data))) {
								break;
							}
							if (!isInline(sibling)) {
								sibling = null;
								break;
							}
						}
						data = data.replace(/[ \r\n]+$/g, sibling ? ' ' : '');
					}
					if (data) {
						child.data = data;
						continue;
					}
				}
				child.remove();
*/
			}
		}
		return node;
	},
	removeEmptyInlines = node => {
		const children = node.childNodes;
		let l = children.length,
			child;
		while (l--) {
			child = children[l];
			if (isElement(child) && !isLeaf(child)) {
				removeEmptyInlines(child);
				if (!child.firstChild && isInline(child)) {
					child.remove();
				}
			} else if (isTextNode(child) && !child.data) {
				child.remove();
			}
		}
	},
	// <br> elements are treated specially, and differently depending on the
	// browser, when in rich text editor mode. When adding HTML from external
	// sources, we must remove them, replacing the ones that actually affect
	// line breaks by wrapping the inline text in a <div>. Browsers that want <br>
	// elements at the end of each block will then have them added back in a later
	// fixCursor method call.
	cleanupBRs = (node, root, keepForBlankLine) => {
		const brs = node.querySelectorAll("BR");
		let l = brs.length;
		let br, parent;
		while (l--) {
			br = brs[l];
			// Cleanup may have removed it
			parent = br.parentNode;
			if (parent) {
				// If it doesn't break a line, just remove it; it's not doing
				// anything useful. We'll add it back later if required by the
				// browser. If it breaks a line, wrap the content in div tags
				// and replace the brs.
				if (!isLineBreak(br, keepForBlankLine)) {
					detach(br);
				} else if (!isInline(parent)) {
					fixContainer(parent, root);
				}
			}
		}
	},
	escapeHTML = text => text.replace('&', '&amp;')
		.replace('<', '&lt;')
		.replace('>', '&gt;')
		.replace('"', '&quot;'),

	// source/node/Block.ts
	getBlockWalker = (node, root) => {
		const walker = createTreeWalker(root, SHOW_ELEMENT, isBlock);
		walker.currentNode = node;
		return walker;
	},
	getPreviousBlock = (node, root) => {
//		node = getClosest(node, root, blockElementNames);
		node = getBlockWalker(node, root).previousNode();
		return node !== root ? node : null;
	},
	getNextBlock = (node, root) => {
//		node = getClosest(node, root, blockElementNames);
		node = getBlockWalker(node, root).nextNode();
		return node !== root ? node : null;
	},
	isEmptyBlock = block => !block.textContent && !block.querySelector("IMG"),

	// source/range/Block.ts
	// Returns the first block at least partially contained by the range,
	// or null if no block is contained by the range.
	getStartBlockOfRange = (range, root) => {
		const container = range.startContainer;
		let block;
		if (isInline(container)) {
			block = getPreviousBlock(container, root);
		} else if (container !== root && container instanceof HTMLElement && isBlock(container)) {
			block = container;
		} else {
			block = getNextBlock(getNodeBefore(container, range.startOffset), root);
		}
		return block && isNodeContainedInRange(range, block) ? block : null;
	},
	// Returns the last block at least partially contained by the range,
	// or null if no block is contained by the range.
	getEndBlockOfRange = (range, root) => {
		const container = range.endContainer;
		let block, child;
		if (isInline(container)) {
			block = getPreviousBlock(container, root);
		} else if (container !== root && container instanceof HTMLElement && isBlock(container)) {
			block = container;
		} else {
			block = getNodeAfterOffset(container, range.endOffset);
			if (!block || !root.contains(block)) {
				block = root;
				while (child = block.lastChild) {
					block = child;
				}
			}
			block = getPreviousBlock(block, root);
		}
		return block && isNodeContainedInRange(range, block) ? block : null;
	},
	rangeDoesStartAtBlockBoundary = (range, root) => {
		const startContainer = range.startContainer;
		const startOffset = range.startOffset;
		let nodeAfterCursor;

		// If in the middle or end of a text node, we're not at the boundary.
		if (isTextNode(startContainer)) {
			const text = startContainer.data;
			let i = startOffset;
			while (i--) {
				if (text.charAt(i) !== ZWS) {
					return false;
				}
			}
			nodeAfterCursor = startContainer;
		} else {
			nodeAfterCursor = getNodeAfterOffset(startContainer, startOffset);
			// The cursor was right at the end of the document
			if (!nodeAfterCursor || !root.contains(nodeAfterCursor)) {
				nodeAfterCursor = getNodeBefore(startContainer, startOffset);
				if (isTextNode(nodeAfterCursor) && nodeAfterCursor.length) {
					return false;
				}
			}
		}
		const block = getStartBlockOfRange(range, root);
		if (block) {
			const contentWalker = newContentWalker(block);
			contentWalker.currentNode = nodeAfterCursor;
			return !contentWalker.previousNode();
		}
	},
	rangeDoesEndAtBlockBoundary = (range, root) => {
		const endContainer = range.endContainer;
		const endOffset = range.endOffset;
		let currentNode;
		// If in a text node with content, and not at the end, we're not at the boundary
		if (isTextNode(endContainer)) {
			const text = endContainer.data;
			const length = text.length;
			for (let i = endOffset; i < length; ++i) {
				if (text.charAt(i) !== ZWS) {
					return false;
				}
			}
			currentNode = endContainer;
		} else {
			currentNode = getNodeBefore(endContainer, endOffset);
		}
		const block = getEndBlockOfRange(range, root);
		if (block) {
			const contentWalker = newContentWalker(block);
			contentWalker.currentNode = currentNode;
			return !contentWalker.nextNode();
		}
	},
	expandRangeToBlockBoundaries = (range, root) => {
		const start = getStartBlockOfRange(range, root);
		const end = getEndBlockOfRange(range, root);
//		let parent;
		if (start && end) {
			range.setStart(start, 0);
			range.setEnd(end, end.childNodes.length);
//			parent = start.parentNode;
//			range.setStart(parent, indexOf(parent.childNodes, start));
//			parent = end.parentNode;
//			range.setEnd(parent, indexOf(parent.childNodes, end) + 1);
		}
	},

	// source/range/InsertDelete.ts
	createRange = (startContainer, startOffset, endContainer, endOffset) => {
		const range = doc.createRange();
		range.setStart(startContainer, startOffset);
		if (endContainer && typeof endOffset === "number") {
			range.setEnd(endContainer, endOffset);
		} else {
			range.setEnd(startContainer, startOffset);
		}
		return range;
	},

	insertNodeInRange = (range, node) => {
		let { startContainer, startOffset, endContainer, endOffset } = range;
		let children, parent;

		// If part way through a text node, split it.
		if (isTextNode(startContainer)) {
			parent = startContainer.parentNode;
			children = parent.childNodes;
			if (startOffset === startContainer.length) {
				startOffset = indexOf(children, startContainer) + 1;
				if (range.collapsed) {
					endContainer = parent;
					endOffset = startOffset;
				}
			} else {
				if (startOffset) {
					const afterSplit = startContainer.splitText(startOffset);
					if (endContainer === startContainer) {
						endOffset -= startOffset;
						endContainer = afterSplit;
					} else if (endContainer === parent) {
						++endOffset;
					}
					startContainer = afterSplit;
				}
				startOffset = indexOf(children,
					startContainer
				);
			}
			startContainer = parent;
		} else {
			children = startContainer.childNodes;
		}
		const childCount = children.length;
		if (startOffset === childCount) {
			startContainer.append(node);
		} else {
			startContainer.insertBefore(node, children[startOffset]);
		}
		if (startContainer === endContainer) {
			endOffset += children.length - childCount;
		}
		range.setStart(startContainer, startOffset);
		range.setEnd(endContainer, endOffset);
	},

	extractContentsOfRange = (range, common, root) => {
		common = common || range.commonAncestorContainer;
		if (isTextNode(common)) {
			common = common.parentNode;
		}

		let endNode = split(range.endContainer, range.endOffset, common, root),
			frag = range.extractContents(),
			startContainer = common,
			startOffset = endNode ? indexOf(common.childNodes, endNode) : common.childNodes.length,
			after = common.childNodes[startOffset],
			before = after?.previousSibling,
			beforeText, afterText;

		// Merge text nodes if adjacent.
		if (isTextNode(before) && isTextNode(after)) {
			startContainer = before;
			startOffset = before.length;
			beforeText = before.data;
			afterText = after.data;

			// If we now have two adjacent spaces, the second one needs to become
			// a nbsp, otherwise the browser will swallow it due to HTML whitespace
			// collapsing.
			if (beforeText.charAt(beforeText.length - 1) === ' ' && afterText.charAt(0) === ' ') {
				afterText = NBSP + afterText.slice(1);
			}
			before.appendData(afterText);
			detach(after);
		}

		range.setStart(startContainer, startOffset);
		range.collapse(true);

		fixCursor(common);

		return frag;
	},
	getAdjacentInlineNode = (iterator, method, node) => {
		iterator.currentNode = node;
		let nextNode;
		while (nextNode = iterator[method]()) {
			if (isTextNode(nextNode) || isLeaf(nextNode)) {
				return nextNode;
			}
			if (!isInline(nextNode)) {
				return null;
			}
		}
		return null;
	},
	deleteContentsOfRange = (range, root) => {
		const startBlock = getStartBlockOfRange(range, root);
		let endBlock = getEndBlockOfRange(range, root);
		const needsMerge = startBlock !== endBlock;
		if (startBlock && endBlock) {
			moveRangeBoundariesDownTree(range);
			moveRangeBoundariesUpTree(range, startBlock, endBlock, root);
		}

		// Remove selected range
		const frag = extractContentsOfRange(range, null, root);

		// Move boundaries back down tree as far as possible.
		moveRangeBoundariesDownTree(range);

		// If we split into two different blocks, merge the blocks.
		if (needsMerge) {
			// endBlock will have been split, so need to refetch
			endBlock = getEndBlockOfRange(range, root);
			if (startBlock && endBlock && startBlock !== endBlock) {
				mergeWithBlock(startBlock, endBlock, range, root);
			}
		}

		// Ensure block has necessary children
		if (startBlock) {
			fixCursor(startBlock);
		}

		// Ensure root has a block-level element in it.
		const child = root.firstChild;
		if (!child || isBrElement(child)) {
			fixCursor(root);
			root.firstChild && range.selectNodeContents(root.firstChild);
		}
		range.collapse(true);
		const startContainer = range.startContainer;
		const startOffset = range.startOffset;
		const iterator = newContentWalker(root);
		let afterNode = startContainer;
		let afterOffset = startOffset;
		if (!isTextNode(afterNode) || afterOffset === afterNode.data.length) {
			afterNode = getAdjacentInlineNode(iterator, "nextNode", afterNode);
			afterOffset = 0;
		}
		let beforeNode = startContainer;
		let beforeOffset = startOffset - 1;
		if (!isTextNode(beforeNode) || beforeOffset === -1) {
			beforeNode = getAdjacentInlineNode(
				iterator,
				"previousPONode",
				afterNode || (isTextNode(startContainer) ? startContainer : startContainer.childNodes[startOffset] || startContainer)
			);
			if (isTextNode(beforeNode)) {
				beforeOffset = beforeNode.data.length;
			}
		}
		let node = null;
		let offset = 0;
		if (isTextNode(afterNode) && afterNode.data.charAt(afterOffset) === " " && rangeDoesStartAtBlockBoundary(range, root)) {
			node = afterNode;
			offset = afterOffset;
		} else if (isTextNode(beforeNode) && beforeNode.data.charAt(beforeOffset) === " ") {
			if (isTextNode(afterNode) && afterNode.data.charAt(afterOffset) === " " || rangeDoesEndAtBlockBoundary(range, root)) {
				node = beforeNode;
				offset = beforeOffset;
			}
		}
		node && node.replaceData(offset, 1, "\xA0");
		range.setStart(startContainer, startOffset);
		range.collapse(true);
		return frag;
	},

	// Contents of range will be deleted.
	// After method, range will be around inserted content
	insertTreeFragmentIntoRange = (range, frag, root) => {
		const firstInFragIsInline = frag.firstChild && isInline(frag.firstChild);
		let node, block, blockContentsAfterSplit, stopPoint, container, offset;
		let replaceBlock, firstBlockInFrag, nodeAfterSplit, nodeBeforeSplit;
		let tempRange;

		// Fixup content: ensure no top-level inline, and add cursor fix elements.
		fixContainer(frag, root);
		node = frag;
		while (node = getNextBlock(node, root)) {
			fixCursor(node);
		}

		// Delete any selected content.
		if (!range.collapsed) {
			deleteContentsOfRange(range, root);
		}

		// Move range down into text nodes.
		moveRangeBoundariesDownTree(range);
		range.collapse(); // collapse to end

		// Where will we split up to? First blockquote parent, otherwise root.
		stopPoint = getClosest(range.endContainer, root, "BLOCKQUOTE") || root;

		// Merge the contents of the first block in the frag with the focused block.
		// If there are contents in the block after the focus point, collect this
		// up to insert in the last block later. This preserves the style that was
		// present in this bit of the page.
		//
		// If the block being inserted into is empty though, replace it instead of
		// merging if the fragment had block contents.
		// e.g. <blockquote><p>Foo</p></blockquote>
		// This seems a reasonable approximation of user intent.

		block = getStartBlockOfRange(range, root);
		firstBlockInFrag = getNextBlock(frag, frag);
		replaceBlock = !firstInFragIsInline && !!block && isEmptyBlock(block);
		if (block && firstBlockInFrag && !replaceBlock && // Don't merge table cells or PRE elements into block
		!getClosest(firstBlockInFrag, frag, "PRE,TABLE")) {
			moveRangeBoundariesUpTree(range, block, block, root);
			range.collapse(true); // collapse to start
			container = range.endContainer;
			offset = range.endOffset;
			// Remove trailing <br> – we don't want this considered content to be
			// inserted again later
			cleanupBRs(block, root, false);
			if (isInline(container)) {
				// Split up to block parent.
				nodeAfterSplit = split(
					container,
					offset,
					getPreviousBlock(container, root) || root,
					root
				);
				container = nodeAfterSplit.parentNode;
				offset = indexOf(container.childNodes, nodeAfterSplit);
			}
			if (/*isBlock(container) && */
				offset !== getLength(container)
			) {
				// Collect any inline contents of the block after the range point
				blockContentsAfterSplit = doc.createDocumentFragment();
				while (node = container.childNodes[offset]) {
					blockContentsAfterSplit.append(node);
				}
			}
			// And merge the first block in.
			mergeWithBlock(container, firstBlockInFrag, range, root);

			// And where we will insert
			offset = indexOf(container.parentNode.childNodes,
				container
			) + 1;
			container = container.parentNode;
			range.setEnd(container, offset);
		}

		// Is there still any content in the fragment?
		if (getLength(frag)) {
			if (replaceBlock && block) {
				range.setEndBefore(block);
				range.collapse();
				detach(block);
			}
			moveRangeBoundariesUpTree(range, stopPoint, stopPoint, root);
			// Now split after block up to blockquote (if a parent) or root
			nodeAfterSplit = split(
				range.endContainer,
				range.endOffset,
				stopPoint,
				root
			);
			nodeBeforeSplit = nodeAfterSplit ? nodeAfterSplit.previousSibling : stopPoint.lastChild;
			stopPoint.insertBefore(frag, nodeAfterSplit);
			if (nodeAfterSplit) {
				range.setEndBefore(nodeAfterSplit);
			} else {
				range.setEnd(stopPoint, getLength(stopPoint));
			}
			block = getEndBlockOfRange(range, root);

			// Get a reference that won't be invalidated if we merge containers.
			moveRangeBoundariesDownTree(range);
			container = range.endContainer;
			offset = range.endOffset;

			// Merge inserted containers with edges of split
			if (nodeAfterSplit && isContainer(nodeAfterSplit)) {
				mergeContainers(nodeAfterSplit, root);
			}
			nodeAfterSplit = nodeBeforeSplit?.nextSibling;
			if (nodeAfterSplit && isContainer(nodeAfterSplit)) {
				mergeContainers(nodeAfterSplit, root);
			}
			range.setEnd(container, offset);
		}

		// Insert inline content saved from before.
		if (blockContentsAfterSplit && block) {
			tempRange = range.cloneRange();
			mergeWithBlock(block, blockContentsAfterSplit, tempRange, root);
			range.setEnd(tempRange.endContainer, tempRange.endOffset);
		}
		moveRangeBoundariesDownTree(range);
	},
/*
	// source/range/Contents.ts
	getTextContentsOfRange = (range) => {
		if (range.collapsed) {
			return "";
		}
		const startContainer = range.startContainer;
		const endContainer = range.endContainer;
		const walker = new TreeIterator(
			range.commonAncestorContainer,
			SHOW_ELEMENT_OR_TEXT,
			(node2) => {
				return isNodeContainedInRange(range, node2, true);
			}
		);
		walker.currentNode = startContainer;
		let node = startContainer;
		let textContent = "";
		let addedTextInBlock = false;
		let value;
		if (!isElement(node) && !isTextNode(node) || !walker.filter(node)) {
			node = walker.nextNode();
		}
		while (node) {
			if (isTextNode(node)) {
				value = node.data;
				if (value && /\S/.test(value)) {
					if (node === endContainer) {
						value = value.slice(0, range.endOffset);
					}
					if (node === startContainer) {
						value = value.slice(range.startOffset);
					}
					textContent += value;
					addedTextInBlock = true;
				}
			} else if (isBrElement(node) || addedTextInBlock && !isInline(node)) {
				textContent += "\n";
				addedTextInBlock = false;
			}
			node = walker.nextNode();
		}
		textContent = textContent.replace(/\xA0/g, " ");
		return textContent;
	},
*/
	// source/Clipboard.ts
	indexOf = (array, value) => Array.prototype.indexOf.call(array, value),
	extractRangeToClipboard = (event, range, root) => {
		// Edge only seems to support setting plain text as of 2016-03-11.
		if (event.clipboardData) {
			// Clipboard content should include all parents within block, or all
			// parents up to root if selection across blocks
			let startBlock = getStartBlockOfRange(range, root),
				endBlock = getEndBlockOfRange(range, root),
				copyRoot = ((startBlock === endBlock) && startBlock) || root,
				contents, parent, newContents;
			// Clone range to mutate, then move up as high as possible without
			// passing the copy root node.
			range = range.cloneRange();
			moveRangeBoundariesDownTree(range);
			moveRangeBoundariesUpTree(range, copyRoot, copyRoot, root);
			// Extract the contents
			contents = range.cloneContents();
			// Add any other parents not in extracted content, up to copy root
			parent = range.commonAncestorContainer;
			if (isTextNode(parent)) {
				parent = parent.parentNode;
			}
			while (parent && parent !== copyRoot) {
				newContents = parent.cloneNode(false);
				newContents.append(contents);
				contents = newContents;
				parent = parent.parentNode;
			}
			// Set clipboard data
			setClipboardData(event, contents, root);
		}
	},
	// The (non-standard but supported enough) innerText property is based on the
	// render tree in Firefox and possibly other browsers, so we must insert the
	// DOM node into the document to ensure the text part is correct.
	setClipboardData = (event, contents, root) => {
		let clipboardData = event.clipboardData;
		let body = doc.body;
		let node = createElement("div");
		let html, text;

		node.append(contents);

		html = node.innerHTML;

		// Firefox will add an extra new line for BRs at the end of block when
		// calculating innerText, even though they don't actually affect
		// display, so we need to remove them first.
		cleanupBRs(node, root, true);
		node.setAttribute("style",
			'position:fixed;overflow:hidden;bottom:100%;right:100%;');
		body.append(node);
		text = (node.innerText || node.textContent).replace(NBSP, ' '); // Replace nbsp with regular space
		node.remove();

		if (text !== html) {
			clipboardData.setData("text/html", html);
		}
		clipboardData.setData("text/plain", text);
		event.preventDefault();
	},
	onCut = function(event) {
		let self = this;
		let range = self.getSelection();
		let root = self._root;
		let startBlock, endBlock, copyRoot, contents, parent, newContents;

		// Nothing to do
		if (range.collapsed) {
			event.preventDefault();
			return;
		}

		// Save undo checkpoint
		self.saveUndoState(range);

		// Edge only seems to support setting plain text as of 2016-03-11.
		if (event.clipboardData) {
			// Clipboard content should include all parents within block, or all
			// parents up to root if selection across blocks
			startBlock = getStartBlockOfRange(range, root);
			endBlock = getEndBlockOfRange(range, root);
			copyRoot = ((startBlock === endBlock) && startBlock) || root;
			// Extract the contents
			contents = deleteContentsOfRange(range, root);
			// Add any other parents not in extracted content, up to copy root
			parent = range.commonAncestorContainer;
			if (isTextNode(parent)) {
				parent = parent.parentNode;
			}
			while (parent && parent !== copyRoot) {
				newContents = parent.cloneNode(false);
				newContents.append(contents);
				contents = newContents;
				parent = parent.parentNode;
			}
			// Set clipboard data
			setClipboardData(event, contents, root);
		} else {
			setTimeout(() => {
				try {
					// If all content removed, ensure div at start of root.
					self._ensureBottomLine();
				} catch (error) {
					didError(error);
				}
			}, 0);
		}

		self.setSelection(range);
	},
	onCopy = function(event) {
		extractRangeToClipboard(
			event,
			this.getSelection(),
			this._root
		);
	},
	onPaste = function(event) {
		const clipboardData = event.clipboardData;
		const items = clipboardData?.items;
		let imageItem = null;
		let plainItem = null;
		let htmlItem = null;
		let self = this;
		let type;

		// Current HTML5 Clipboard interface
		// ---------------------------------
		// https://html.spec.whatwg.org/multipage/interaction.html
		if (items) {
			[...items].forEach(item => {
				type = item.type;
				if (type === 'text/html') {
					htmlItem = item;
				// iOS copy URL gives you type text/uri-list which is just a list
				// of 1 or more URLs separated by new lines. Can just treat as
				// plain text.
				} else if (type === 'text/plain' || type === 'text/uri-list') {
					plainItem = item;
				} else if (item.kind === "file" && /^image\/(png|jpeg|webp)/.test(type)) {
					imageItem = item;
				}
			});
			if (htmlItem || plainItem || imageItem) {
				event.preventDefault();
				if (imageItem) {
					let reader = new FileReader();
					reader.onload = event => {
						let img = createElement("img", {src: event.target.result}),
							canvas = createElement("canvas"),
							ctx = canvas.getContext('2d');
						img.onload = ()=>{
							ctx.drawImage(img, 0, 0);
							let width = img.width, height = img.height;
							if (width > height) {
								// Landscape
								if (width > 1024) {
									height = height * 1024 / width;
									width = 1024;
								}
							} else if (height > 1024) {
								// Portrait
								width = width * 1024 / height;
								height = 1024;
							}
							canvas.width = width;
							canvas.height = height;
							ctx.drawImage(img, 0, 0, width, height);
							self.insertHTML('<img alt="" style="width:100%;max-width:'+width+'px" src="'+canvas.toDataURL()+'">', true);
						};
					}
					reader.readAsDataURL(imageItem.getAsFile());
				} else if (htmlItem && (!self.isShiftDown || !plainItem)) {
					htmlItem.getAsString(html => self.insertHTML(html, true));
				} else if (plainItem) {
					plainItem.getAsString(text => self.insertPlainText(text, true));
				}
			}
		}
	},

	// source/keyboard/KeyHelpers.ts
	// If you delete the content inside a span with a font styling, Webkit will
	// replace it with a <font> tag (!). If you delete all the text inside a
	// link in Opera, it won't delete the link. Let's make things consistent. If
	// you delete all text inside an inline tag, remove the inline tag.
	afterDelete = (self, range) => {
		try {
			range = range || self.getSelection();
			let node = range.startContainer,
				parent;
			// Climb the tree from the focus point while we are inside an empty
			// inline element
			if (isTextNode(node)) {
				node = node.parentNode;
			}
			parent = node;
			while (isInline(parent) && (!parent.textContent || parent.textContent === ZWS)) {
				node = parent;
				parent = node.parentNode;
			}
			// If focused in empty inline element
			if (node !== parent) {
				// Move focus to just before empty inline(s)
				range.setStart(parent, indexOf(parent.childNodes, node));
				range.collapse(true);
				// Remove empty inline(s)
				node.remove();
				// Fix cursor in block
				if (!isBlock(parent)) {
					parent = getPreviousBlock(parent, self._root) || parent;
				}
				fixCursor(parent);
				// Move cursor into text node
				moveRangeBoundariesDownTree(range);
			}
			// If you delete the last character in the sole <div> in Chrome,
			// it removes the div and replaces it with just a <br> inside the
			// root. Detach the <br>; the _ensureBottomLine call will insert a new
			// block.
			if (node === self._root && (node = node.firstChild) && isBrElement(node)) {
				detach(node);
			}
			self._ensureBottomLine();
			self.setRange(range);
		} catch (error) {
			didError(error);
		}
	},

	// source/keyboard/Backspace.ts
	Backspace = (self, event, range) => {
		const root = self._root;
		self._removeZWS();
		// Record undo checkpoint.
		self.saveUndoState(range);
		// If not collapsed, delete contents
		if (!range.collapsed) {
			event.preventDefault();
			deleteContentsOfRange(range, root);
			afterDelete(self, range);
		}
		// If at beginning of block, merge with previous
		else if (rangeDoesStartAtBlockBoundary(range, root)) {
			event.preventDefault();
			let current = getStartBlockOfRange(range, root);
			let previous;
			if (!current) {
				return;
			}
			// In case inline data has somehow got between blocks.
			fixContainer(current.parentNode, root);
			// Now get previous block
			previous = getPreviousBlock(current, root);
			// Must not be at the very beginning of the text area.
			if (previous) {
				// If not editable, just delete whole block.
				if (!previous.isContentEditable) {
					detachUneditableNode(previous, root);
					return;
				}
				// Otherwise merge.
				mergeWithBlock(previous, current, range, root);
				// If deleted line between containers, merge newly adjacent
				// containers.
				current = previous.parentNode;
				while (current !== root && !current.nextSibling) {
					current = current.parentNode;
				}
				if (current !== root && (current = current.nextSibling)) {
					mergeContainers(current, root);
				}
				self.setSelection(range);
			}
			// If at very beginning of text area, allow backspace
			// to break lists/blockquote.
			else if (current) {
				if (decreaseLevel(self, range, current)) {
					return;
				}
				self.setRange(range);
			}
		}
		// Otherwise, leave to browser but check afterwards whether it has
		// left behind an empty inline tag.
		else {
			self.setSelection(range);
			setTimeout(() => afterDelete(self), 0);
		}
	},

	// source/keyboard/Delete.ts
	Delete = (self, event, range) => {
		const root = self._root;
		let current;
		let next;
		let originalRange;
		let cursorContainer;
		let cursorOffset;
		let nodeAfterCursor;
		self._removeZWS();
		// Record undo checkpoint.
		self.saveUndoState(range);
		if (!range.collapsed) {
			// If not collapsed, delete contents
			event.preventDefault();
			deleteContentsOfRange(range, root);
			afterDelete(self, range);
		} else if (rangeDoesEndAtBlockBoundary(range, root)) {
			// If at end of block, merge next into this block
			event.preventDefault();
			if (current = getStartBlockOfRange(range, root)) {
				// In case inline data has somehow got between blocks.
				fixContainer(current.parentNode, root);
				// Now get next block
				// Must not be at the very end of the text area.
				if (next = getNextBlock(current, root)) {
					// If not editable, just delete whole block.
					if (!next.isContentEditable) {
						detachUneditableNode(next, root);
						return;
					}
					// Otherwise merge.
					mergeWithBlock(current, next, range, root);
					// If deleted line between containers, merge newly adjacent
					// containers.
					next = current.parentNode;
					while (next !== root && !next.nextSibling) {
						next = next.parentNode;
					}
					if (next !== root && (next = next.nextSibling)) {
						mergeContainers(next, root);
					}
					self.setRange(range);
				}
			}
		} else {
			// Otherwise, leave to browser but check afterwards whether it has
			// left behind an empty inline tag.
			// But first check if the cursor is just before an IMG tag. If so,
			// delete it ourselves, because the browser won't if it is not
			// inline.
			originalRange = range.cloneRange();
			moveRangeBoundariesUpTree(range, root, root, root);
			cursorContainer = range.endContainer;
			cursorOffset = range.endOffset;
			if (isElement(cursorContainer)) {
				nodeAfterCursor = cursorContainer.childNodes[cursorOffset];
				if (nodeAfterCursor?.nodeName === "IMG") {
					event.preventDefault();
					detach(nodeAfterCursor);
					moveRangeBoundariesDownTree(range);
					afterDelete(self, range);
					return;
				}
			}
			self.setSelection(originalRange);
			setTimeout(() => afterDelete(self), 0);
		}
	},

	// source/keyboard/Tab.ts
	Tab = (self, event, range) => {
		self._removeZWS();
		// If no selection and at start of block
		if (range.collapsed && rangeDoesStartAtBlockBoundary(range, self._root)) {
			getClosest(range.startContainer, self._root, "UL,OL,BLOCKQUOTE")
			&& self.changeIndentationLevel('increase')
			&& event.preventDefault();
		}
	},
	ShiftTab = (self, event, range) => {
		self._removeZWS();
		// If no selection and at start of block
		if (range.collapsed && rangeDoesStartAtBlockBoundary(range, self._root)) {
			// Break list
			decreaseLevel(self, range, range.startContainer)
			&& event.preventDefault();
		}
	},

	// source/keyboard/Space.ts
	Space = (self, event, range) => {
/*
		var _a;
		let node;
		const root = self._root;
		self._recordUndoState(range);
		self._getRangeAndRemoveBookmark(range);
		if (!range.collapsed) {
			deleteContentsOfRange(range, root);
			self._ensureBottomLine();
			self.setSelection(range);
			self._updatePath(range, true);
		} else if (rangeDoesEndAtBlockBoundary(range, root)) {
			const block = getStartBlockOfRange(range, root);
			if (block && block.nodeName !== "PRE") {
				const text = (_a = block.textContent) == null ? void 0 : _a.trimEnd().replace(ZWS, "");
				if (text === "*" || text === "1.") {
					event.preventDefault();
					const walker = new TreeIterator(block, SHOW_TEXT);
					let textNode;
					while (textNode = walker.nextNode()) {
						textNode.data = cantFocusEmptyTextNodes ? ZWS : "";
					}
					if (text === "*") {
						self.makeUnorderedList();
					} else {
						self.makeOrderedList();
					}
					return;
				}
			}
		}
		node = range.endContainer;
		if (range.endOffset === getLength(node)) {
			do {
				if (node.nodeName === "A") {
					range.setStartAfter(node);
					break;
				}
			} while (!node.nextSibling && (node = node.parentNode) && node !== root);
		}
		if (self._config.addLinks) {
			const linkRange = range.cloneRange();
			moveRangeBoundariesDownTree(linkRange);
			const textNode = linkRange.startContainer;
			const offset = linkRange.startOffset;
			setTimeout(() => {
				linkifyText(self, textNode, offset);
			}, 0);
		}
		self.setSelection(range);
*/
		const root = self._root;
		self._recordUndoState(range);
		self._config.addLinks && addLinks(range.startContainer, root);
		self._getRangeAndRemoveBookmark(range);
/*
		// If the cursor is at the end of a link (<a>foo|</a>) then move it
		// outside of the link (<a>foo</a>|) so that the space is not part of
		// the link text.
		// SnappyMail: disabled as it fails in Firefox
		let node = range.endContainer;
		if (range.collapsed && range.endOffset === getLength(node)) {
			do {
				if (node.nodeName === "A") {
					range.setStartAfter(node);
					break;
				}
			} while (!node.nextSibling && (node = node.parentNode) && node !== root);
		}
*/
		// Delete the selection if not collapsed
		if (!range.collapsed) {
			deleteContentsOfRange(range, root);
			self._ensureBottomLine();
		}
		self.setRange(range);
	},

	// source/keyboard/KeyHandlers.ts
	onKey = function(event) {
		if (event.defaultPrevented) {
			return;
		}

		let key = event.key.toLowerCase(),
			range = this.getSelection(),
			root = this._root;

		// We need to apply the backspace/delete handlers regardless of control key modifiers.
		// ctrl-shift-key or meta-shift-key
		if (key !== "backspace" && key !== "delete") {
			if (event.shiftKey) { key = 'shift-' + key; }
			if (event[osKey]) { key = ctrlKey + key; }
//			if (event.altKey) { key = 'alt-' + key; }
		}
		if (this._keyHandlers[key]) {
			this._keyHandlers[key](this, event, range);
		// !event.isComposing stops us from blatting Kana-Kanji conversion in Safari
		} else if (!range.collapsed && !event.isComposing && !event[osKey] && key.length === 1) {
			// Record undo checkpoint.
			this.saveUndoState(range);
			// Delete the selection
			deleteContentsOfRange(range, root);
			this._ensureBottomLine();
			this.setRange(range);
		} else if (range.collapsed && range.startContainer === root && root.children.length > 0) {
			// Under certain conditions, cursor/range can be positioned directly
			// under this._root (not wrapped) and when this happens, an inline(TEXT)
			// element is attached directly to this._root. There might be other
			// issues, but squire.makeUnorderedList(), squire.makeOrderedList() and
			// maybe other functions that call squire.modifyBlocks() do NOT work
			// on inline(TEXT) nodes that are direct children of this._root.
			// Therefor, we try to detect this case here and wrap the cursor/range
			// before the text is inserted.

			const nextElement = root.children[range.startOffset];
			if (nextElement && !isBlock(nextElement)) {
				// create a new wrapper
				range = createRange(root.insertBefore(
					this.createDefaultBlock(), nextElement
				), 0);
				if (isBrElement(nextElement)) {
					// delete it because a new <br> is created by createDefaultBlock()
					root.removeChild(nextElement);
				}
				const restore = this._restoreSelection;
				this.setSelection(range);
				this._restoreSelection = restore;
			}
		}
	},
	keyHandlers = {
//		"backspace": Backspace,
//		"delete": Delete,
		tab: Tab,
		'shift-tab': ShiftTab,
		space: Space,
		arrowleft: self => self._removeZWS(),
		arrowright: self => self._removeZWS()
	},
	mapKeyToFormat = (tag, remove) => {
		return (self, event) => {
			event.preventDefault();
			self.toggleTag(tag, remove);
		};
	},
	mapKeyTo = method => (self, event) => {
		event.preventDefault();
		self[method]();
	},

	blockTag = "DIV",
	DOCUMENT_POSITION_PRECEDING = 2, // Node.DOCUMENT_POSITION_PRECEDING

	NBSP = '\u00A0',

	win = doc.defaultView,

	osKey = isMac ? "metaKey" : "ctrlKey",

	filterAccept = NodeFilter.FILTER_ACCEPT,
/*
	typeToBitArray = {
		// ELEMENT_NODE
		1: 1,
		// ATTRIBUTE_NODE
		2: 2,
		// TEXT_NODE
		3: 4,
		// COMMENT_NODE
		8: 128,
		// DOCUMENT_NODE
		9: 256,
		// DOCUMENT_FRAGMENT_NODE
		11: 1024
	},
*/

	isElement = node => node instanceof Element,
	isTextNode = node => node instanceof Text,
//	isBrElement = node => node instanceof HTMLBRElement,
	isBrElement = node => "BR" === node?.nodeName,

	createTreeWalker = (root, whatToShow, filter) => {
		const walker = doc.createTreeWalker(root, whatToShow, filter ? {
			acceptNode: node => filter(node) ? filterAccept : NodeFilter.FILTER_SKIP
		} : null);
		walker.previousPONode = () => previousPONode(walker, filter);
		return walker;
	},

	setAttributes = (node, props) => {
		props && Object.entries(props).forEach(([k,v]) => {
			if (null == v) {
				node.removeAttribute(k);
			} else if ("style" === k && typeof v === "object") {
				Object.entries(v).forEach(([k,v]) => node.style[k] = v);
			} else {
				node.setAttribute(k, v);
			}
		});
	},

	newContentWalker = root => createTreeWalker(root,
		SHOW_ELEMENT_OR_TEXT,
		node => isTextNode(node) ? notWS.test(node.data) : node.nodeName === "IMG"
	),

	didError = error => console.error(error),

	detachUneditableNode = (node, root) => {
		let parent;
		while (parent = node.parentNode) {
			if (parent === root || parent.isContentEditable) {
				break;
			}
			node = parent;
		}
		detach(node);
	},

	changeIndentationLevel = direction => (self, event) => {
		event.preventDefault();
		self.changeIndentationLevel(direction);
	},

	toggleList = (type, methodIfNotInList) => (self, event) => {
		event.preventDefault();
		let parent = self.getSelectionClosest('UL,OL');
		if (type == parent?.nodeName) {
			self.removeList();
		} else {
			self[methodIfNotInList]();
		}
	},

	// TreeIterator Previous node in post-order.
	previousPONode = (walker,  filter) => {
		let current = walker.currentNode,
			root = walker.root,
			node;
		const isAcceptableNode = node => {
			const nodeType = node.nodeType;
			const nodeFilterType = nodeType === ELEMENT_NODE ? SHOW_ELEMENT : nodeType === TEXT_NODE ? SHOW_TEXT : 0;
			return !!(nodeFilterType & walker.whatToShow) && filter(node);
		}
		while (true) {
			if (current === root) {
				return null;
			}
			node = current.previousSibling;
			if (node) {
				while (current = node.lastChild) {
					node = current;
				}
			} else {
				node = current.parentNode;
			}
			if (!node) {
				return null;
			}
			if (isAcceptableNode(node)) {
				walker.currentNode = node;
				return node;
			}
			current = node;
		}
	},

	mergeObjects = (base, extras, mayOverride) => {
		base = base || {};
		extras && Object.entries(extras).forEach(([prop,value])=>{
			if (mayOverride || !(prop in base)) {
				base[prop] = (value?.constructor === Object) ?
					mergeObjects(base[prop], value, mayOverride) :
					value;
			}
		});
		return base;
	},

	// --- Events ---

	// Subscribing to these events won't automatically add a listener to the
	// document node, since these events are fired in a custom manner by the
	// editor code.
	customEvents = {
		pathChange: 1, select: 1, input: 1, undoStateChange: 1
	},

	// --- Bookmarking ---

	startSelectionId = 'squire-selection-start',
	endSelectionId = 'squire-selection-end',

	createBookmarkNodes = () => [
		createElement("INPUT", {
			id: startSelectionId,
			type: "hidden"
		}),
		createElement("INPUT", {
			id: endSelectionId,
			type: "hidden"
		})
	],

	// --- Block formatting ---

	tagAfterSplit = {
		DT:  "DD",
		DD:  "DT",
		LI:  "LI",
		PRE: "PRE"
	},

	getListSelection = (range, root) => {
		// Get start+end li in single common ancestor
		let list = range.commonAncestorContainer;
		let startLi = range.startContainer;
		let endLi = range.endContainer;
		while (list && list !== root && !listNodeNames.has(list.nodeName)) {
			list = list.parentNode;
		}
		if (!list || list === root) {
			return null;
		}
		if (startLi === list) {
			startLi = startLi.childNodes[range.startOffset];
		}
		if (endLi === list) {
			endLi = endLi.childNodes[range.endOffset];
		}
		while (startLi && startLi.parentNode !== list) {
			startLi = startLi.parentNode;
		}
		while (endLi && endLi.parentNode !== list) {
			endLi = endLi.parentNode;
		}
		return [list, startLi, endLi];
	},

	makeList = (self, frag, type) => {
		let walker = getBlockWalker(frag, self._root),
			node, tag, prev, newLi;

		while (node = walker.nextNode()) {
			if (node.parentNode.nodeName === "LI") {
				node = node.parentNode;
				walker.currentNode = node.lastChild;
			}
			if (node.nodeName !== "LI") {
				newLi = createElement("LI");
				if (node.dir) {
					newLi.dir = node.dir;
				}

				// Have we replaced the previous block with a new <ul>/<ol>?
				if ((prev = node.previousSibling) && prev.nodeName === type) {
					prev.append(newLi);
					detach(node);
				}
				// Otherwise, replace this block with the <ul>/<ol>
				else {
					node.replaceWith(
						createElement(type, null, [
							newLi
						])
					);
				}
				newLi.append(empty(node));
				walker.currentNode = newLi;
			} else {
				node = node.parentNode;
				tag = node.nodeName;
				if (tag !== type && listNodeNames.has(tag)) {
					node.replaceWith(
						createElement(type, null, [empty(node)])
					);
				}
			}
		}

		return frag;
	},

	setDirection = (self, frag, dir) => {
		let walker = getBlockWalker(frag, self._root),
			node;

		while (node = walker.nextNode()) {

			if (node.nodeName === "LI") {
				node.parentNode.setAttribute("dir", dir);
				break;
			}

			node.setAttribute("dir", dir);
		}
		return frag;
	},

	decreaseLevel = (self, range, node) =>
		getClosest(node, self._root, 'UL,OL,BLOCKQUOTE') && self.changeIndentationLevel('decrease'),

	linkRegExp = /\b(?:((https?:\/\/)?(?:www\d{0,3}\.|[a-z0-9][a-z0-9.-]*\.[a-z]{2,}\/)(?:[^\s()<>]+|\([^\s()<>]+\))+(?:[^\s?&`!()[\]{};:'".,<>«»“”‘’]|\([^\s()<>]+\)))|([\w\-.%+]+@(?:[\w-]+\.)+[a-z]{2,}\b(?:\?[^&?\s]+=[^\s?&`!()[\]{};:'".,<>«»“”‘’]+(?:&[^&?\s]+=[^\s?&`!()[\]{};:'".,<>«»“”‘’]+)*)?))/i,

	addLinks = (frag, root) => {
		let walker = createTreeWalker(frag, SHOW_TEXT, node => !getClosest(node, root, "A"));
		let node, data, parent, match, index, endIndex, child;
		while (node = walker.nextNode()) {
			data = node.data;
			parent = node.parentNode;
			while (match = linkRegExp.exec(data)) {
				index = match.index;
				endIndex = index + match[0].length;
				if (index) {
					child = doc.createTextNode(data.slice(0, index));
					parent.insertBefore(child, node);
				}
				child = createElement("A", {
					href: match[1]
						? (match[2] ? match[1] : 'https://' + match[1])
						: 'mailto:' + match[0]
				}, [data.slice(index, endIndex)]);
				parent.insertBefore(child, node);
				node.data = data = data.slice(endIndex);
			}
		}
	};

let nodeCategoryCache = new WeakMap();

keyHandlers[ctrlKey + "b"] = mapKeyToFormat("B");
keyHandlers[ctrlKey + "i"] = mapKeyToFormat("I");
keyHandlers[ctrlKey + "u"] = mapKeyToFormat("U");
keyHandlers[ctrlKey + 'shift-7'] = mapKeyToFormat("S");
keyHandlers[ctrlKey + 'shift-5'] = mapKeyToFormat("SUB", "SUP");
keyHandlers[ctrlKey + 'shift-6'] = mapKeyToFormat("SUP", "SUB");
keyHandlers[ctrlKey + 'shift-8'] = toggleList("UL", "makeUnorderedList");
keyHandlers[ctrlKey + 'shift-9'] = toggleList("OL", "makeOrderedList");
keyHandlers[ctrlKey + '['] = changeIndentationLevel("decrease");
keyHandlers[ctrlKey + ']'] = changeIndentationLevel("increase");
keyHandlers[ctrlKey + "d"] = mapKeyTo("toggleCode");
keyHandlers[ctrlKey + "y"] = mapKeyTo("redo");
//keyHandlers[ctrlKey + "z"] = mapKeyTo("undo");
keyHandlers[ctrlKey + 'shift-z'] = mapKeyTo("redo");
keyHandlers["redo"] = mapKeyTo("redo");
//keyHandlers["undo"] = mapKeyTo("undo");

class EditStack extends Array
{
	constructor(squire) {
		super();
		this.squire = squire;
		this.index = -1;
		this.inUndoState = false;

		this.threshold = -1; // -1 means no threshold
		this.limit = -1; // -1 means no limit
	}

	clear() {
		this.index = -1;
		this.length = 0;
	}

	stateChanged(/*canUndo, canRedo*/) {
		this.squire.fireEvent("undoStateChange", {
			canUndo: this.index > 0,
			canRedo: this.index + 1 < this.length
		});
		this.squire.fireEvent("input");
	}

	docWasChanged() {
		if (this.inUndoState) {
			this.inUndoState = false;
			this.stateChanged(/*true, false*/);
		} else
			this.squire.fireEvent("input");
	}

	/**
	 * Leaves bookmark.
	 */
	recordUndoState(range, replace) {
		// Don't record if we're already in an undo state
		if (!this.inUndoState || replace) {
			// Advance pointer to new position
			let undoIndex = this.index;
			let undoThreshold = this.threshold;
			let undoLimit = this.limit;
			let squire = this.squire;

			replace || ++undoIndex;
			undoIndex = Math.max(0, undoIndex);

			// Truncate stack if longer (i.e. if has been previously undone)
			this.length = Math.min(undoIndex + 1, this.length);

			// Get data
			range && squire._saveRangeToBookmark(range);
			const html = squire._getHTML();

			// If this document is above the configured size threshold,
			// limit the number of saved undo states.
			// Threshold is in bytes, JS uses 2 bytes per character
			if (undoThreshold > -1 && html.length * 2 > undoThreshold
			 && undoLimit > -1 && undoIndex > undoLimit) {
				this.splice(0, undoIndex - undoLimit);
				undoIndex = undoLimit;
			}

			// Save data
			this[undoIndex] = html;
			this.index = undoIndex;
			this.inUndoState = true;
		}
	}

	saveUndoState(range) {
		let squire = this.squire;
		range = range || squire.getSelection();
		this.recordUndoState(range, true);
		squire._getRangeAndRemoveBookmark(range);
	}

	undo() {
		let squire = this.squire,
			undoIndex = this.index - 1;
		// Sanity check: must not be at beginning of the history stack
		if (undoIndex >= 0 || !this.inUndoState) {
			// Make sure any changes since last checkpoint are saved.
			this.recordUndoState(squire.getSelection());
			this.index = undoIndex;
			squire._setHTML(this[undoIndex]);
			let range = squire._getRangeAndRemoveBookmark();
			if (range) {
				squire.setSelection(range);
			}
			this.inUndoState = true;
			this.stateChanged(/*undoIndex > 0, true*/);
		}
	}

	redo() {
		// Sanity check: must not be at end of stack and must be in an undo state.
		let squire = this.squire,
			undoIndex = this.index + 1;
		if (undoIndex < this.length && this.inUndoState) {
			this.index = undoIndex;
			squire._setHTML(this[undoIndex]);
			let range = squire._getRangeAndRemoveBookmark();
			if (range) {
				squire.setSelection(range);
			}
			this.stateChanged(/*true, undoIndex + 1 < this.length*/);
		}
	}
}

class Squire
{
	constructor(root, config) {
		this._root = root;
		this.setConfig(config);
		this._isFocused = false;
		this._lastRange = null;
		this._restoreSelection = false;
		this._mayHaveZWS = false;
		this._path = "";
		this._pathRange = null;
		this._events = {}; // new Map();
		this._ignoreChange = false;
		this.editStack = new EditStack(this);
		doc.addEventListener("selectionchange", () => this._isFocused && this._updatePath(this.getSelection()));
		this.addEventListener("blur", () => this._restoreSelection = true)
			.addEventListener("pointerdown mousedown touchstart", () => this._restoreSelection = false)
			.addEventListener("focus", () => this._restoreSelection && this.setSelection(this._lastRange))
			.addEventListener("cut", onCut)
			.addEventListener("copy", onCopy)
			// Need to monitor for shift key like this, as event.shiftKey is not available in paste event.
			.addEventListener("keydown keyup", event => this.isShiftDown = event.shiftKey)
			.addEventListener("paste", onPaste)
			// On Windows you can drag an drop text. We can"t handle this ourselves, because
			// as far as I can see, there"s no way to get the drop insertion point. So just
			// save an undo state and hope for the best.
			.addEventListener("drop", event => {
				let types = event.dataTransfer.types;
				if (types.includes("text/plain") || types.includes("text/html")) {
					this.saveUndoState();
				}
			})
			.addEventListener("keydown", onKey)
			.addEventListener("pointerup keyup mouseup touchend", ()=>this.getSelection());

		this._keyHandlers = Object.create(keyHandlers);
		this._mutation = new MutationObserver(() => this._docWasChanged());
		this._mutation.observe(root, {
			childList: true,
			attributes: true,
			characterData: true,
			subtree: true
		});
		root.setAttribute("contenteditable", "true");
		this.addEventListener(
			"beforeinput",
			this._beforeInput
		);
		this.setHTML("");

		this._beforeInputTypes = {
//			insertFromPaste: event => {},
//			insertReplacementText: event => {},
			insertText: event => {
				if (isAndroid && event.data && event.data.includes("\n")) {
					event.preventDefault();
				}
			},
			// shift + enter
			insertLineBreak: event => {
				event.preventDefault();
				this.splitBlock(true);
			},
			// enter
			insertParagraph: event => {
				event.preventDefault();
				this.splitBlock(false);
			},
			insertOrderedList: event => {
				event.preventDefault();
				this.makeOrderedList();
			},
			insertUnoderedList: event => {
				event.preventDefault();
				this.makeUnorderedList();
			},
			historyUndo: event => {
				event.preventDefault();
				this.undo();
			},
			historyRedo: event => {
				event.preventDefault();
//				this.redo();
			},
			formatRemove: event => {
				event.preventDefault();
				this.setStyle();
			},
			formatSetBlockTextDirection: event => {
				event.preventDefault();
				let dir = event.data;
				this.bidi(dir === "null" ? null : dir);
			},
			formatBackColor: event => {
				event.preventDefault();
				this.setStyle({backgroundColor:event.data});
			},
			formatFontColor: event => {
				event.preventDefault();
				this.setStyle({color:event.data});
			},
			formatFontName: event => {
				event.preventDefault();
				this.setStyle({fontFamily:event.data});
			},
/*
			formatIndent: event => {
				event.preventDefault();
				this.changeIndentationLevel("increase");
			},
			formatOutdent: event => {
				event.preventDefault();
				this.changeIndentationLevel("decrease");
			},
//			deleteByCut: event => {
				this.saveUndoState();
			},
*/
			deleteContentBackward: event => {
				Backspace(this, event, this.getSelection());
			},
			deleteContentForward: event => {
				Delete(this, event, this.getSelection());
			}
		}
	}

	// https://developer.mozilla.org/en-US/docs/Web/API/InputEvent/inputType
	// https://github.com/w3c/input-events
	_beforeInput(event) {
		let type = event.inputType;
		// event.getTargetRanges();
		switch (type) {
			case "formatBold":
			case "formatItalic":
			case "formatUnderline":
			case "formatStrikeThrough":
			case "formatSuperscript":
			case "formatSubscript":
				event.preventDefault();
				this[type.slice(6).toLowerCase()]();
				break;
			case "formatJustifyFull":
			case "formatJustifyCenter":
			case "formatJustifyRight":
			case "formatJustifyLeft": {
				event.preventDefault();
				let alignment = type.slice(13).toLowerCase();
				this.setStyle({textAlign:alignment === "full" ? "justify" : alignment});
				break;
			}
			default:
				this._beforeInputTypes[type]?.(event);
		}
	}

	// --- Events

	handleEvent(event) {
		this.fireEvent(event.type, event);
	}

	fireEvent(type, event) {
		let handlers = this._events[type];
		let isFocused, l, obj;
		// UI code, especially modal views, may be monitoring for focus events and
		// immediately removing focus. In certain conditions, this can cause the
		// focus event to fire after the blur event, which can cause an infinite
		// loop. So we detect whether we're actually focused/blurred before firing.
		if (/^(?:focus|blur)/.test(type)) {
			isFocused = this._root === doc.activeElement;
			if (type === "focus") {
				if (!isFocused || this._isFocused) {
					return this;
				}
				this._isFocused = true;
			} else {
				if (isFocused || !this._isFocused) {
					return this;
				}
				this._isFocused = false;
			}
		}
		if (handlers) {
			event = event || {};
			if (event.type !== type) {
				event.type = type;
			}
			// Clone handlers array, so any handlers added/removed do not affect it.
			handlers = handlers.slice();
			l = handlers.length;
			while (l--) {
				obj = handlers[l];
				try {
					obj.handleEvent ? obj.handleEvent(event) : obj.call(this, event);
				} catch (error) {
					error.details = 'Squire: fireEvent error. Event type: ' + type;
					didError(error);
				}
			}
		}
		return this;
	}

	addEventListener(type, fn) {
		type.split(/\s+/).forEach(type=>{
			if (!fn) {
				didError({
					name: 'Squire: addEventListener with null or undefined fn',
					message: 'Event type: ' + type
				});
				return this;
			}
			let handlers = this._events[type];
			if (!handlers) {
				handlers = this._events[type] = [];
				customEvents[type] || this._root.addEventListener(type, this, {capture:true,passive:"touchstart"===type});
			}
			handlers.push(fn);
		});
		return this;
	}

	removeEventListener(type, fn) {
		const handlers = this._events[type];
		let l;
		if (handlers) {
			if (fn) {
				l = handlers.length;
				while (l--) {
					if (handlers[l] === fn) {
						handlers.splice(l, 1);
					}
				}
			} else {
				handlers.length = 0;
			}
			if (!handlers.length) {
				delete this._events[type];
				customEvents[type] || this._root.removeEventListener(type, this, true);
			}
		}
		return this;
	}

	// --- Focus

	focus() {
		this._root.focus({ preventScroll: true });
		return this;
	}

	blur() {
		this._root.blur();
		return this;
	}

	// --- Selection and bookmarking

	// --- Workaround for browsers that can't focus empty text nodes ---
	// WebKit bug: https://bugs.webkit.org/show_bug.cgi?id=15256
	_removeZWS() {
		if (this._mayHaveZWS) {
			removeZWS(this._root);
			this._mayHaveZWS = false;
		}
	}

	_saveRangeToBookmark(range) {
		let [startNode, endNode] = createBookmarkNodes(),
			temp;

		insertNodeInRange(range, startNode);
		range.collapse();
		insertNodeInRange(range, endNode);

		// In a collapsed range, the start is sometimes inserted after the end!
		if (startNode.compareDocumentPosition(endNode) & DOCUMENT_POSITION_PRECEDING) {
			startNode.id = endSelectionId;
			endNode.id = startSelectionId;
			temp = startNode;
			startNode = endNode;
			endNode = temp;
		}

		range.setStartAfter(startNode);
		range.setEndBefore(endNode);
	}

	_getRangeAndRemoveBookmark(range) {
		const root = this._root,
			start = root.querySelector('#' + startSelectionId),
			end = root.querySelector('#' + endSelectionId);

		if (start && end) {
			let startContainer = start.parentNode,
				endContainer = end.parentNode,
				startOffset = indexOf(startContainer.childNodes, start),
				endOffset = indexOf(endContainer.childNodes, end);

			if (startContainer === endContainer) {
				--endOffset;
			}

			detach(start);
			detach(end);

			range = range || doc.createRange();
			range.setStart(startContainer, startOffset);
			range.setEnd(endContainer, endOffset);

			// Merge any text nodes we split
			mergeInlines(startContainer, range);
			if (startContainer !== endContainer) {
				mergeInlines(endContainer, range);
			}

			// If we didn't split a text node, we should move into any adjacent
			// text node to current selection point
			if (range.collapsed) {
				startContainer = range.startContainer;
				if (isTextNode(startContainer)) {
					endContainer = startContainer.childNodes[range.startOffset];
					if (!endContainer || !isTextNode(endContainer)) {
						endContainer =
							startContainer.childNodes[range.startOffset - 1];
					}
					if (isTextNode(endContainer)) {
						range.setStart(endContainer, 0);
						range.collapse(true);
					}
				}
			}
		}
		return range || null;
	}

	// --- Selection and Path ---

	getSelection() {
		let sel = win.getSelection();
		let root = this._root;
		let range, startContainer, endContainer;
		// If not focused, always rely on cached range; another function may
		// have set it but the DOM is not modified until focus again
		if (this._isFocused && sel?.rangeCount) {
			range = sel.getRangeAt(0).cloneRange();
			startContainer = range.startContainer;
			endContainer = range.endContainer;
			// FF can return the range as being inside an <img>. WTF?
			if (startContainer && isLeaf(startContainer)) {
				range.setStartBefore(startContainer);
			}
			if (endContainer && isLeaf(endContainer)) {
				range.setEndBefore(endContainer);
			}
		}
		if (range && root.contains(range.commonAncestorContainer)) {
			this._lastRange = range;
		} else {
			range = this._lastRange;
			// Check the editor is in the live document; if not, the range has
			// probably been rewritten by the browser and is bogus
			if (!doc.contains(range.commonAncestorContainer)) {
				range = null;
			}
		}
		return range || createRange(root.firstChild, 0);
	}

	setSelection(range) {
		this._lastRange = range;
		// If we're setting selection, that automatically, and synchronously,
		// triggers a focus event. So just store the selection and mark it as
		// needing restore on focus.
		if (this._isFocused) {
			const selection = win.getSelection();
			if (selection) {
				selection.setBaseAndExtent(
					range.startContainer,
					range.startOffset,
					range.endContainer,
					range.endOffset
				);
			}
		} else {
			this._restoreSelection = true;
		}
		return this;
	}

	getSelectionClosest(selector) {
		return getClosest(this.getSelection().commonAncestorContainer, this._root, selector);
	}

	// --- Path

	getPath() {
		return this._path;
	}

	_updatePath(range, force) {
		const anchor = range.startContainer,
			focus = range.endContainer;
		if (force || anchor !== this._pathRange.startContainer || focus !== this._pathRange.endContainer) {
			this._pathRange = range.cloneRange();
			let node = anchor === focus ? focus : null,
				newPath = (anchor && focus) ? (node ? this._getPath(focus) : "(selection)") : "";
			if (this._path !== newPath) {
				this._path = newPath;
				this.fireEvent("pathChange", {
					path: newPath,
					element: (!node || isElement(node)) ? node : node.parentElement
				});
			}
		}
		this.fireEvent(range.collapsed ? "cursor" : "select", {
			range
		});
	}

	_getPath(node) {
		const root = this._root;
		let path = "", style;
		if (node && node !== root) {
			path = this._getPath(node.parentNode, root);
			if (isElement(node)) {
				path += (path ? '>' : '') + node.nodeName;
				if (node.id) {
					path += '#' + node.id;
				}
				if (node.dir) {
					path += '[dir=' + node.dir + ']';
				}
				if (style = node.style.cssText) {
					path += '[style=' + style + ']';
				}
			}
		}
		return path;
	}

	// --- History

	_docWasChanged() {
		nodeCategoryCache = new WeakMap();
		this._mayHaveZWS = cantFocusEmptyTextNodes;
		if (this._ignoreChange) {
			this._ignoreChange = false;
		} else {
			this.editStack.docWasChanged();
		}
	}

	/**
	 * Leaves bookmark.
	 */
	_recordUndoState(range, replace) {
		this.editStack.recordUndoState(range, replace);
	}

	saveUndoState(range) {
		this.editStack.saveUndoState(range);
	}

	undo() {
		this.editStack.undo();
//		this.focus();
	}

	redo() {
		this.editStack.redo();
//		this.focus();
	}

	// --- Get and set data

	getRoot() {
		return this._root;
	}

	_getHTML() {
		return this._root.innerHTML;
	}

	// Called by undo() and redo()
	_setHTML(html) {
		if (html !== undefined) {
			const root = this._root;
			let node = root;
			root.innerHTML = html;
			do {
				fixCursor(node);
			} while (node = getNextBlock(node, root));
			this._ignoreChange = true;
		}
	}

	getHTML(withBookMark) {
		let html, range;
		if (withBookMark) {
			range = this.getSelection();
			this._saveRangeToBookmark(range);
		}
		html = this._getHTML().replace(/\u200B/g, "");
		withBookMark && this._getRangeAndRemoveBookmark(range);
		return html;
	}

	setHTML(html) {
		const root = this._root,
			// Parse HTML into DOM tree
			frag = this._config.sanitizeToDOMFragment(html, false);

		cleanTree(frag);
		cleanupBRs(frag, root, false);

		fixContainer(frag, root);

		// Fix cursor
		let node, walker = getBlockWalker(frag, root);
		while ((node = walker.nextNode()) && node !== root) {
			fixCursor(node);
		}

		// Don't fire an input event
		this._ignoreChange = true;

		if (root.replaceChildren) {
			root.replaceChildren(frag);
		} else {
			// Safari < 14
			while (node = root.lastChild) detach(node);
			root.append(frag);
		}
		fixCursor(root);

		// Reset the undo stack
		this.editStack.clear();

		// Record undo state
		const range = this._getRangeAndRemoveBookmark() || createRange(root.firstElementChild || root, 0);
		this.saveUndoState(range);
		this.setRange(range);
		return this;
	}

	/**
	 * Insert HTML at the cursor location. If the selection is not collapsed
	 * insertTreeFragmentIntoRange will delete the selection so that it is
	 * replaced by the html being inserted.
	 */
	insertHTML(html, isPaste) {
		let range = this.getSelection();

		// Edge doesn't just copy the fragment, but includes the surrounding guff
		// including the full <head> of the page. Need to strip this out.
		if (isPaste) {
			let startFragmentIndex = html.indexOf('<!--StartFragment-->'),
				endFragmentIndex = html.lastIndexOf('<!--EndFragment-->');
			if (startFragmentIndex > -1 && endFragmentIndex > -1) {
				html = html.slice(startFragmentIndex + 20, endFragmentIndex);
			}
		}

		let frag = this._config.sanitizeToDOMFragment(html, isPaste);

		// Record undo checkpoint
		this.saveUndoState(range);

		try {
			let root = this._root, node = frag;

			addLinks(frag, frag);
			cleanTree(frag);
			cleanupBRs(frag, root, false);
			removeEmptyInlines(frag);
			frag.normalize();

			while (node = getNextBlock(node, frag)) {
				fixCursor(node);
			}

			insertTreeFragmentIntoRange(range, frag, root);
			range.collapse();

			// After inserting the fragment, check whether the cursor is inside
			// an <a> element and if so if there is an equivalent cursor
			// position after the <a> element. If there is, move it there.
			moveRangeBoundaryOutOf(range, "A", root);

			this._ensureBottomLine();

			this.setRange(range);
			// Safari sometimes loses focus after paste. Weird.
			isPaste && this.focus();
		} catch (error) {
			didError(error);
		}
		return this;
	}

	insertElement(el, range) {
		range = range || this.getSelection();
		range.collapse(true);
		if (isInline(el)) {
			insertNodeInRange(range, el);
			range.setStartAfter(el);
		} else {
			// Get containing block node.
			let root = this._root;
			let splitNode = getStartBlockOfRange(range, root) || root;
			let parent, nodeAfterSplit;
			// While at end of container node, move up DOM tree.
			while (splitNode !== root && !splitNode.nextSibling) {
				splitNode = splitNode.parentNode;
			}
			// If in the middle of a container node, split up to root.
			if (splitNode !== root) {
				parent = splitNode.parentNode;
				nodeAfterSplit = split(parent, splitNode.nextSibling, root, root);
			}
			if (nodeAfterSplit) {
				nodeAfterSplit.before(el);
			} else {
				root.append(el);
				// Insert blank line below block.
				nodeAfterSplit = this.createDefaultBlock();
				root.append(nodeAfterSplit);
			}
			range.setStart(nodeAfterSplit, 0);
			range.setEnd(nodeAfterSplit, 0);
			moveRangeBoundariesDownTree(range);
		}
		this.focus();
		this.setSelection(range);
		this._updatePath(range);

		return this;
	}

	insertImage(src, attributes) {
		const img = createElement("IMG", mergeObjects({
			src: src
		}, attributes, true));
		this.insertElement(img);
		return img;
	}

	insertPlainText(plainText, isPaste) {
		const range = this.getSelection();
		if (range.collapsed && getClosest(range.startContainer, this._root, "PRE")) {
			let node = range.startContainer;
			let offset = range.startOffset;
			let text;
			if (!isTextNode(node)) {
				text = doc.createTextNode('');
				node?.childNodes[offset].before(text);
				node = text;
				offset = 0;
			}

			node.insertData(offset, plainText);
			range.setStart(node, offset + plainText.length);
			range.collapse(true);
			this.setSelection(range);
			return this;
		}
		const lines = plainText.split(/\r?\n/),
			closeBlock = '</' + blockTag + '>',
			openBlock = '<' + blockTag + '>';

		lines.forEach((line, i) => {
			line = escapeHTML(line).replace(/ (?=(?: |$))/g, NBSP);
			// We don't wrap the first line in the block, so if it gets inserted
			// into a blank line it keeps that line's formatting.
			// Wrap each line in <div></div>
			lines[i] = i ? openBlock + (line || '<BR>') + closeBlock : line;
		});
		return this.insertHTML(lines.join(''), isPaste);
	}

	// --- Inline formatting ---

	/**
	 * Extracts the font-family and font-size (if any) of the element
	 * holding the cursor. If there's a selection, returns an empty object.
	 */
	getFontInfo(range) {
		const fontInfo = {
			color: void 0,
			backgroundColor: void 0,
			family: void 0,
			size: void 0
		};
		range = range || this.getSelection();
		let seenAttributes = 0;
		let element = range.commonAncestorContainer, style, attr;
		if (range.collapsed || isTextNode(element)) {
			if (isTextNode(element)) {
				element = element.parentNode;
			}
			while (seenAttributes < 4 && element) {
				if (style = element.style) {
					if (!fontInfo.color && (attr = style.color)) {
						fontInfo.color = attr;
						++seenAttributes;
					}
					if (!fontInfo.backgroundColor && (attr = style.backgroundColor)) {
						fontInfo.backgroundColor = attr;
						++seenAttributes;
					}
					if (!fontInfo.family && (attr = style.fontFamily)) {
						fontInfo.family = attr;
						++seenAttributes;
					}
					if (!fontInfo.size && (attr = style.fontSize)) {
						fontInfo.size = attr;
						++seenAttributes;
					}
				}
				element = element.parentNode;
			}
		}
		return fontInfo;
	}

	/**
	 * Looks for matching tag and attributes, so won't work if <strong>
	 * instead of <b> etc.
	 */
	hasFormat(tag, attributes, range) {
		// 1. Normalise the arguments and get selection
		tag = tag.toUpperCase();
		range = range || this.getSelection()

		// Sanitize range to prevent weird IE artifacts
		if (!range.collapsed && isTextNode(range.startContainer) && range.startOffset === range.startContainer.length && range.startContainer.nextSibling) {
			range.setStartBefore(range.startContainer.nextSibling);
		}
		if (!range.collapsed && isTextNode(range.endContainer) && range.endOffset === 0 && range.endContainer.previousSibling) {
			range.setEndAfter(range.endContainer.previousSibling);
		}

		// If the common ancestor is inside the tag we require, we definitely
		// have the format.
		const root = this._root;
		const common = range.commonAncestorContainer;
		if (getNearest(common, root, tag, attributes)) {
			return true;
		}

		// If common ancestor is a text node and doesn't have the format, we
		// definitely don't have it.
		if (isTextNode(common)) {
			return false;
		}

		// Otherwise, check each text node at least partially contained within
		// the selection and make sure all of them have the format we want.
		const walker = createTreeWalker(common, SHOW_TEXT, node => isNodeContainedInRange(range, node));

		let seenNode = false;
		let node;
		while (node = walker.nextNode()) {
			if (!getNearest(node, root, tag, attributes)) {
				return false;
			}
			seenNode = true;
		}

		return seenNode;
	}

	changeFormat(add, remove, range, partial) {
		// Normalise the arguments and get selection
		range = range || this.getSelection();
		// Save undo checkpoint
		this.saveUndoState(range);

		if (remove) {
			range = this._removeFormat(remove.tag.toUpperCase(),
				remove.attributes || {}, range, partial);
		}

		if (add) {
			range = this._addFormat(add.tag.toUpperCase(),
				add.attributes || {}, range);
		}
		this.setRange(range);
		return this.focus();
	}

	_addFormat(tag, attributes, range) {
		// If the range is collapsed we simply insert the node by wrapping
		// it round the range and focus it.
		const root = this._root;
		let el, walker, startContainer, endContainer, startOffset, endOffset,
			node, block;

		if (range.collapsed) {
			el = fixCursor(createElement(tag, attributes));
			insertNodeInRange(range, el);
			range.setStart(el.firstChild, el.firstChild.length);
			range.collapse(true);

			// Clean up any previous formats that may have been set on this block
			// that are unused.
			block = el;
			while (isInline(block)) {
				block = block.parentNode;
			}
			removeZWS(block, el);
		}
		// Otherwise we find all the textnodes in the range (splitting
		// partially selected nodes) and if they're not already formatted
		// correctly we wrap them in the appropriate tag.
		else {
			// Create an iterator to walk over all the text nodes under this
			// ancestor which are in the range and not already formatted
			// correctly.
			//
			// In Blink/WebKit, empty blocks may have no text nodes, just a <br>.
			// Therefore we wrap this in the tag as well, as this will then cause it
			// to apply when the user types something in the block, which is
			// presumably what was intended.
			//
			// IMG tags are included because we may want to create a link around
			// them, and adding other styles is harmless.
			walker = createTreeWalker(
				range.commonAncestorContainer,
				SHOW_ELEMENT_OR_TEXT,
				node => (isTextNode(node) ||
							isBrElement(node) ||
							node.nodeName === "IMG"
						) && isNodeContainedInRange(range, node)
			);

			// Start at the beginning node of the range and iterate through
			// all the nodes in the range that need formatting.
			startContainer = range.startContainer;
			startOffset = range.startOffset;
			endContainer = range.endContainer;
			endOffset = range.endOffset;

			// Make sure we start with a valid node.
			walker.currentNode = startContainer;
			if (filterAccept != walker.filter.acceptNode(startContainer)) {
				startContainer = walker.nextNode();
				startOffset = 0;
			}

			// If there are interesting nodes in the selection
			if (startContainer) {
				do {
					node = walker.currentNode;
					if (!getNearest(node, root, tag, attributes)) {
						// <br> can never be a container node, so must have a text node
						// if node == (end|start)Container
						if (node === endContainer && node.length > endOffset) {
							node.splitText(endOffset);
						}
						if (node === startContainer && startOffset) {
							node = node.splitText(startOffset);
							if (endContainer === startContainer) {
								endContainer = node;
								endOffset -= startOffset;
							}
							startContainer = node;
							startOffset = 0;
						}
						el = createElement(tag, attributes);
						node.replaceWith(el);
						el.append(node);
					}
				} while (walker.nextNode());

				// If we don't finish inside a text node, offset may have changed.
				if (!isTextNode(endContainer)) {
					if (isTextNode(node)) {
						endContainer = node;
						endOffset = node.length;
					} else {
						// If <br>, we must have just wrapped it, so it must have only
						// one child
						endContainer = node.parentNode;
						endOffset = 1;
					}
				}

				// Now set the selection to as it was before
				range = createRange(
					startContainer,
					startOffset,
					endContainer,
					endOffset
				);
			}
		}
		return range;
	}

	_removeFormat(tag, attributes, range, partial) {
		// Add bookmark
		this._saveRangeToBookmark(range);

		// We need a node in the selection to break the surrounding
		// formatted text.
		let fixer;
		if (range.collapsed) {
			this._mayHaveZWS = cantFocusEmptyTextNodes;
			fixer = doc.createTextNode(cantFocusEmptyTextNodes ? ZWS : "");
			insertNodeInRange(range, fixer);
		}

		// Find block-level ancestor of selection
		let root = range.commonAncestorContainer;
		while (isInline(root)) {
			root = root.parentNode;
		}

		// Find text nodes inside formatTags that are not in selection and
		// add an extra tag with the same formatting.
		const startContainer = range.startContainer,
			startOffset = range.startOffset,
			endContainer = range.endContainer,
			endOffset = range.endOffset,
			toWrap = [],
			examineNode = (node, exemplar) => {
				// If the node is completely contained by the range then
				// we're going to remove all formatting so ignore it.
				if (isNodeContainedInRange(range, node, false)) {
					return;
				}

				let isText = isTextNode(node),
					child, next;

				// If not at least partially contained, wrap entire contents
				// in a clone of the tag we're removing and we're done.
				if (!isNodeContainedInRange(range, node)) {
					// Ignore bookmarks and empty text nodes
					if (node.nodeName !== "INPUT" && (!isText || node.data)) {
						toWrap.push([exemplar, node]);
					}
					return;
				}

				// Split any partially selected text nodes.
				if (isText) {
					if (node === endContainer && endOffset !== node.length) {
						toWrap.push([exemplar, node.splitText(endOffset)]);
					}
					if (node === startContainer && startOffset) {
						node.splitText(startOffset);
						toWrap.push([exemplar, node]);
					}
				}
				// If not a text node, recurse onto all children.
				// Beware, the tree may be rewritten with each call
				// to examineNode, hence find the next sibling first.
				else {
					for (child = node.firstChild; child; child = next) {
						next = child.nextSibling;
						examineNode(child, exemplar);
					}
				}
			},
			formatTags = Array.prototype.filter.call(
				root.getElementsByTagName(tag),
				el => isNodeContainedInRange(range, el) && hasTagAttributes(el, tag, attributes)
			);

		partial || formatTags.forEach(node => examineNode(node, node));

		// Now wrap unselected nodes in the tag
		toWrap.forEach(([el, node]) => {
			el = el.cloneNode(false);
			node.replaceWith(el);
			el.append(node);
		});
		// and remove old formatting tags.
		formatTags.forEach(el => el.replaceWith(empty(el)));

		// Merge adjacent inlines:
		if (cantFocusEmptyTextNodes && fixer) {
			fixer = fixer.parentNode;
			let block = fixer;
			while (block && isInline(block)) {
				block = block.parentNode;
			}
			if (block) {
				removeZWS(block, fixer);
			}
		}
		this._getRangeAndRemoveBookmark(range);
		fixer && range.collapse();
		mergeInlines(root, range);

		return range;
	}

	// ---
	bold() { this.toggleTag("B"); }
	italic() { this.toggleTag("I"); }
	underline() { this.toggleTag("U"); }
	strikethrough() { this.toggleTag("S"); }
	subscript() { this.toggleTag("SUB", "SUP"); }
	superscript() { this.toggleTag("SUP", "SUB"); }

	makeLink(url, attributes) {
		const range = this.getSelection();
		if (range.collapsed) {
			insertNodeInRange(
				range,
				doc.createTextNode(url.replace(/^[^:]*:\/*/, ''))
			);
		}
		attributes = mergeObjects(
			mergeObjects({
				href: url
			}, attributes, true),
			null,
			false
		);

		return this.changeFormat(
			{
				tag: "A",
				attributes: attributes
			},
			{
				tag: "A"
			},
			range
		);
	}

	removeLink() {
		return this.changeFormat(
			null,
			{
				tag: "A"
			},
			this.getSelection(),
			true
		);
	}

	// --- Block formatting

	_ensureBottomLine() {
		const root = this._root;
		const last = root.lastElementChild;
		if (!last || last.nodeName !== blockTag || !isBlock(last)) {
			root.append(this.createDefaultBlock());
		}
	}

	createDefaultBlock(children) {
		return fixCursor(
			createElement(blockTag, null, children)
		);
	}

	splitBlock(lineBreakOnly, range) {
		range = range || this.getSelection();

		const root = this._root;
		let block;
		let parent;
		let node;
		let nodeAfterSplit;

		// TODO: why was _docWasChanged() not triggered?
		this.editStack.inUndoState && this._docWasChanged();
		this._recordUndoState(range);
//		this._config.addLinks && addLinks(range.startContainer, root);
		this._removeZWS();
		this._getRangeAndRemoveBookmark(range);

		if (!range.collapsed) {
			deleteContentsOfRange(range, root);
		}
		if (this._config.addLinks) {
			moveRangeBoundariesDownTree(range);
//			const textNode = range.startContainer;
//			const offset2 = range.startOffset;
			setTimeout(() => {
//				linkifyText(this, textNode, offset2);
				addLinks(range.startContainer, this._root);
			}, 0);
		}
		block = getStartBlockOfRange(range, root);
		if (block && (parent = getClosest(block, root, "PRE"))) {
			// TODO: also check style="white-space:pre*" ?
			moveRangeBoundariesDownTree(range);
			node = range.startContainer;
			const offset2 = range.startOffset;
			if (!isTextNode(node)) {
				node = doc.createTextNode("");
				parent.insertBefore(node, parent.firstChild);
			}
			if (!lineBreakOnly && isTextNode(node) && (node.data.charAt(offset2 - 1) === "\n" || rangeDoesStartAtBlockBoundary(range, root)) && (node.data.charAt(offset2) === "\n" || rangeDoesEndAtBlockBoundary(range, root))) {
				node.deleteData(offset2 && offset2 - 1, offset2 ? 2 : 1);
				nodeAfterSplit = split(
					node,
					offset2 && offset2 - 1,
					root,
					root
				);
				node = nodeAfterSplit.previousSibling;
				if (!node.textContent) {
					detach(node);
				}
				node = this.createDefaultBlock();
				nodeAfterSplit.before(node);
				if (!nodeAfterSplit.textContent) {
					detach(nodeAfterSplit);
				}
				range.setStart(node, 0);
			} else {
				node.insertData(offset2, "\n");
				fixCursor(parent);
				// Firefox bug: if you set the selection in the text node after
				// the new line, it draws the cursor before the line break still
				// but if you set the selection to the equivalent position
				// in the parent, it works.
				if (node.length === offset2 + 1) {
					range.setStartAfter(node);
				} else {
					range.setStart(node, offset2 + 1);
				}
			}
			range.collapse(true);
			this.setRange(range);
			this._docWasChanged();
			return;
		}
		// If this is a malformed bit of document or in a table;
		// just play it safe and insert a <br>.
		if (!block || lineBreakOnly || /^T[HD]$/.test(block.nodeName)) {
			moveRangeBoundaryOutOf(range, "A", root);
			insertNodeInRange(range, createElement("BR"));
			range.collapse();
			this.setRange(range);
			return;
		}
		block = getClosest(block, root, "LI") || block;
		if (isEmptyBlock(block) && (parent = getClosest(block, root, "UL,OL,BLOCKQUOTE"))) {
			return "BLOCKQUOTE" === parent.nodeName
				// Break blockquote
				? this.modifyBlocks((/* frag */) => this.createDefaultBlock(createBookmarkNodes()), range)
				// Break list
				: this.decreaseListLevel(range);
		}
		node = range.startContainer;
		const offset = range.startOffset;
		let splitTag = tagAfterSplit[block.nodeName] || blockTag;
		nodeAfterSplit = split(
			node,
			offset,
			block.parentNode,
			root
		);
		if (!hasTagAttributes(nodeAfterSplit, splitTag)) {
			block = createElement(splitTag);
			if (nodeAfterSplit.dir) {
				block.dir = nodeAfterSplit.dir;
			}
			nodeAfterSplit.replaceWith(block);
			block.append(empty(nodeAfterSplit));
			nodeAfterSplit = block;
		}
		removeZWS(block);
		removeEmptyInlines(block);
		fixCursor(block);
		while (isElement(nodeAfterSplit)) {
			let child = nodeAfterSplit.firstChild;
			let next;
			if (nodeAfterSplit.nodeName === "A" && (!nodeAfterSplit.textContent || nodeAfterSplit.textContent === ZWS)) {
				child = doc.createTextNode("");
				nodeAfterSplit.replaceWith(child);
				nodeAfterSplit = child;
				break;
			}
			while (isTextNode(child) && !child.data) {
				next = child.nextSibling;
				if (!next || isBrElement(next)) {
					break;
				}
				detach(child);
				child = next;
			}
			if (!child || isBrElement(child) || isTextNode(child)) {
				break;
			}
			nodeAfterSplit = child;
		}
		range = createRange(nodeAfterSplit, 0);
		this.setRange(range);
	}

	modifyBlocks(modify, range) {
		range = range || this.getSelection();
		// 1. Save undo checkpoint and bookmark selection
		this._recordUndoState(range, true);

		const root = this._root;

		// 2. Expand range to block boundaries
		expandRangeToBlockBoundaries(range, root);

		// 3. Remove range.
		moveRangeBoundariesUpTree(range, root, root, root);
		const frag = extractContentsOfRange(range, root, root);

		// 4. Modify tree of fragment and reinsert.
		insertNodeInRange(range, modify.call(this, frag));

		// 5. Merge containers at edges
		if (range.endOffset < range.endContainer.childNodes.length) {
			mergeContainers(
				range.endContainer.childNodes[range.endOffset],
				root
			);
		}
		mergeContainers(
			range.startContainer.childNodes[range.startOffset],
			root
		);

		// 6. Restore selection
		this._getRangeAndRemoveBookmark(range);
		this.setRange(range);
		return this;
	}

	increaseListLevel(range) {
		range = range || this.getSelection();
		const root = this._root;
		const listSelection = getListSelection(range, root);
		if (listSelection) {
			let [list, startLi, endLi] = listSelection;
			if (startLi && startLi !== list.firstChild) {
				// Save undo checkpoint and bookmark selection
				this._recordUndoState(range, true);

				// Increase list depth
				const type = list.nodeName;
				let newParent = startLi.previousSibling;
				let next;
				if (newParent.nodeName !== type) {
					newParent = createElement(type);
					startLi.before(newParent);
				}
				do {
					next = startLi === endLi ? null : startLi.nextSibling;
					newParent.append(startLi);
				} while (startLi = next);
				next = newParent.nextSibling;
				next && mergeContainers(next, root);

				// Restore selection
				this._getRangeAndRemoveBookmark(range);
				this.setRange(range);
			}
		}
		return this.focus();
	}

	decreaseListLevel(range) {
		range = range || this.getSelection();
		const root = this._root;
		const listSelection = getListSelection(range, root);
		if (listSelection) {
			let list = listSelection[0];
			let startLi = listSelection[1] || list.firstChild;
			let endLi = listSelection[2] || list.lastChild;
			let newParent, next, insertBefore, makeNotList;

			// Save undo checkpoint and bookmark selection
			this._recordUndoState(range, true);

			if (startLi) {
				// Find the new parent list node
				newParent = list.parentNode;

				// Split list if necesary
				insertBefore = !endLi.nextSibling ?
					list.nextSibling :
					split(list, endLi.nextSibling, newParent, root);

				if (newParent !== root && newParent.nodeName === "LI") {
					newParent = newParent.parentNode;
					while (insertBefore) {
						next = insertBefore.nextSibling;
						endLi.append(insertBefore);
						insertBefore = next;
					}
					insertBefore = list.parentNode.nextSibling;
				}

				makeNotList = !/^[OU]L$/.test(newParent.nodeName);
				do {
					next = startLi === endLi ? null : startLi.nextSibling;
					startLi.remove();
					if (makeNotList && startLi.nodeName === "LI") {
						startLi = this.createDefaultBlock([empty(startLi)]);
					}
					newParent.insertBefore(startLi, insertBefore);
				} while (startLi = next);
			}

			list.firstChild || detach(list);

			insertBefore && mergeContainers(insertBefore, root);

			// Restore selection
			this._getRangeAndRemoveBookmark(range);
			this.setRange(range);
		}
		return this.focus();
	}

	makeUnorderedList() {
		return this.modifyBlocks(frag => makeList(this, frag, "UL")).focus();
	}

	makeOrderedList() {
		return this.modifyBlocks(frag => makeList(this, frag, "OL")).focus();
	}

	removeList() {
		return this.modifyBlocks(frag => {
			let root = this._root,
				listFrag;
			frag.querySelectorAll('UL, OL').forEach(list => {
				listFrag = empty(list);
				fixContainer(listFrag, root);
				list.replaceWith(listFrag);
			});

			frag.querySelectorAll("LI").forEach(item => {
				if (isBlock(item)) {
					item.replaceWith(this.createDefaultBlock([empty(item)]));
				} else {
					fixContainer(item, root);
					item.replaceWith(empty(item));
				}
			});

			return frag;
		}).focus();
	}

	// ---

	increaseQuoteLevel(range) {
		return this.modifyBlocks(
			frag => createElement(
				"BLOCKQUOTE",
				null,
				[frag]
			),
			range
		).focus();
	}

	decreaseQuoteLevel(range) {
		return this.modifyBlocks(
			frag => {
				Array.prototype.filter.call(frag.querySelectorAll("blockquote"), el =>
					!getClosest(el.parentNode, frag, "BLOCKQUOTE")
				).forEach(el =>
					el.replaceWith(empty(el))
				);
				return frag;
			},
			range
		).focus();
	}

	// ---

	code() {
		let range = this.getSelection();
		if (range.collapsed || isContainer(range.commonAncestorContainer)) {
			return this.modifyBlocks(frag => {
				let root = this._root;
				let output = doc.createDocumentFragment();
				let walker = getBlockWalker(frag, root);
				let node;
				// 1. Extract inline content; drop all blocks and contains.
				while (node = walker.nextNode()) {
					// 2. Replace <br> with \n in content
					node.querySelectorAll("BR").forEach(br => {
						if (!isLineBreak(br, false)) {
							detach(br);
						} else {
							br.replaceWith(doc.createTextNode("\n"));
						}
					});
					// 3. Remove <code>; its format clashes with <pre>
					node.querySelectorAll("CODE").forEach(el => detach(el));
					if (output.childNodes.length) {
						output.append(doc.createTextNode("\n"));
					}
					output.append(empty(node));
				}
				// 4. Replace nbsp with regular sp
				walker = createTreeWalker(output, SHOW_TEXT);
				while (node = walker.nextNode()) {
					node.data = node.data.replace(NBSP, ' '); // nbsp -> sp
				}
				output.normalize();
				return fixCursor(
					createElement("PRE", null, [
						output
					])
				);
			}, range).focus();
		}
		return this.changeFormat({ tag: "CODE" }, null, range);
	}

	removeCode() {
		let range = this.getSelection();
		let ancestor = range.commonAncestorContainer;
		let inPre = getClosest(ancestor, this._root, "PRE");
		if (inPre) {
			return this.modifyBlocks(frag => {
				let root = this._root;
				let pres = frag.querySelectorAll("PRE");
				let l = pres.length;
				let pre, walker, node, value, contents, index;
				while (l--) {
					pre = pres[l];
					walker = createTreeWalker(pre, SHOW_TEXT);
					while (node = walker.nextNode()) {
						value = node.data;
						value = value.replace(/ (?=)/g, NBSP); // sp -> nbsp
						contents = doc.createDocumentFragment();
						while ((index = value.indexOf("\n")) > -1) {
							contents.append(
								doc.createTextNode(value.slice(0, index))
							);
							contents.append(createElement("BR"));
							value = value.slice(index + 1);
						}
						node.before(contents);
						node.data = value;
					}
					fixContainer(pre, root);
					pre.replaceWith(empty(pre));
				}
				return frag;
			}, range).focus();
		}
		return this.changeFormat(null, { tag: "CODE" }, range);
	}

	toggleCode() {
		return (this.hasFormat("PRE") || this.hasFormat("CODE"))
			? this.removeCode()
			: this.code();
	}

	toggleTag(name, remove) {
		let range = this.getSelection();
		if (this.hasFormat(name, null, range)) {
			this.changeFormat(null, { tag: name }, range);
		} else {
			this.changeFormat({ tag: name }, remove ? { tag: remove } : null, range);
		}
	}

	// --- Formatting ---

	setStyle(style) {
		this.setAttribute("style", style);
	}

	setAttribute(name, value) {
		let range = this.getSelection();
		let start = range?.startContainer || {};
		let end = range ? range.endContainer : 0;
		// When the selection is all the text inside an element, set style on the element itself
		if ("dir" == name || (isTextNode(start) && 0 === range.startOffset && start === end && end.length === range.endOffset)) {
			this._recordUndoState(range);
			setAttributes(start.parentNode, {[name]: value});
//			this.setRange(range);
			this._docWasChanged();
		}
		// Else when it should remove the attribute
		else if (null == value) {
			this._recordUndoState(range);
			let node = getClosest(range.commonAncestorContainer, this._root, '*');
			range.collapsed
				? setAttributes(node, {[name]: value})
				: node.querySelectorAll('*').forEach(el => setAttributes(el, {[name]: value}));
//			this.setRange(range);
			this._docWasChanged();
		}
		// Else create a span element
		else {
			this.changeFormat({
				tag: "SPAN",
				attributes: {[name]: value}
			}, null, range);
		}
		return this.focus();
	}

	// ---

	changeIndentationLevel(direction) {
		let parent = this.getSelectionClosest('UL,OL,BLOCKQUOTE');
		if (parent || "increase" === direction) {
			let method = (!parent || "BLOCKQUOTE" === parent.nodeName) ? "Quote" : "List";
			return this[direction + method + "Level"]();
		}
	}

	bidi(dir) {
		return this.modifyBlocks(frag => setDirection(this, frag, dir)).focus();
	}

	setRange(range) {
		this.setSelection(range);
		// Path may have changed
		this._updatePath(range, true);
	}

	setConfig(config) {
		this._config = mergeObjects({
			addLinks: true
		}, config, true);
		return this;
	}
}

win.Squire = Squire;

})(document);

/* eslint max-len: 0 */

(doc => {

const
	removeElements = 'HEAD,LINK,META,NOSCRIPT,SCRIPT,TEMPLATE,TITLE',
	allowedElements = 'A,B,BLOCKQUOTE,BR,DIV,FONT,H1,H2,H3,H4,H5,H6,HR,IMG,LI,OL,P,SPAN,STRONG,TABLE,TD,TH,TR,U,UL',
	allowedAttributes = 'abbr,align,background,bgcolor,border,cellpadding,cellspacing,class,color,colspan,dir,face,frame,height,href,hspace,id,lang,rowspan,rules,scope,size,src,style,target,type,usemap,valign,vspace,width'.split(','),

	i18n = (str, def) => rl.i18n(str) || def,

	ctrlKey = shortcuts.getMetaKey() + ' + ',

	createElement = name => doc.createElement(name),

	tpl = createElement('template'),

	trimLines = html => html.trim().replace(/^(<div>\s*<br\s*\/?>\s*<\/div>)+/, '').trim(),
	htmlToPlain = html => rl.Utils.htmlToPlain(html).trim(),
	plainToHtml = text => rl.Utils.plainToHtml(text),

	forEachObjectValue = (obj, fn) => Object.values(obj).forEach(fn),

	getFragmentOfChildren = parent => {
		let frag = doc.createDocumentFragment();
		frag.append(...parent.childNodes);
		return frag;
	},

	SquireDefaultConfig = {
/*
		addLinks: true // allow_smart_html_links
*/
		sanitizeToDOMFragment: (html, isPaste/*, squire*/) => {
			tpl.innerHTML = (html||'')
				.replace(/<\/?(BODY|HTML)[^>]*>/gi,'')
				.replace(/<!--[^>]+-->/g,'')
				.replace(/<span[^>]*>\s*<\/span>/gi,'')
				.trim();
			tpl.querySelectorAll('a:empty,span:empty').forEach(el => el.remove());
			if (isPaste) {
				tpl.querySelectorAll(removeElements).forEach(el => el.remove());
				tpl.querySelectorAll('*').forEach(el => {
					if (!el.matches(allowedElements)) {
						el.replaceWith(getFragmentOfChildren(el));
					} else if (el.hasAttributes()) {
						[...el.attributes].forEach(attr => {
							let name = attr.name.toLowerCase();
							if (!allowedAttributes.includes(name)) {
								el.removeAttribute(name);
							}
						});
					}
				});
			}
			return tpl.content;
		}
	};

class SquireUI
{
	constructor(container) {
		const
			clr = createElement('input'),
			doClr = name => input => {
				// https://github.com/the-djmaze/snappymail/issues/826
				clr.style.left = (input.offsetLeft + input.parentNode.offsetLeft) + 'px';
				clr.style.width = input.offsetWidth + 'px';

				clr.value = '';
				clr.onchange = () => squire.setStyle({[name]:clr.value});
				// Chrome 110+ https://github.com/the-djmaze/snappymail/issues/1199
//				clr.oninput = () => squire.setStyle({[name]:clr.value});
				setTimeout(()=>clr.click(),1);
			},

			actions = {
				mode: {
					plain: {
//						html: '〈〉',
//						cmd: () => this.setMode('plain' == this.mode ? 'wysiwyg' : 'plain'),
						select: [
							[i18n('SETTINGS_GENERAL/EDITOR_HTML'),'wysiwyg'],
							[i18n('SETTINGS_GENERAL/EDITOR_PLAIN'),'plain']
						],
						cmd: s => this.setMode('plain' == s.value ? 'plain' : 'wysiwyg')
					}
				},
				font: {
					fontFamily: {
						select: {
							'sans-serif': {
								Arial: "'Nimbus Sans L', 'Liberation sans', 'Arial Unicode MS', Arial, Helvetica, Garuda, Utkal, FreeSans, sans-serif",
								Tahoma: "'Luxi Sans', Tahoma, Loma, Geneva, Meera, sans-serif",
								Trebuchet: "'DejaVu Sans Condensed', Trebuchet, 'Trebuchet MS', sans-serif",
								Lucida: "'Lucida Sans Unicode', 'Lucida Sans', 'DejaVu Sans', 'Bitstream Vera Sans', 'DejaVu LGC Sans', sans-serif",
								Verdana: "'DejaVu Sans', Verdana, Geneva, 'Bitstream Vera Sans', 'DejaVu LGC Sans', sans-serif"
							},
							monospace: {
								Courier: "'Liberation Mono', 'Courier New', FreeMono, Courier, monospace",
								Lucida: "'DejaVu Sans Mono', 'DejaVu LGC Sans Mono', 'Bitstream Vera Sans Mono', 'Lucida Console', Monaco, monospace"
							},
							sans: {
								Times: "'Nimbus Roman No9 L', 'Times New Roman', Times, FreeSerif, serif",
								Palatino: "'Bitstream Charter', 'Palatino Linotype', Palatino, Palladio, 'URW Palladio L', 'Book Antiqua', Times, serif",
								Georgia: "'URW Palladio L', Georgia, Times, serif"
							}
						},
						cmd: s => squire.setStyle({ fontFamily: s.value })
					},
					fontSize: {
						select: ['11px','13px','16px','20px','24px','30px'],
						defaultValueIndex: 2,
						cmd: s => squire.setStyle({ fontSize: s.value })
						// TODO: maybe consider using https://developer.mozilla.org/en-US/docs/Web/CSS/font-size#values
						// example:
						// select: ['xx-small', 'x-small',' small',' medium', 'large', 'x-large', 'xx-large', 'xxx-large'],
						// defaultValueIndex: 3,
					},
// 					dir: {
// 						select: [
// 							[i18n('EDITOR/DIR_LTR', 'LTR'),'ltr'],
// 							[i18n('EDITOR/DIR_RTL', 'RTL'),'rtl'],
// 							[i18n('EDITOR/DIR_AUTO', 'Auto'),'auto'],
// 							['',''],
// 						],
// 						cmd: s => {
// 							squire.setAttribute('dir', s.value || null);
// //							squire.setStyle({ 'unicode-bidi': 'plaintext' });
// 						}
// 					}
				},
				dir: {
					dir_ltr: {
						html: '⁋',
						cmd: () => squire.bidi('ltr')
					},
					dir_rtl: {
						html: '¶',
						cmd: () => squire.bidi('rtl')
					}
				},
				colors: {
					textColor: {
						html: 'A<sub>▾</sub>',
						cmd: doClr('color')
					},
					backgroundColor: {
						html: '🎨', /* ▧ */
						cmd: doClr('backgroundColor')
					},
				},
				inline: {
					bold: {
						html: 'B',
						cmd: () => this.doAction('bold'),
						key: 'B',
						matches: 'B,STRONT'
					},
					italic: {
						html: 'I',
						cmd: () => this.doAction('italic'),
						key: 'I',
						matches: 'I'
					},
					underline: {
						html: '<u>U</u>',
						cmd: () => this.doAction('underline'),
						key: 'U',
						matches: 'U'
					},
					strike: {
						html: '<s>S</s>',
						cmd: () => this.doAction('strikethrough'),
						key: 'Shift + 7',
						matches: 'S'
					},
					sub: {
						html: 'Xₙ',
						cmd: () => this.doAction('subscript'),
						key: 'Shift + 5',
						matches: 'SUB'
					},
					sup: {
						html: 'Xⁿ',
						cmd: () => this.doAction('superscript'),
						key: 'Shift + 6',
						matches: 'SUP'
					}
				},
				block: {
					ol: {
						html: '#',
						cmd: () => this.doList('OL'),
						key: 'Shift + 8',
						matches: 'OL'
					},
					ul: {
						html: '⋮',
						cmd: () => this.doList('UL'),
						key: 'Shift + 9',
						matches: 'UL'
					},
					quote: {
						html: '"',
						cmd: () => {
							let parent = squire.getSelectionClosest('UL,OL,BLOCKQUOTE')?.nodeName;
							('BLOCKQUOTE' == parent) ? squire.decreaseQuoteLevel() : squire.increaseQuoteLevel();
						},
						matches: 'BLOCKQUOTE'
					},
					indentDecrease: {
						html: '⇤',
						cmd: () => squire.changeIndentationLevel('decrease'),
						key: ']'
					},
					indentIncrease: {
						html: '⇥',
						cmd: () => squire.changeIndentationLevel('increase'),
						key: '['
					}
				},
				targets: {
					link: {
						html: '🔗',
						cmd: () => {
							let node = squire.getSelectionClosest('A'),
								url = prompt("Link", node?.href || "https://");
							if (url != null) {
								url.length ? squire.makeLink(url) : (node && squire.removeLink());
							}
						},
						matches: 'A'
					},
					imageUrl: {
						html: '🖼️',
						cmd: () => {
							let node = squire.getSelectionClosest('IMG'),
								src = prompt("Image", node?.src || "https://");
							src?.length ? squire.insertImage(src) : (node && squire.detach(node));
						},
						matches: 'IMG'
					},
					imageUpload: {
						html: '📂️',
						cmd: () => browseImage.click(),
						matches: 'IMG'
					}
				},
/*
				table: {
					// TODO
				},
*/
				changes: {
					undo: {
						html: '↶',
						cmd: () => squire.undo(),
						key: 'Z'
					},
					redo: {
						html: '↷',
						cmd: () => squire.redo(),
						key: 'Y'
					},
					source: {
						html: '👁',
						cmd: btn => {
							this.setMode('source' == this.mode ? 'wysiwyg' : 'source');
							btn.classList.toggle('active', 'source' == this.mode);
						}
					}
				},

				clear: {
					removeStyle: {
						html: '⎚',
						cmd: () => squire.setStyle()
					}
				}
			},

			plain = createElement('textarea'),
			wysiwyg = createElement('div'),
			toolbar = createElement('div'),
			browseImage = createElement('input'),
			squire = new Squire(wysiwyg, SquireDefaultConfig);

		clr.type = 'color';
		toolbar.append(clr);
		// Chrome https://github.com/the-djmaze/snappymail/issues/1199
		let clrid = 'squire-colors',
			colorlist = doc.getElementById(clrid),
			add = hex => colorlist.append(new Option(hex));
		if (!colorlist) {
			colorlist = createElement('datalist');
			colorlist.id = clrid;
			// Color blind safe Tableau 10 by Maureen Stone
			add('#4E79A7');
			add('#F28E2B');
			add('#E15759');
			add('#76B7B2');
			add('#59A14F');
			add('#EDC948');
			add('#B07AA1');
			add('#FF9DA7');
			add('#9C755F');
			add('#BAB0AC');
			doc.body.append(colorlist);
		}
		clr.setAttribute('list', clrid);

		browseImage.type = 'file';
		browseImage.accept = 'image/*';
		browseImage.style.display = 'none';
		browseImage.onchange = () => {
			if (browseImage.files.length) {
				let reader = new FileReader();
				reader.readAsDataURL(browseImage.files[0]);
				reader.onloadend = () => reader.result && squire.insertImage(reader.result);
			}
		}

		plain.className = 'squire-plain';
		wysiwyg.className = 'squire-wysiwyg';
		wysiwyg.dir = 'auto';
		this.mode = ''; // 'plain' | 'wysiwyg'
		this.container = container;
		this.squire = squire;
		this.plain = plain;
		this.wysiwyg = wysiwyg;

		dispatchEvent(new CustomEvent('squire-toolbar', {detail:{squire:this,actions:actions}}));

		toolbar.className = 'squire-toolbar btn-toolbar';
		let group, action/*, touchTap*/;
		for (group in actions) {
			let toolgroup = createElement('div');
			toolgroup.className = 'btn-group';
			toolgroup.id = 'squire-toolgroup-'+group;
			for (action in actions[group]) {
				let cfg = actions[group][action], input, ev = 'click';
				if (cfg.input) {
					input = createElement('input');
					input.type = cfg.input;
					ev = 'change';
				} else if (cfg.select) {
					input = createElement('select');
					input.className = 'btn';
					if (Array.isArray(cfg.select)) {
						cfg.select.forEach(value => {
							value = Array.isArray(value) ? value : [value, value];
							var option = new Option(value[0], value[1]);
							option.style[action] = value[1];
							input.append(option);
						});
					} else {
						Object.entries(cfg.select).forEach(([label, options]) => {
							let group = createElement('optgroup');
							group.label = label;
							Object.entries(options).forEach(([text, value]) => {
								var option = new Option(text, value);
								option.style[action] = value;
								group.append(option);
							});
							input.append(group);
						});
					}
					ev = 'input';
				} else {
					input = createElement('button');
					input.type = 'button';
					input.className = 'btn';
					input.innerHTML = cfg.html;
					input.action_cmd = cfg.cmd;
/*
					input.addEventListener('pointerdown', () => touchTap = input, {passive:true});
					input.addEventListener('pointermove', () => touchTap = null, {passive:true});
					input.addEventListener('pointercancel', () => touchTap = null);
					input.addEventListener('pointerup', e => {
						if (touchTap === input) {
							e.preventDefault();
							cfg.cmd(input);
						}
						touchTap = null;
					});
*/
				}
				input.addEventListener(ev, () => cfg.cmd(input));
				cfg.hint = i18n('EDITOR/' + action.toUpperCase());
				if (cfg.hint) {
					input.title = cfg.key ? cfg.hint + ' (' + ctrlKey + cfg.key + ')' : cfg.hint;
				} else if (cfg.key) {
					input.title = ctrlKey + cfg.key;
				}
				input.dataset.action = action;
				input.tabIndex = -1;
				cfg.input = input;
				toolgroup.append(input);
			}
			toolgroup.children.length && toolbar.append(toolgroup);
		}

		this.modeSelect = actions.mode.plain.input;

		let changes = actions.changes;
		changes.undo.input.disabled = changes.redo.input.disabled = true;
		squire.addEventListener('undoStateChange', state => {
			changes.undo.input.disabled = !state.canUndo;
			changes.redo.input.disabled = !state.canRedo;
		});

		actions.font.fontSize.input.selectedIndex = actions.font.fontSize.defaultValueIndex;

//		squire.addEventListener('focus', () => shortcuts.off());
//		squire.addEventListener('blur', () => shortcuts.on());

		container.append(toolbar, wysiwyg, plain);

		/**
		 * @param {string} fontName
		 * @return {string}
		 */
		const normalizeFontName = (fontName) => fontName.trim().replace(/(^["']*|["']*$)/g, '').trim().toLowerCase();

		/** @type {string[]} - lower cased array of available font families*/
		const fontFamiliesLowerCase = Object.values(actions.font.fontFamily.input.options).map(option => option.value.toLowerCase());

		/**
		 * A theme might have CSS like div.squire-wysiwyg[contenteditable="true"] {
		 * font-family: 'Times New Roman', Times, serif; }
		 * so let's find the best match squire.getRoot()'s font
		 * it will also help to properly handle generic font names like 'sans-serif'
		 * @type {number}
		 */
		let defaultFontFamilyIndex = 0;
		const squireRootFonts = getComputedStyle(squire.getRoot()).fontFamily.split(',').map(normalizeFontName);
		fontFamiliesLowerCase.some((family, index) => {
			const matchFound = family.split(',').some(availableFontName => {
				const normalizedFontName = normalizeFontName(availableFontName);
				return squireRootFonts.some(squireFontName => squireFontName === normalizedFontName);
			});
			if (matchFound) {
				defaultFontFamilyIndex = index;
			}
			return matchFound;
		});

		/**
		 * Instead of comparing whole 'font-family' strings,
		 * we are going to look for individual font names, because we might be
		 * editing a Draft started in another email client for example
		 *
		 * @type {Object.<string,number>}
		 */
		const fontNamesMap = {};
		/**
		 * @param {string} fontFamily
		 * @param {number} index
		 */
		const processFontFamilyString = (fontFamily, index) => {
			fontFamily.split(',').forEach(fontName => {
				const key = normalizeFontName(fontName);
				if (fontNamesMap[key] === undefined) {
					fontNamesMap[key] = index;
				}
			});
		};
		// first deal with the default font family
		processFontFamilyString(fontFamiliesLowerCase[defaultFontFamilyIndex], defaultFontFamilyIndex);
		// and now with the rest of the font families
		fontFamiliesLowerCase.forEach((fontFamily, index) => {
			if (index !== defaultFontFamilyIndex) {
				processFontFamilyString(fontFamily, index);
			}
		});

		// -----

		squire.addEventListener('pathChange', e => {

			const squireRoot = squire.getRoot();

			forEachObjectValue(actions, entries => {
				forEachObjectValue(entries, cfg => {
//					cfg.matches && cfg.input.classList.toggle('active', e.element && e.element.matches(cfg.matches));
					cfg.matches && cfg.input.classList.toggle('active', e.element && e.element.closestWithin(cfg.matches, squireRoot));
				});
			});

			if (e.element) {
				// try to find font-family and/or font-size and set "select" elements' values

				let sizeSelectedIndex = actions.font.fontSize.defaultValueIndex;
				let familySelectedIndex = defaultFontFamilyIndex;

				let elm = e.element;
				let familyFound = false;
				let sizeFound = false;
				do {
					if (!familyFound && elm.style.fontFamily) {
						familyFound = true;
						familySelectedIndex = -1; // show empty select if we don't know the font
						const fontNames = elm.style.fontFamily.split(',');
						for (let i = 0; i < fontNames.length; i++) {
							const index = fontNamesMap[normalizeFontName(fontNames[i])];
							if (index !== undefined) {
								familySelectedIndex = index;
								break;
							}
						}
					}

					if (!sizeFound && elm.style.fontSize) {
						sizeFound = true;
						// -1 is ok because it will just show a black <select>
						sizeSelectedIndex = actions.font.fontSize.select.indexOf(elm.style.fontSize);
					}

					elm = elm.parentElement;
				} while ((!familyFound || !sizeFound) && elm && elm !== squireRoot);

				actions.font.fontFamily.input.selectedIndex = familySelectedIndex;
				actions.font.fontSize.input.selectedIndex = sizeSelectedIndex;
			}
		});
/*
		squire.addEventListener('cursor', e => {
			console.dir({cursor:e.range});
		});
		squire.addEventListener('select', e => {
			console.dir({select:e.range});
		});
*/
	}

	doAction(name) {
		this.squire[name]();
		this.squire.focus();
	}

	doList(type) {
		let parent = this.squire.getSelectionClosest('UL,OL')?.nodeName,
			fn = {UL:'makeUnorderedList',OL:'makeOrderedList'};
		(parent == type) ? this.squire.removeList() : this.squire[fn[type]]();
	}
/*
	testPresenceinSelection(format, validation) {
		return validation.test(this.squire.getPath()) || this.squire.hasFormat(format);
	}
*/
	setMode(mode) {
		if (this.mode != mode) {
			let cl = this.container.classList, source = 'source' == this.mode;
			cl.remove('squire-mode-'+this.mode);
			if ('plain' == mode) {
				this.plain.value = htmlToPlain(source ? this.plain.value : this.squire.getHTML(), true);
			} else if ('source' == mode) {
				this.plain.value = this.squire.getHTML();
			} else {
				this.setData(source ? this.plain.value : plainToHtml(this.plain.value, true));
				mode = 'wysiwyg';
			}
			this.mode = mode; // 'wysiwyg' or 'plain'
			cl.add('squire-mode-'+mode);
			this.onModeChange?.();
			setTimeout(()=>this.focus(),1);
		}
		this.modeSelect.selectedIndex = 'plain' == this.mode ? 1 : 0;
	}

	on(type, fn) {
		if ('mode' == type) {
			this.onModeChange = fn;
		} else {
			this.squire.addEventListener(type, fn);
			this.plain.addEventListener(type, fn);
		}
	}

	execCommand(cmd, cfg) {
		if ('insertSignature' == cmd) {
			cfg = Object.assign({
				clearCache: false,
				isHtml: false,
				insertBefore: false,
				signature: ''
			}, cfg);

			if (cfg.clearCache) {
				this._prev_txt_sig = null;
			} else try {
				const signature = cfg.isHtml ? htmlToPlain(cfg.signature) : cfg.signature;
				if ('plain' === this.mode) {
					let
						text = this.plain.value,
						prevSignature = this._prev_txt_sig;
					if (prevSignature) {
						text = text.replace(prevSignature, '').trim();
					}
					this.plain.value = cfg.insertBefore ? '\n\n' + signature + '\n\n' + text : text + '\n\n' +  signature;
				} else {
					const squire = this.squire,
						root = squire.getRoot(),
						br = createElement('br'),
						div = createElement('div');
					div.className = 'rl-signature';
					div.innerHTML = cfg.isHtml ? cfg.signature : plainToHtml(cfg.signature);
					root.querySelectorAll('div.rl-signature').forEach(node => node.remove());
					cfg.insertBefore ? root.prepend(div) : root.append(div);
					// Move cursor above signature
					div.before(br);
					div.before(br.cloneNode());
//					squire._docWasChanged();
				}
				this._prev_txt_sig = signature;
			} catch (e) {
				console.error(e);
			}
		}
	}

	getData() {
		return 'source' == this.mode ? this.plain.value : trimLines(this.squire.getHTML());
	}

	setData(html) {
//		this.plain.value = html;
		const squire = this.squire;
		squire.setHTML(trimLines(html));
		const node = squire.getRoot(),
			range = squire.getSelection();
		range.setStart(node, 0);
		range.setEnd(node, 0);
		squire.setSelection( range );
	}

	getPlainData() {
		return this.plain.value;
	}

	setPlainData(text) {
		this.plain.value = text;
	}

	blur() {
		this.squire.blur();
	}

	focus() {
		if ('plain' == this.mode) {
			this.plain.focus();
			this.plain.setSelectionRange(0, 0);
		} else {
			this.squire.focus();
		}
	}
}

this.SquireUI = SquireUI;

})(document);
